/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <strstream.h>

#include "FieldImage.h"
#include "LogAction.h"

const float SP_pitch_length         = 105.0;
const float SP_pitch_width          = 68.0;
const float SP_pitch_margin         = 5.0;
const float SP_penalty_area_length  = 16.5;
const float SP_penalty_area_width   = 40.32;
const float SP_goal_area_length     = 5.5;
const float SP_goal_area_width      = 18.32;
const float SP_free_kick_buffer     = 9.15;

const int FieldImage::font_size = 12;
const int FieldImage::field_legend_buffer = 5;
const int FieldImage::horiz_legend_buffer = 4;
const int FieldImage::max_path_elements = 1000;

using namespace Magick;

FieldImage::FieldImage(int legend_lines, float scale_factor)
  : scale(scale_factor),
    img(Geometry(RoundToInt(SP_pitch_length*scale_factor),
		 RoundToInt(SP_pitch_width*scale_factor)+
		 ((legend_lines > 0) ? field_legend_buffer : 0)+
		 legend_lines*font_size),
	Color("white"))
{
  num_legend_lines = legend_lines;
  curr_legend_line = 0;
}


void FieldImage::addFieldLines(Magick::Color c)
{
  std::list<Magick::Drawable> drawList;

  drawList.push_back(DrawableStrokeColor(c)); // Outline color
  drawList.push_back(DrawableStrokeWidth(0.4)); // Stroke width
  drawList.push_back(DrawableTranslation(SP_pitch_length/2.0 * scale,
					 SP_pitch_width/2.0 * scale));
  drawList.push_back(DrawableScaling(scale, scale));

  //main lines
  drawList.push_back(DrawableRectangle(-SP_pitch_length/2.0,-SP_pitch_width/2.0,
				        SP_pitch_length/2.0, SP_pitch_width/2.0));
  drawList.push_back(DrawableLine(0, -SP_pitch_width/2.0, 0, SP_pitch_width/2.0));
  //penalty boxes
  drawList.push_back(DrawableRectangle(-SP_pitch_length/2.0,
				       -SP_penalty_area_width/2,
				       -SP_pitch_length/2.0 + SP_penalty_area_length,
				       SP_penalty_area_width/2));
  drawList.push_back(DrawableRectangle(SP_pitch_length/2.0 - SP_penalty_area_length,
				       -SP_penalty_area_width/2,
				       SP_pitch_length/2.0,
				       SP_penalty_area_width/2));
  //goalie_boxes
  drawList.push_back(DrawableRectangle(-SP_pitch_length/2.0,
				       -SP_goal_area_width/2,
				       -SP_pitch_length/2.0 + SP_goal_area_length,
				       SP_goal_area_width/2));
  drawList.push_back(DrawableRectangle(SP_pitch_length/2.0 - SP_goal_area_length,
				       -SP_goal_area_width/2,
				       SP_pitch_length/2.0,
				       SP_goal_area_width/2));
  //center circle
  drawList.push_back(DrawableCircle(0,0, 0 - SP_free_kick_buffer, 0));

  img.draw(drawList);
}

bool FieldImage::writeImage(char* fn)
{
  try {
    img.write( fn );
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::writeImage error for '%s': %2", fn, s.str());
    s.freeze(0);
    return false;
  }
  return true;
}

list<Magick::Coordinate> FieldImage::computeArrowHeadPts(Vector from, Vector to)
{
  float headlen = 2;
  float headwidang = 15;
  list<Coordinate> l;
  Vector pt;

  l.push_back(Coordinate(to.x, to.y));
  
  Vector disp = from - to;
  pt = to + disp.rotate(headwidang).scaleTo(headlen);
  l.push_back(Coordinate(pt.x, pt.y));
  pt = to + disp.rotate(-headwidang).scaleTo(headlen);
  l.push_back(Coordinate(pt.x, pt.y));

  return l;
}

void FieldImage::addPointWithHalo(Vector pt, float haloRadX, float haloRadY,
				  Magick::Color c, float pscale)
{
  vector<Vector> vPts;
  vPts.push_back(pt);
  vector<pair<float,float> > vRad;
  vRad.push_back(pair<float,float>(haloRadX, haloRadY));
  addPointsWithHalo(vPts, vRad, c, pscale);
}
void FieldImage::addPointsWithHalo(vector<Vector> vPts, vector<pair<float,float> > vHaloRad,
				   Magick::Color c, float pscale)
{
  std::list<Magick::Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(c)); // Outline color
  drawList.push_back(DrawableStrokeWidth(0.4)); // Stroke width
  drawList.push_back(DrawableTranslation(SP_pitch_length/2.0 * scale,
					 SP_pitch_width/2.0 * scale));
  drawList.push_back(DrawableScaling(scale, scale));

  vector<Vector>::iterator iPts;
  vector<pair<float,float> >::iterator iRad;
  
  for(iPts = vPts.begin();
      iPts != vPts.end();
      ++iPts) {
    drawList.push_back(DrawableCircle((*iPts).x, (*iPts).y, (*iPts).x-pscale/2, (*iPts).y));
  }

  drawList.push_back(DrawableStrokeWidth(0.2));
  for(iRad = vHaloRad.begin(), iPts = vPts.begin();
      iRad != vHaloRad.end() && iPts != vPts.end();
      ++iRad, ++iPts) {
    drawList.push_back(DrawableEllipse((*iPts).x, (*iPts).y, (*iRad).first, (*iRad).second,
				       0, 360));
  }
  
  try {
    img.draw(drawList);  
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addPointsWithHalo error: %s", s.str());
    s.freeze(0);
  }  
}

void FieldImage::addArrows(vector<pair<Vector,Vector> > vLines, Magick::Color c,
			   float lscale, bool with_head)
{
  std::list<Magick::Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(c)); // Outline color
  drawList.push_back(DrawableStrokeWidth(0.4*lscale)); // Stroke width
  drawList.push_back(DrawableTranslation(SP_pitch_length/2.0 * scale,
					 SP_pitch_width/2.0 * scale));
  drawList.push_back(DrawableScaling(scale, scale));
  drawList.push_back(DrawableFillColor(c));
  for(vector<pair<Vector,Vector> >::iterator i = vLines.begin();
      i != vLines.end();
      ++i) {
    drawList.push_back(DrawableLine( (*i).first.x, (*i).first.y, (*i).second.x, (*i).second.y));
    if (with_head)
      drawList.push_back(DrawablePolygon(computeArrowHeadPts((*i).first, (*i).second)));
  }
  try {
    img.draw(drawList);
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addArrows error: %s", s.str());
    s.freeze(0);
  }      
}


void FieldImage::addPath(list<Magick::VPath> path, Magick::Color c, float lscale)
{
  std::list<Magick::Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(c)); // Outline color
  drawList.push_back(DrawableStrokeWidth(0.4*lscale)); // Stroke width
  drawList.push_back(DrawableTranslation(SP_pitch_length/2.0 * scale,
					 SP_pitch_width/2.0 * scale));
  drawList.push_back(DrawableScaling(scale, scale));
  try {
    while (!path.empty()) {
      list<Magick::VPath> l;
      list<Magick::VPath>::iterator lastpos = path.begin();
      for (int i=0; i<max_path_elements; i++) {
	if (lastpos == path.end())
	  break;
	++lastpos;
      }
      l.splice(l.begin(), path, path.begin(), lastpos);
      drawList.push_back(DrawablePath(l));
      img.draw(drawList);
      drawList.pop_back();
    }
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addPath error: %s", s.str());
    s.freeze(0);
  }      
}

  
void FieldImage::addText(const char* str, Vector pt, int point_size, Magick::Color textc) 
{
  pt *= scale;
  pt.x += SP_pitch_length/2.0 * scale;
  pt.y += SP_pitch_width/2.0 * scale;
  std::list<Magick::Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(textc));
  drawList.push_back(DrawableFillColor(textc));
  drawList.push_back(DrawableStrokeOpacity((double)0));
  drawList.push_back(DrawableFillOpacity((double)0));
  drawList.push_back(DrawablePointSize(point_size));
  drawList.push_back(DrawableText(pt.x, pt.y, str));
  try {
    img.draw(drawList);
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addText error: %s", s.str());
    s.freeze(0);
  }      
}

  


void FieldImage::addLegendLine(string str, Magick::Color textc)
{
  if (curr_legend_line >= num_legend_lines) {
    my_error("Trying to add too many legend lines! %d >= %d",
	     curr_legend_line, num_legend_lines);
    return;
  }

  if (str[0] == 0) {
    my_error("You tried to add a null string to the FieldImage (1)");
    return;
  }
  
  LogAction3(100, "FieldImage: adding legend line: '%s'", str.c_str());
  
  //we advance first because the point seems to be the baseline from where to draw
  curr_legend_line++;

  std::list<Magick::Drawable> drawList;
  /* I am pretty sure that the only color that matters is the FillColor, but this is
     not what the documentation claims. I'll set both in case this is "fixed"
     in later versions */
  drawList.push_back(DrawableStrokeColor(textc));
  drawList.push_back(DrawableFillColor(textc));
  drawList.push_back(DrawablePointSize(font_size));
  drawList.push_back(DrawableText(horiz_legend_buffer*scale,
				  RoundToInt(SP_pitch_width*scale)+
				  field_legend_buffer+curr_legend_line*font_size,
				  str));
  try {
    img.draw(drawList);
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addLegendLine error: %s", s.str());
    s.freeze(0);
  }      
}

void FieldImage::addLegendLineWithDot(string str, Magick::Color dotc,
				      Magick::Color textc)
{
  if (curr_legend_line >= num_legend_lines) {
    my_error("Trying to add too many legend lines! %d >= %d",
	     curr_legend_line, num_legend_lines);
    return;
  }

  if (str[0] == 0) {
    my_error("You tried to add a null string to the FieldImage (2)");
    return;
  }
  
  //we advance first because the point seems to be the baseline from where to draw
  curr_legend_line++;
  std::list<Magick::Drawable> drawList;
  //the little circle
  drawList.push_back(DrawableStrokeColor(dotc));
  drawList.push_back(DrawableFillColor(dotc));
  float circle_y = RoundToInt(SP_pitch_width*scale)+
    field_legend_buffer + curr_legend_line*font_size - font_size / 2.0;
  drawList.push_back(DrawableCircle(horiz_legend_buffer*scale / 2.0, circle_y,
				    (horiz_legend_buffer*scale+font_size) / 2.0,
				    circle_y));
  //the text
  /* I am pretty sure that the only color that matters is the FillColor, but this is
     not what the documentation claims. I'll set both in case this is "fixed"
     in later versions */
  drawList.push_back(DrawableStrokeColor(textc));
  drawList.push_back(DrawableFillColor(textc));
  drawList.push_back(DrawablePointSize(font_size));
  drawList.push_back(DrawableText(horiz_legend_buffer*scale,
				  RoundToInt(SP_pitch_width*scale)+
				  field_legend_buffer+curr_legend_line*font_size,
				  str));
  try {
    img.draw(drawList);
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addLegendLineWithDot error: %s", s.str());
    s.freeze(0);
  }      
}


void FieldImage::addDrawableObjects(const list<Magick::Drawable>& drawListC)
{
  list<Magick::Drawable> drawList;
  drawList.insert(drawList.end(), drawListC.begin(), drawListC.end());
  drawList.push_front(DrawableScaling(scale, scale));
  drawList.push_front(DrawableTranslation(SP_pitch_length/2.0 * scale,
					 SP_pitch_width/2.0 * scale));
  try {
    img.draw(drawList);
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addDrawableObjects error: %s", s.str());
    s.freeze(0);
  }      
}

void FieldImage::addDrawableObjects(list<Magick::Drawable>& drawList)
{
  drawList.push_front(DrawableScaling(scale, scale));
  drawList.push_front(DrawableTranslation(SP_pitch_length/2.0 * scale,
					 SP_pitch_width/2.0 * scale));
  try {
    img.draw(drawList);
  }
  catch( Exception &error_ ) {
    ostrstream s;
    s << error_.what() << ends;
    my_error("FieldImage::addDrawableObjects error: %s", s.str());
    s.freeze(0);
  }      
  drawList.pop_front();
  drawList.pop_front();  
}

