/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _FIELD_IMAGE_H_
#define _FIELD_IMAGE_H_

#include <vector>
#include "Magick++.h"
#include "geometry.h"

class FieldImage 
{
 public:
  FieldImage(int legend_lines = 0, float scale_factor = 6.0);
  ~FieldImage() {}
  FieldImage(FieldImage& fi) : scale(fi.scale), img(fi.img)  {}
  FieldImage& operator=(const FieldImage& fi)
    { scale = fi.scale; img = fi.img; return *this; }
  //the ImageMagick++ library does smart reference counting and handles = correctly

  void addFieldLines(Magick::Color c = Magick::Color("black"));
  bool writeImage(char* fn);
  void erase() { img.erase(); curr_legend_line = 0; }

  void addPoint(Vector pt,
		Magick::Color c = Magick::Color("green"), float pscale = 1.0)
  { vector<Vector> v; v.push_back(pt); addPoints(v, c, pscale); }
  void addPoints(vector<Vector> vPts,
		 Magick::Color c = Magick::Color("green"), float pscale = 1.0)
  { vector<pair<float,float> > v; addPointsWithHalo(vPts, v, c, pscale); }
  //The "halo" are extra circles surrounding the points. This is useful
  //for showing variances 
  void addPointWithHalo(Vector pt, float haloRadX, float haloRadY,
			Magick::Color c = Magick::Color("green"), float pscale = 1.0);
  void addPointsWithHalo(vector<Vector> vPts, vector<pair<float,float> > vHaloRad,
			 Magick::Color c = Magick::Color("green"), float pscale = 1.0);
  void addLine(Vector from, Vector to,
	       Magick::Color c = Magick::Color("red"), float lscale = 1.0)
  { addArrow(from, to, c, lscale, false); }
  void addLines(vector<pair<Vector,Vector> > vLines,
	       Magick::Color c = Magick::Color("red"), float lscale = 1.0)
  { addArrows(vLines, c, lscale, false); }
  void addArrow(Vector from, Vector to,
		Magick::Color c = Magick::Color("red"), float lscale = 1.0,
		bool with_head = true)
  { vector<pair<Vector,Vector> > v; v.push_back(pair<Vector,Vector>(from,to));
  addArrows(v, c, lscale, with_head); }
  void addArrows(vector<pair<Vector,Vector> > vLines,
		 Magick::Color c = Magick::Color("red"), float lscale = 1.0,
		 bool with_head = true);

  void addPath(list<Magick::VPath> path,
	       Magick::Color c = Magick::Color("blue"), float lscale = 1.0);

  void addText(const char* str, Vector pt, int point_size = 12,
	       Magick::Color textc = Magick::Color("black"));
  void addText(const char* str, Vector pt, Magick::Color textc)
  { addText(str, pt, 12, textc); }
  
  void addLegendLine(string str, Magick::Color textc = Magick::Color("black"));
  void addLegendLineWithDot(string str, Magick::Color dotc,
			    Magick::Color textc = Magick::Color("black"));
  int getLegendLinesLeft() { return num_legend_lines - curr_legend_line; }
  
  //These functions will add the appropriate scaling and translation
  void addDrawableObjects(const list<Magick::Drawable>& drawList);
  void addDrawableObjects(list<Magick::Drawable>& drawList);
  
 private:
  //note that order matters here!
  float scale;
  int num_legend_lines;
  int curr_legend_line; //0 based
  Magick::Image img;

  list<Magick::Coordinate> computeArrowHeadPts(Vector from, Vector to);

  static const int font_size;
  static const int field_legend_buffer;
  static const int horiz_legend_buffer;

  static const int max_path_elements;
};

/* utility functions */
inline Magick::Coordinate Vector2Coord(const Vector& v)
{ return Magick::Coordinate(v.x, v.y); }


#endif
