/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Itsuki Noda,	Mikhail Prokopenko, Jan Wendler 
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* This is a modified version of a file distributed with the SoccerServer system
   http://sserver.sourceforge.net/
   Patrick Riley (pfr@cs.cmu.edu) 08/2001
*/

#ifndef _H_COACH_LANG
#define _H_COACH_LANG

#include <stream.h>
#include <vector>
#include <string>

#include "region.h"
#include "coach_lang_comp.h"
#include "coach_lang_iter.h"

//using namespace std;

enum CoachMessageType {
  CMT_Meta,
  CMT_Freeform,
  CMT_Advice,
  CMT_Info,
  CMT_Define
} ;


/********* CM_Message *******************/

class CM_Message 
{
public:
  CM_Message() {}
  virtual ~CM_Message() {}

  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_Message& m)
    { m.Print(os); return os; }

  virtual CoachMessageType getType() = 0;

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedCondition(const char* name, CM_Condition* p) {}
  virtual void updateNamedDirective(const char* name, CM_Directive* p) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}
  virtual void updateNamedAction   (const char* name, CM_Action* p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_Message* clone() = 0;
  
#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi, const char* fstem) = 0;
#endif

  int getTimeRecv() { return time_recv; }
  void setTimeRecv(int time_recv) { this->time_recv = time_recv; }

protected:
  int time_recv;
} ;


/******** Meta Message ********************/

class CM_MetaMessage 
  : public CM_Message 
{
public:
  CM_MetaMessage() : CM_Message() {}
  CM_MetaMessage(vector<CM_MetaToken*>& vTokens) : CM_Message()
    { this->vTokens = vTokens; }
  ~CM_MetaMessage() { clearTokens(); }

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CoachMessageType getType() { return CMT_Meta; }

  void iterate(CoachMessageIterator* pIter, IteratorType it);
  
  CM_Message* clone();

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const char* fstem);
#endif

  vector<CM_MetaToken*>& getTokens() { return vTokens; }
  void setTokens(vector<CM_MetaToken*>& vTokens)
    { clearTokens(); this->vTokens = vTokens; }
  void clearTokens();
  void forgetTokens() { vTokens.clear(); }
  
private:
  vector<CM_MetaToken*> vTokens;
} ;


/******* Freeform Message *********************/

class CM_FreeformMessage 
  : public CM_Message 
{
public:
  CM_FreeformMessage(char* str = NULL);
  ~CM_FreeformMessage();

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CoachMessageType getType() { return CMT_Freeform; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterMessage(this); }

  CM_Message* clone() { return new CM_FreeformMessage(str); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const char* fstem);
#endif

  void setStr(char* s);
  void setStrQuoteDelim(char* s);
    
private:
  char* str;
} ;

/******** Info Message ********************/
class CM_InfoMessage 
  : public CM_Message 
{
public:
  CM_InfoMessage() : CM_Message() {}
  CM_InfoMessage(vector<CM_Token*>& vTokens) : CM_Message()
    { this->vTokens = vTokens; }
  ~CM_InfoMessage() { clearTokens(); }

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CoachMessageType getType() { return CMT_Info; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedDirective(const char* name, CM_Directive* p);
  void updateNamedRegion   (const char* name, RegionPtr p);
  void updateNamedAction   (const char* name, CM_Action* p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);
  
  CM_Message* clone();

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const char* fstem);
#endif

  vector<CM_Token*>& getTokens() { return vTokens; }
  void setTokens(vector<CM_Token*>& vTokens)
    { clearTokens(); this->vTokens = vTokens; }
  void clearTokens();
  void forgetTokens() { vTokens.clear(); }
  
private:
  vector<CM_Token*> vTokens;
} ;

/******** Advice Message ********************/
class CM_AdviceMessage 
  : public CM_Message 
{
public:
  CM_AdviceMessage() : CM_Message() {}
  CM_AdviceMessage(vector<CM_Token*>& vTokens) : CM_Message()
    { this->vTokens = vTokens; }
  ~CM_AdviceMessage() { clearTokens(); }

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CoachMessageType getType() { return CMT_Advice; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedDirective(const char* name, CM_Directive* p);
  void updateNamedRegion   (const char* name, RegionPtr p);
  void updateNamedAction   (const char* name, CM_Action* p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);  

  CM_Message* clone();

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const char* fstem);
#endif

  vector<CM_Token*>& getTokens() { return vTokens; }
  void setTokens(vector<CM_Token*>& vTokens)
    { clearTokens(); this->vTokens = vTokens; }
  void clearTokens();
  void forgetTokens() { vTokens.clear(); }
  
private:
  vector<CM_Token*> vTokens;
} ;

/******** Define Message ********************/
class CM_DefineMessage
  : public CM_Message 
{
public:
  CM_DefineMessage() : CM_Message() {}
  CM_DefineMessage(vector<CM_DefineToken*>& vTokens) : CM_Message()
    { this->vTokens = vTokens; }
  virtual ~CM_DefineMessage() {}

  virtual void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  
  CoachMessageType getType() { return CMT_Define; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedDirective(const char* name, CM_Directive* p);
  void updateNamedRegion   (const char* name, RegionPtr p);
  void updateNamedAction   (const char* name, CM_Action* p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Message* clone();

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const char* fstem);
#endif

  vector<CM_DefineToken*>& getTokens() { return vTokens; }
  void setTokens(vector<CM_DefineToken*>& vTokens)
    { clearTokens(); this->vTokens = vTokens; }
  void clearTokens();
  void forgetTokens() { vTokens.clear(); }
  
private:
  vector<CM_DefineToken*> vTokens;
};


#endif //_H_COACH_LANG
