/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Itsuki Noda,	Mikhail Prokopenko, Jan Wendler 
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* This is a modified version of a file distributed with the SoccerServer system
   http://sserver.sourceforge.net/
   Patrick Riley (pfr@cs.cmu.edu) 08/2001
*/

#ifndef _H_COACH_LANG_COMP
#define _H_COACH_LANG_COMP

#include <stream.h>
#include <vector.h>

#include "region.h"
#include "coach_lang_iter.h"
#include "region_ptr.h"

class CoachMessageEvaluator; /* defined in coach_mess_eval.h */
class CoachMessageRepository; /* defined in coach_mess.h */

/******** type definitions **********/
/* these have been moved to types_all.h */


/******** Misc Function **********************/
/* This modifies the string in place! and returns a pointer to the new string start */
char* strip_quotes(char* str);

/******** Utility Classes ********************/
//adding the number 0 to the set turns on all the bits
class UnumSet {
public:
  explicit UnumSet(int team_size = 11) { this->team_size = team_size; clear(); }

  void clear() { entries = 0; }
  void addNum(int i);
  void removeNum(int i);
  bool isNumMember(int i) const;

  void Print(ostream& out) const;
  friend ostream& operator <<(ostream & os, const UnumSet& s)
    { s.Print(os); return os; } 
private:
  int team_size;
  unsigned entries;
} ;

class BallMoveTokenSet {
public:
  BallMoveTokenSet() { clear(); }
  BallMoveTokenSet(BallMoveToken init_entry) { clear(); addToken(init_entry); }

  void clear() { entries = 0; }
  void addToken(BallMoveToken t);
  void removeToken(BallMoveToken t);
  bool isTokenMember(BallMoveToken t) const;

  void Print(ostream& out) const;
  friend ostream& operator <<(ostream & os, const BallMoveTokenSet& s)
    { s.Print(os); return os; } 
private:
  unsigned entries;
} ;

/******** MetaToken primitives ********************/
class CM_MetaToken 
{
public:
  CM_MetaToken() {}
  virtual ~CM_MetaToken() {}

  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_MetaToken& mt)
    { mt.Print(os); return os; } 

  virtual CM_MetaTokenType getType() = 0;

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_MetaToken* clone() = 0;

} ;

class CM_MetaToken_Version 
  : public CM_MetaToken 
{
public:
  CM_MetaToken_Version(float ver = 0.0) : CM_MetaToken() {this->ver = ver;}
  ~CM_MetaToken_Version() {}

  void Print(ostream& out) {out << "(ver " << ver << ")";}
  void PrintPretty(ostream& out, string line_header) 
    { out << line_header << "version: " << ver << endl; }
  CM_MetaTokenType getType() { return CMMT_Version; }

  void setVer(float argver) {ver = argver;}
  float getVer() { return ver; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterMetaToken(this); }

  CM_MetaToken* clone() { return new CM_MetaToken_Version(ver); }

private:
  float ver;
} ;

/******** Action primitives ********************/
class CM_Action 
{
public:
  CM_Action() {}
  virtual ~CM_Action() {}

  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_Action& a)
    { a.Print(os); return os; } 

  virtual CM_ActionType getType() = 0;

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}
  virtual void updateNamedAction   (const char* name, CM_Action* p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_Action* clone() = 0;

#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi) = 0;
#endif

};

class CM_ActPosition
  : public CM_Action
{
public:
  CM_ActPosition() : CM_Action(), pReg() { }
  CM_ActPosition(RegionPtr r) : CM_Action(), pReg(NULL) { setRegion(r); }
  ~CM_ActPosition() {}

  void Print(ostream& out)
    { out << "(pos " << *pReg << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "position self at: " << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_ActionType getType() { return CMA_Position; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Action* clone() { return new CM_ActPosition(pReg->clone()); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif

  /* class specific stuff */
  RegionPtr getRegion() { return pReg; }
  void    setRegion(RegionPtr r) { /* smartptr, don't need to delete! */ pReg = r; }
  //void    forgetRegion() { pReg = NULL; }
private:
  RegionPtr pReg;
} ;

class CM_ActHome
  : public CM_Action
{
public:
  CM_ActHome() : CM_Action(), pReg() {}
  CM_ActHome(RegionPtr r) : CM_Action(), pReg() { setRegion(r); }
  ~CM_ActHome() {}

  void Print(ostream& out)
    { out << "(home " << *pReg << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "home position at: " << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_ActionType getType() { return CMA_Home; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Action* clone() { return new CM_ActHome(pReg->clone()); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif

  /* class specific stuff */
  RegionPtr getRegion() { return pReg; }
  void    setRegion(RegionPtr r) { /* smartptr, don't need to delete! */ pReg = r; }
  //void    forgetRegion() { pReg = NULL; }
private:
  RegionPtr pReg;
} ;


class CM_ActBallToRegion
  : public CM_Action
{
public:
  CM_ActBallToRegion() : CM_Action(), pReg() {  }
  CM_ActBallToRegion(RegionPtr pReg, const BallMoveTokenSet& bmtset) : CM_Action()
    { this->pReg=pReg; this->bmtset = bmtset; }
  ~CM_ActBallToRegion() { /* smartptr, don't need to delete! */}

  void Print(ostream& out) { out << "(bto " << *pReg << " " << bmtset << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "ball to: use " << bmtset << " to go to:" << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_ActionType getType() { return CMA_BallToRegion; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Action* clone() { return new CM_ActBallToRegion(pReg->clone(), bmtset); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif

  /* class specific stuff */
  RegionPtr getRegion() { return pReg; }
  void    setRegion(RegionPtr r) { /* smartptr, don't need to delete! */ pReg = r; }
  //void    forgetRegion() { pReg = NULL; }

  const BallMoveTokenSet& getBMTSet() { return bmtset; }
  void setBMTSet(const BallMoveTokenSet& bmtset) { this->bmtset = bmtset; }
  void setAddBMT(BallMoveToken t) { bmtset.addToken(t); }
  void setRemoveBMT(BallMoveToken t) { bmtset.removeToken(t); }
  void setClearAllBMT() { bmtset.clear(); }
private:
  RegionPtr pReg;
  BallMoveTokenSet bmtset;
} ;

class CM_ActBallToPlayer
  : public CM_Action
{
public:
  CM_ActBallToPlayer() : CM_Action() {}
  CM_ActBallToPlayer(UnumSet& players) : CM_Action()
    { this->players = players; }
  ~CM_ActBallToPlayer() {}

  void Print(ostream& out) { out << "(bto " << players << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "ball to players: " << players << endl; }
  CM_ActionType getType() { return CMA_BallToPlayer; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterAction(this); }

  CM_Action* clone() { return new CM_ActBallToPlayer(players); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  //static const Magick::Color draw_color; //No color needed since nothing drawn
#endif

  /* class specific stuff */
  const UnumSet& getPlayers() { return players; }
  void setPlayerSet(const UnumSet& players) { this->players = players; }
  void setAddPlayer(int i) { players.addNum(i); }
  void setRemovePlayer(int i) { players.removeNum(i); }
  void setClearAllPlayer() { players.clear(); }
private:
  UnumSet players;
} ;


class CM_ActMark
  : public CM_Action
{
public:
  //CM_ActMark() : CM_Action() {}
  CM_ActMark(UnumSet& players) : CM_Action() {this->players = players;}
  ~CM_ActMark() {}

  void Print(ostream& out)
    { out << "(mark " << players << ")";}
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "mark opponent " << players << endl; }
  CM_ActionType getType() { return CMA_Mark; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterAction(this); }

  CM_Action* clone() { return new CM_ActMark(players); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  //static const Magick::Color draw_color; //No color needed since nothing drawn
#endif

  /* class specific stuff */
  const UnumSet& getPlayers() { return players; }
  void setPlayerSet(const UnumSet& players) { this->players = players; }
  void setAddPlayer(int i) { players.addNum(i); }
  void setRemovePlayer(int i) { players.removeNum(i); }
  void setClearAllPlayer() { players.clear(); }
private:
  UnumSet players;
} ;

class CM_ActMarkLinePlayer
  : public CM_Action
{
public:
  CM_ActMarkLinePlayer(UnumSet& players) : CM_Action() {this->players = players;}
  ~CM_ActMarkLinePlayer() {}

  void Print(ostream& out)
    { out << "(markl " << players << ")";}
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "mark line to opponent " << players << endl; }
  CM_ActionType getType() { return CMA_MarkLinePlayer; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterAction(this); }

  CM_Action* clone() { return new CM_ActMarkLinePlayer(players); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  //static const Magick::Color draw_color; //No color needed since nothing drawn
#endif

  /* class specific stuff */
  const UnumSet& getPlayers() { return players; }
  void setPlayerSet(const UnumSet& players) { this->players = players; }
  void setAddPlayer(int i) { players.addNum(i); }
  void setRemovePlayer(int i) { players.removeNum(i); }
  void setClearAllPlayer() { players.clear(); }
private:
  UnumSet players;
} ;

class CM_ActMarkLineRegion
  : public CM_Action
{
public:
  CM_ActMarkLineRegion() : CM_Action(), pReg() { }
  CM_ActMarkLineRegion(RegionPtr p) : CM_Action(), pReg(p) {}
  ~CM_ActMarkLineRegion() { /* smartptr, don't need to delete! */ }

  void Print(ostream& out)
    { out << "(markl " << *pReg << ")";}
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "mark line to region:" << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_ActionType getType() { return CMA_MarkLineRegion; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Action* clone() { return new CM_ActMarkLineRegion(pReg->clone()); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif

  /* class specific stuff */
  RegionPtr getRegion() { return pReg; }
  void    setRegion(RegionPtr r) { /* smartptr, don't need to delete! */pReg = r; }
  //  void    forgetRegion() { pReg = NULL; }

private:
  RegionPtr pReg;
} ;

class CM_ActOffsidesLine
  : public CM_Action
{
public:
  CM_ActOffsidesLine() : CM_Action(), pReg() {}
  CM_ActOffsidesLine(RegionPtr r) : CM_Action(), pReg() { setRegion(r); }
  ~CM_ActOffsidesLine() {}

  void Print(ostream& out)
    { out << "(oline " << *pReg << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "offsides line at: " << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_ActionType getType() { return CMA_OffsidesLine; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Action* clone() { return new CM_ActOffsidesLine(pReg->clone()); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif

  /* class specific stuff */
  RegionPtr getRegion() { return pReg; }
  void    setRegion(RegionPtr r) { /* smartptr, don't need to delete! */ pReg = r; }
  //void    forgetRegion() { pReg = NULL; }
private:
  RegionPtr pReg;
} ;

class CM_ActHetType
  : public CM_Action
{
public:
  //CM_ActHetType() : CM_Action() {player_type = 0;}
  CM_ActHetType(int player_type = 0) : CM_Action() {setPlayerType(player_type);}
  ~CM_ActHetType() {}

  void Print(ostream& out)
    { out << "(htype " << player_type << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "heterogenous type: " << player_type << endl; }
  CM_ActionType getType() { return CMA_HetType; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterAction(this); }

  CM_Action* clone() { return new CM_ActHetType(player_type); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  //static const Magick::Color draw_color; //No color needed since nothing drawn
#endif

  /* class specific stuff */
  int getPlayerType() { return player_type; }
  void setPlayerType(int player_type) { this->player_type = player_type; }
private:
  int player_type;
} ;

class CM_ActNamed
  : public CM_Action
{
public:
  CM_ActNamed(const char* name = NULL) : CM_Action() { setName(name); pAct = NULL; }
  ~CM_ActNamed() {}

  void Print(ostream& out) { out << "\"" << name << "\""; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "action named \"" << name << "\"" << endl;
      if (pAct) pAct->PrintPretty(out, line_header+" ");
      else out << line_header << " UNLINKED" << endl; }
  CM_ActionType getType() { return CMA_Named; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedAction   (const char* name, CM_Action* p)
  { if (strcmp(name, this->name) == 0) pAct = p; }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Action* clone() { return new CM_ActNamed(name); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  //static const Magick::Color draw_color; //No color needed since nothing drawn
#endif

  /* class specific */
  char* getName() { return name; }
  void  setName(const char* name)
    { if (name == NULL) this->name[0] = 0;
    else strncpy(this->name, name, max_name_len); this->name[max_name_len] = 0; }

private:
  char name[max_name_len+1];
  CM_Action* pAct;
} ;

/******** Directive ********************/
class CM_Directive
{
public:
  CM_Directive() {}
  virtual ~CM_Directive() {}
  
  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_Directive& d)
    { d.Print(os); return os; } 
  virtual CM_DirectiveType getType() = 0;

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedDirective(const char* name, CM_Directive* p) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}
  virtual void updateNamedAction   (const char* name, CM_Action* p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_Directive* clone() = 0;

#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi) = 0;
#endif
};


class CM_DirCommand
  : public CM_Directive
{
public:
  CM_DirCommand( bool positive = true, bool our_side = true,
		const UnumSet& players = UnumSet(), CM_Action* pAct = NULL);
  ~CM_DirCommand() {if (pAct) delete pAct;}

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CM_DirectiveType getType() { return CMD_Command; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pAct->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pAct->updateNamedRegion(name, p); }
  void updateNamedAction   (const char* name, CM_Action* p)
  { pAct->updateNamedAction(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Directive* clone() { return new CM_DirCommand(positive, our_side, players, pAct->clone()); }
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color_pos;
  static const Magick::Color draw_color_neg;
#endif
  
  bool isPositive() { return positive; }
  bool isNegative() { return !positive; }
  bool isOurSide()  { return our_side; }
  bool isTheirSide(){ return !our_side; }
  const UnumSet& getPlayerSet() { return players; }
  CM_Action* getAction() { return pAct; }

  void setSense(bool positive) { this->positive = positive; }
  void setOurSide(bool our_side) { this->our_side = our_side; }
  void setPlayerSet(const UnumSet& players) { this->players = players; }
  void setAddPlayer(int i) { players.addNum(i); }
  void setRemovePlayer(int i) { players.removeNum(i); }
  void setClearAllPlayer() { players.clear(); }
  void setAction(CM_Action* pAct) { if (this->pAct) delete this->pAct; this->pAct = pAct; }
  
private:
  bool positive; //true = do; false = dont
  bool our_side; //true = our; false = opp
  UnumSet players;
  CM_Action* pAct;
} ;

class CM_DirNamed
  : public CM_Directive
{
public:
  CM_DirNamed(const char* name = NULL) : CM_Directive() { setName(name); pDir = NULL; }
  ~CM_DirNamed() {}

  void Print(ostream& out) { out << "\"" << name << "\""; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "directive named \"" << name << "\"" << endl; 
      if (pDir) pDir->PrintPretty(out, line_header+" ");
      else out << line_header << " UNLINKED" << endl; }
  CM_DirectiveType getType() { return CMD_Named; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedDirective(const char* name, CM_Directive* p)
  { if (strcmp(name, this->name) == 0) pDir = p; }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Directive* clone() { return new CM_DirNamed(name); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif

  /* class specific */
  char* getName() { return name; }
  void  setName(const char* name)
    { if (name == NULL) this->name[0] = 0;
    else strncpy(this->name, name, max_name_len); this->name[max_name_len] = 0; }

private:
  char name[max_name_len+1];
  CM_Directive* pDir;
};

/******** Conditions ********************/
class CM_Condition
{
public:
  CM_Condition() {}
  virtual ~CM_Condition() {}

  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_Condition& c)
    { c.Print(os); return os; }

  virtual CM_ConditionType getType() = 0;

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedCondition(const char* name, CM_Condition* p) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_Condition* clone() = 0;

#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi) = 0;
#endif

  virtual CMeval_t eval() = 0;

  static void setEvaluator(CoachMessageEvaluator* pe);
  static CoachMessageEvaluator* getEvaluator();
  static void forgetEvaluator();
  
protected:
  static CoachMessageEvaluator* pEval;
};

class CM_CondTrue
  : public CM_Condition
{
public:
  CM_CondTrue() : CM_Condition() {}
  ~CM_CondTrue() {}

  void Print(ostream& out) { out << "(true)"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "true" << endl; }
  CM_ConditionType getType() { return CMC_True; }
  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterCondition(this); }
  CM_Condition* clone() { return new CM_CondTrue(); }
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif
  CMeval_t eval() { return true; }
};
class CM_CondFalse
  : public CM_Condition
{
public:
  CM_CondFalse() : CM_Condition() {}
  ~CM_CondFalse() {}

  void Print(ostream& out) { out << "(false)"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "false" << endl; }
  CM_ConditionType getType() { return CMC_False; }
  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterCondition(this); }
  CM_Condition* clone() { return new CM_CondFalse(); }
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif
  CMeval_t eval() { return false; }
};
class CM_CondPlayerPosition
  : public CM_Condition
{
public:
  CM_CondPlayerPosition(bool our_side = false,
			const UnumSet& players = UnumSet(),
			int min_match=1, int max_match=11,
			RegionPtr pReg = RegionPtr())
    : CM_Condition()
    {this->our_side=our_side; this->players=players;
    this->min_match = min_match; this->max_match = max_match;
    this->pReg = pReg;}
  ~CM_CondPlayerPosition() {/* smartptr, don't need to delete! */}

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CM_ConditionType getType() { return CMC_PlayerPosition; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Condition* clone()
  { return new CM_CondPlayerPosition(our_side, players,
				     min_match, max_match, pReg->clone()); }
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color_left;
  static const Magick::Color draw_color_right;
#endif

  CMeval_t eval();

  /* class specific stuff */
  bool isOurSide()  { return our_side; }
  bool isTheirSide(){ return !our_side; }
  int  getMinMatch() { return min_match; }
  int  getMaxMatch() { return max_match; }
  RegionPtr getRegion() { return pReg; }
  const UnumSet& getPlayerSet() { return players; }

  void setMinMatch(int x) { min_match = x; }
  void setMaxMatch(int x) { max_match = x; }
  void setRegion(RegionPtr r) { /* smartptr, don't need to delete! */ pReg = r; }
  //void forgetRegion() { pReg = NULL; }
  void setOurSide(bool our_side) { this->our_side = our_side; }
  void setPlayerSet(const UnumSet& players) { this->players = players; }
  void setAddPlayer(int i) { players.addNum(i); }
  void setRemovePlayer(int i) { players.removeNum(i); }
  void setClearAllPlayer() { players.clear(); }
private:
  bool our_side;
  UnumSet players;
  int min_match;
  int max_match;
  RegionPtr pReg;
};
class CM_CondBallPosition
  : public CM_Condition
{
public:
  CM_CondBallPosition(RegionPtr pReg = RegionPtr()) : CM_Condition() {this->pReg = pReg;}
  ~CM_CondBallPosition() {/* smartptr, don't need to delete! */}

  void Print(ostream& out) { out << "(bpos " << *pReg << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "is ball position at: " << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_ConditionType getType() { return CMC_BallPosition; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Condition* clone() { return new CM_CondBallPosition(pReg->clone()); }
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
  static const Magick::Color draw_color;
#endif

  CMeval_t eval();
  
  /* class specific stuff */
  RegionPtr getRegion() { return pReg; }
  void setRegion(RegionPtr r) { /* smartptr, don't need to delete! */pReg = r; }
  //void forgetRegion() { pReg = NULL; }
private:
  RegionPtr pReg;
};
class CM_CondBallOwner
  : public CM_Condition
{
public:
  CM_CondBallOwner(bool our_side = false, const UnumSet& players = UnumSet())
    : CM_Condition() {this->our_side=our_side; this->players=players; }
  ~CM_CondBallOwner() {}

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CM_ConditionType getType() { return CMC_BallOwner; }
  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterCondition(this); }
  CM_Condition* clone() { return new CM_CondBallOwner(our_side, players); }
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif
  CMeval_t eval();

  /* class specific stuff */
  bool isOurSide()  { return our_side; }
  bool isTheirSide(){ return !our_side; }
  const UnumSet& getPlayerSet() { return players; }

  void setOurSide(bool our_side) { this->our_side = our_side; }
  void setPlayerSet(const UnumSet& players) { this->players = players; }
  void setAddPlayer(int i) { players.addNum(i); }
  void setRemovePlayer(int i) { players.removeNum(i); }
  void setClearAllPlayer() { players.clear(); }
private:
  bool our_side;
  UnumSet players;

private:
};
class CM_CondPlayMode
  : public CM_Condition
{
public:
  CM_CondPlayMode(CM_PlayModeType pm = CMPM_None) : CM_Condition() {this->pm= pm;}
  ~CM_CondPlayMode() {}

  void Print(ostream& out) { out << "(playm " << CM_PlayModeType_strings[pm] << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "is play mode '"
	  << CM_PlayModeType_strings[pm] << "'" << endl; }
  CM_ConditionType getType() { return CMC_PlayMode; }
  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterCondition(this); }
  CM_Condition* clone() { return new CM_CondPlayMode(pm); }
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif
  CMeval_t eval();
  
  /* class specific stuff */
  CM_PlayModeType getPlayMode() { return pm; }
  void setRegion(CM_PlayModeType) { this->pm = pm; }
private:
  CM_PlayModeType pm;
};
class CM_CondAnd
  : public CM_Condition
{
public:
  CM_CondAnd() : CM_Condition() {}
  CM_CondAnd(vector<CM_Condition*>& vCond) : CM_Condition() {this->vCond = vCond;}
  ~CM_CondAnd() {clearConditions();}

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CM_ConditionType getType() { return CMC_And; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedRegion   (const char* name, RegionPtr p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Condition* clone();

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif

  CMeval_t eval();

  vector<CM_Condition*>& getConditions() { return vCond; }
  void setConditions(vector<CM_Condition*>& vCond) 
    { clearConditions(); this->vCond = vCond; }
  void addCondition(CM_Condition* pCond) { vCond.push_back(pCond); }
  void clearConditions();

private:
  vector<CM_Condition*> vCond;
};
class CM_CondOr
  : public CM_Condition
{
public:
  CM_CondOr() : CM_Condition() {}
  CM_CondOr(vector<CM_Condition*>& vCond) : CM_Condition() {this->vCond = vCond;}
  ~CM_CondOr() {clearConditions();}

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CM_ConditionType getType() { return CMC_Or; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedRegion   (const char* name, RegionPtr p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Condition* clone();
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif

  CMeval_t eval();

  vector<CM_Condition*>& getConditions() { return vCond; }
  void setConditions(vector<CM_Condition*>& vCond) 
    { clearConditions(); this->vCond = vCond; }
  void addCondition(CM_Condition* pCond) { vCond.push_back(pCond); }
  void clearConditions();

private:
  vector<CM_Condition*> vCond;
  
private:
};
class CM_CondNot
  : public CM_Condition
{
public:
  CM_CondNot(CM_Condition* pCond = NULL) : CM_Condition() {this->pCond = pCond;}
  ~CM_CondNot() {if (pCond) delete pCond;}

  void Print(ostream& out) { if (pCond) out << "(not " << *pCond << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "not " << endl;
    pCond->PrintPretty(out, line_header + " +"); }
  CM_ConditionType getType() { return CMC_Not; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pCond->linkNamedElements(pCMR); }
  void updateNamedCondition(const char* name, CM_Condition* p)
  { pCond->updateNamedCondition(name, p); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pCond->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Condition* clone() { return new CM_CondNot(pCond->clone()); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif

  CMeval_t eval();
  
  /* class specific stuff */
  CM_Condition* getCondition() { return pCond; }
  void setCondition(CM_Condition* pCond)
  { if (this->pCond) delete this->pCond; this->pCond = pCond; }
  void forgetCondition() { pCond = NULL; }
private:
  CM_Condition* pCond;
};
class CM_CondNamed
  : public CM_Condition
{
public:
  CM_CondNamed(const char* name = NULL) : CM_Condition() { setName(name); pCond = NULL; }
  ~CM_CondNamed() {}

  void Print(ostream& out) { out << "\"" << name << "\""; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "condition named \"" << name << "\"" << endl; 
      if (pCond) pCond->PrintPretty(out, line_header+" ");
      else out << line_header << " UNLINKED" << endl; }
  CM_ConditionType getType() { return CMC_Named; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p)
  { if (strcmp(name, this->name) == 0) pCond = p; }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Condition* clone() { return new CM_CondNamed(name); }
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif

  CMeval_t eval()
  { if (pCond == NULL) {my_error("CondNamed::eval NULL"); return 0.0;}
  return pCond->eval(); }

  /* class specific */
  char* getName() { return name; }
  void  setName(const char* name)
    { if (name == NULL) this->name[0] = 0;
    else strncpy(this->name, name, max_name_len); this->name[max_name_len] = 0; }

private:
  char name[max_name_len+1];
  CM_Condition* pCond; 
};

/******** Tokens ********************/
class CM_Token {
public:
  CM_Token() {}
  virtual ~CM_Token() {}

  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_Token& t)
    { t.Print(os); return os; } 
  virtual CM_TokenType getType() = 0;

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedCondition(const char* name, CM_Condition* p) {}
  virtual void updateNamedDirective(const char* name, CM_Directive* p) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}
  virtual void updateNamedAction   (const char* name, CM_Action* p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_Token* clone() = 0;

#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi) = 0;
#endif
};
class CM_TokClear
  : public CM_Token
{
public:
  CM_TokClear() : CM_Token() {}
  ~CM_TokClear() {}

  void Print(ostream& out) { out << "(clear)"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "CLEAR token" << endl; }
  CM_TokenType getType() { return CMT_Clear; }
  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterToken(this); }
  CM_Token* clone() { return new CM_TokClear(); }
#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif
};
class CM_TokRule
  : public CM_Token
{
public:
  CM_TokRule(int ttl, CM_Condition* pCond = NULL) : CM_Token()
    { time_to_live = ttl; this->pCond = pCond; }
  CM_TokRule(int ttl, CM_Condition* pCond, vector<CM_Directive*>& vDir) : CM_Token()
    { time_to_live = ttl; this->pCond = pCond; this->vDir = vDir; }
  ~CM_TokRule() {if (pCond) delete pCond; clearDirectives(); }

  void Print(ostream& out);
  void PrintPretty(ostream& out, string line_header);
  CM_TokenType getType() { return CMT_Rule; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedDirective(const char* name, CM_Directive* p);
  void updateNamedRegion   (const char* name, RegionPtr p);
  void updateNamedAction   (const char* name, CM_Action* p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_Token* clone();

#ifdef WITH_DRAW
  void draw(FieldImage* pfi);
#endif

  /* class specific stuff */
  int getTimeToLive() { return time_to_live; }
  void setTimeToLive(int ttl) { time_to_live = ttl; }
  CM_Condition* getCondition() { return pCond; }
  void setCondition(CM_Condition* pCond)
  { if (this->pCond) delete this->pCond; this->pCond = pCond; }
  void forgetCondition() { pCond = NULL; }
  vector<CM_Directive*>& getDirectives() { return vDir; }
  void setDirectives(vector<CM_Directive*>& vDir) 
    { clearDirectives(); this->vDir = vDir; }
  void addDirective(CM_Directive* pDir) { vDir.push_back(pDir); }
  void clearDirectives();
private:
  int time_to_live;
  CM_Condition* pCond;
  vector<CM_Directive*> vDir;
};

/******** DefineTokens ********************/

class CM_DefineToken {
public:
  CM_DefineToken() { name[0] = 0; }
  virtual ~CM_DefineToken() {}

  virtual void Print(ostream& out) = 0;
  virtual void PrintPretty(ostream& out, string line_header) = 0;
  friend ostream& operator <<(ostream & os, CM_DefineToken& t)
    { t.Print(os); return os; }

  virtual CM_DefineTokenType getType() = 0;
  
  char* getName() { return name; }
  void  setName(const char* name)
    { if (name == NULL) this->name[0] = 0;
    else strncpy(this->name, name, max_name_len); this->name[max_name_len] = 0; }

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedCondition(const char* name, CM_Condition* p) {}
  virtual void updateNamedDirective(const char* name, CM_Directive* p) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}
  virtual void updateNamedAction   (const char* name, CM_Action* p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual CM_DefineToken* clone() = 0;
  
protected:
  char name[max_name_len+1];
};

class CM_DefTokCondition
  : public CM_DefineToken
{
public:
  CM_DefTokCondition(const char* name = NULL, CM_Condition* pCond = NULL)
    : CM_DefineToken()
    { this->pCond = pCond; setName(name);}
  ~CM_DefTokCondition() { if (pCond) delete pCond; }

  void Print(ostream& out) { out << "(definec \"" << name << "\" " << *pCond << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "Condition \"" << name << "\"" << endl;
    pCond->PrintPretty(out, line_header + " "); }
  CM_DefineTokenType getType() { return CMDT_Condition; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pCond->linkNamedElements(pCMR); }
  void updateNamedCondition(const char* name, CM_Condition* p)
  { pCond->updateNamedCondition(name, p); }
  void updateNamedRegion   (const char* name, RegionPtr p)
  { pCond->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_DefineToken* clone() { return new CM_DefTokCondition(name, pCond->clone()); }

  CM_Condition* getCondition() { return pCond; }
  
private:
  CM_Condition* pCond;
};

class CM_DefTokDirective
  : public CM_DefineToken
{
public:
  CM_DefTokDirective(const char* name = NULL, CM_Directive* pDir = NULL)
    : CM_DefineToken()
    { this->pDir = pDir; setName(name);}
  ~CM_DefTokDirective() { if (pDir) delete pDir; }

  void Print(ostream& out) { out << "(defined \"" << name << "\" " << *pDir << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "Directive \"" << name << "\"" << endl;
    pDir->PrintPretty(out, line_header + " "); }
  CM_DefineTokenType getType() { return CMDT_Directive; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pDir->linkNamedElements(pCMR); }
  void updateNamedDirective(const char* name, CM_Directive* p)
  { pDir->updateNamedDirective(name, p); }
  void updateNamedRegion(const char* name, RegionPtr p)
  { pDir->updateNamedRegion(name, p); }
  void updateNamedAction(const char* name, CM_Action* p)
  { pDir->updateNamedAction(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_DefineToken* clone() { return new CM_DefTokDirective(name, pDir->clone()); }

  CM_Directive* getDirective() { return pDir; }

private:
  CM_Directive* pDir;
};

class CM_DefTokRegion
  : public CM_DefineToken
{
public:
  CM_DefTokRegion(const char* name = NULL, RegionPtr pReg = RegionPtr())
    : CM_DefineToken()
    { this->pReg = pReg; setName(name);}
  ~CM_DefTokRegion() { /* smartptr, don't need to delete! */ }

  void Print(ostream& out) { out << "(definer \"" << name << "\" " << *pReg << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "Region \"" << name << "\"" << endl;
    pReg->PrintPretty(out, line_header + " "); }
  CM_DefineTokenType getType() { return CMDT_Region; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pReg->linkNamedElements(pCMR); }
  void updateNamedRegion(const char* name, RegionPtr p)
  { pReg->updateNamedRegion(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_DefineToken* clone() { return new CM_DefTokRegion(name, pReg->clone()); }

  RegionPtr getRegion() { return pReg; }

private:
  RegionPtr pReg;
};

class CM_DefTokAction
  : public CM_DefineToken
{
public:
  CM_DefTokAction(const char* name = NULL, CM_Action* pAct = NULL)
    : CM_DefineToken()
    { this->pAct = pAct; setName(name);}
  ~CM_DefTokAction() { if (pAct) delete pAct; }

  void Print(ostream& out) { out << "(definea \"" << name << "\" " << *pAct << ")"; }
  void PrintPretty(ostream& out, string line_header)
    { out << line_header << "Action \"" << name << "\"" << endl;
    pAct->PrintPretty(out, line_header + " "); }
  CM_DefineTokenType getType() { return CMDT_Action; }

  void linkNamedElements(const CoachMessageRepository* pCMR)
  { pAct->linkNamedElements(pCMR); }
  void updateNamedRegion(const char* name, RegionPtr p)
  { pAct->updateNamedRegion(name, p); }
  void updateNamedAction(const char* name, CM_Action* p)
  { pAct->updateNamedAction(name, p); }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  CM_DefineToken* clone() { return new CM_DefTokAction(name, pAct->clone()); }

  CM_Action* getAction() { return pAct; }

private:
  CM_Action* pAct;
};

/**************************************************************************/
/* This is some code to be inlined. we had to put it here because of
   circular class references */

#include "coach_mess_eval.h"

inline void CM_Condition::setEvaluator(CoachMessageEvaluator* pe)
{ if (pEval) delete pEval; pEval = pe; }
inline CoachMessageEvaluator* CM_Condition::getEvaluator() { return pEval; }
inline void CM_Condition::forgetEvaluator() { pEval = NULL; }

inline CMeval_t CM_CondPlayerPosition::eval()
{ return pEval->evalPlayerPosition(our_side, players, min_match, max_match, pReg); }
inline CMeval_t CM_CondBallPosition::eval()
{ return pEval->evalBallPosition(pReg); }
inline CMeval_t CM_CondBallOwner::eval()
{ return pEval->evalBallOwner(our_side, players); }
inline CMeval_t CM_CondPlayMode::eval()
{ return pEval->evalPlayMode(pm); }
inline CMeval_t CM_CondAnd::eval()
{ return pEval->evalAnd(vCond); }
inline CMeval_t CM_CondOr::eval()
{ return pEval->evalOr(vCond); }
inline CMeval_t CM_CondNot::eval()
{ return pEval->evalNot(pCond); }



#endif
