/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <strstream.h>
#include <algorithm>

#include "coach_mess.h"
#include "LogAction.h"

/********** Support Classes **************/

CMRepTokenElement::CMRepTokenElement(int tokenID,
				     int timeSent, int timeReceived,
				     int timeToLive, CM_TokRule* pTok)
{
  this->tokenID = tokenID;
  this->timeSent = timeSent;
  this->timeReceived = timeReceived;
  this->timeDeath = timeSent + timeToLive;
  this->pTok = pTok;
}

ostream& operator <<(ostream & os, const CMRepTokenElement& te)
{
  os << te.tokenID << ": "
     << te.timeSent << " "
     << te.timeReceived << " "
     << te.timeDeath << " "
     << *te.pTok;
  return os;
}
void CMRepTokenElement::PrintPretty(ostream& out, string line_header) const 
{
  out << line_header
      << "ID: " << tokenID << "\t"
      << "sent: " << timeSent << "\t"
      << "received: " << timeReceived << "\t"
      << "death: " << timeDeath << endl;
  pTok->PrintPretty(out, "  ");
}


CMRepDefineTokenElement::CMRepDefineTokenElement(int tokenID,
						 int timeSent, int timeReceived,
						 CM_DefineToken* pTok)
{
  this->tokenID = tokenID;
  this->timeSent = timeSent;
  this->timeReceived = timeReceived;
  this->pTok = pTok;
}

ostream& operator <<(ostream & os, const CMRepDefineTokenElement& te)
{
  os << te.tokenID << ": "
     << te.timeSent << " "
     << te.timeReceived << " "
     << *te.pTok;
  return os;
}
void CMRepDefineTokenElement::PrintPretty(ostream& out, string line_header) const 
{
  out << line_header
      << "ID: " << tokenID << "\t"
      << "sent: " << timeSent << "\t"
      << "received: " << timeReceived << endl;
  pTok->PrintPretty(out, "  ");
}



/*********** CoachMessageRepository *************/



void CoachMessageRepository::updateNamedCondition(const char* name, CM_Condition* p)
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;

  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    (*itert)->getToken()->updateNamedCondition(name, p);
  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    (*itert)->getToken()->updateNamedCondition(name, p);
  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    iterd->second->getToken()->updateNamedCondition(name, p);
  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    iterd->second->getToken()->updateNamedCondition(name, p);
  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    iterd->second->getToken()->updateNamedCondition(name, p);
  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    iterd->second->getToken()->updateNamedCondition(name, p);
}

void CoachMessageRepository::updateNamedDirective(const char* name, CM_Directive* p)
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;

  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    (*itert)->getToken()->updateNamedDirective(name, p);
  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    (*itert)->getToken()->updateNamedDirective(name, p);
  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    iterd->second->getToken()->updateNamedDirective(name, p);
  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    iterd->second->getToken()->updateNamedDirective(name, p);
  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    iterd->second->getToken()->updateNamedDirective(name, p);
  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    iterd->second->getToken()->updateNamedDirective(name, p);
}

void CoachMessageRepository::updateNamedRegion(const char* name, RegionPtr p)
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;

  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    (*itert)->getToken()->updateNamedRegion(name, p);
  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    (*itert)->getToken()->updateNamedRegion(name, p);
  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    iterd->second->getToken()->updateNamedRegion(name, p);
  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    iterd->second->getToken()->updateNamedRegion(name, p);
  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    iterd->second->getToken()->updateNamedRegion(name, p);
  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    iterd->second->getToken()->updateNamedRegion(name, p);
}

void CoachMessageRepository::updateNamedAction(const char* name, CM_Action* p)
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;

  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    (*itert)->getToken()->updateNamedAction(name, p);
  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    (*itert)->getToken()->updateNamedAction(name, p);
  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    iterd->second->getToken()->updateNamedAction(name, p);
  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    iterd->second->getToken()->updateNamedAction(name, p);
  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    iterd->second->getToken()->updateNamedAction(name, p);
  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    iterd->second->getToken()->updateNamedAction(name, p);
}



void CoachMessageRepository::clearAdviceQueue()
{
  list<CMRepTokenElement*>::iterator iter;
  for (iter = lAdvTok.begin(); iter != lAdvTok.end(); ++iter)
    delete *iter;
  lAdvTok.clear();
}

void CoachMessageRepository::clearInfoQueue()
{
  list<CMRepTokenElement*>::iterator iter;
  for (iter = lInfTok.begin(); iter != lInfTok.end(); ++iter)
    delete *iter;
  lInfTok.clear();
}

void CoachMessageRepository::clear_define_queue(define_token_container& dq) 
{
  define_token_container::iterator iter;
  for (iter = dq.begin(); iter != dq.end(); ++iter)
    delete iter->second;
  dq.clear();
}

void CoachMessageRepository::clearDefineQueue()
{
  clear_define_queue(cDefCon);
  clear_define_queue(cDefDir);
  clear_define_queue(cDefReg);
  clear_define_queue(cDefAct);
}


/*********************************/

CM_Condition* CoachMessageRepository::lookupCondition(char* name) const
{
  define_token_container::const_iterator iter;
  iter = cDefCon.find(name);
  if (iter == cDefCon.end())
    return NULL;
  return ((CM_DefTokCondition*)(iter->second->getToken()))->getCondition();
}

CM_Directive* CoachMessageRepository::lookupDirective(char* name) const
{
  define_token_container::const_iterator iter;
  iter = cDefDir.find(name);
  if (iter == cDefDir.end())
    return NULL;
  return ((CM_DefTokDirective*)(iter->second->getToken()))->getDirective();
}

RegionPtr CoachMessageRepository::lookupRegion(char* name) const
{
  define_token_container::const_iterator iter;
  iter = cDefReg.find(name);
  if (iter == cDefReg.end())
    return RegionPtr();
  return ((CM_DefTokRegion*)(iter->second->getToken()))->getRegion();
}

CM_Action* CoachMessageRepository::lookupAction(char* name) const
{
  define_token_container::const_iterator iter;
  iter = cDefAct.find(name);
  if (iter == cDefAct.end())
    return NULL;
  return ((CM_DefTokAction*)(iter->second->getToken()))->getAction();
}


/*********************************/
CM_Condition* CoachMessageRepository::resolveCondition(CM_Condition* p) const
{
  if (p==NULL) return NULL;
  if (p->getType() != CMC_Named) return p;
  return resolveCondition(lookupCondition(((CM_CondNamed*)p)->getName()));
}

CM_Directive* CoachMessageRepository::resolveDirective(CM_Directive* p) const
{
  if (p==NULL) return NULL;
  if (p->getType() != CMD_Named) return p;
  return resolveDirective(lookupDirective(((CM_DirNamed*)p)->getName()));
}

RegionPtr       CoachMessageRepository::resolveRegion(RegionPtr p) const
{
  if (p==NULL) return RegionPtr();
  if (p->getType() != RT_Named) return p;
  return resolveRegion(lookupRegion(((RegNamed*)p.get())->getName()));
}

CM_Action*    CoachMessageRepository::resolveAction(CM_Action* p) const
{
  if (p==NULL) return NULL;
  if (p->getType() != CMA_Named) return p;
  return resolveAction(lookupAction(((CM_ActNamed*)p)->getName()));
}


/*********************************/
void CoachMessageRepository::Print(ostream& out)
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;

  out << "Current Advice Messages" << endl;
  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    out << **itert << endl;
  out << endl;

  out << "Current Info Messages" << endl;
  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    out << **itert << endl;
  out << endl;

  out << "Current Define Condition Messages" << endl;
  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    out << *(iterd->second) << endl;
  out << endl;

  out << "Current Define Directive Messages" << endl;
  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    out << *(iterd->second) << endl;
  out << endl;

  out << "Current Define Region Messages" << endl;
  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    out << *(iterd->second) << endl;
  out << endl;

  out << "Current Define Action Messages" << endl;
  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    out << *(iterd->second) << endl;
  out << endl;
}

void CoachMessageRepository::PrintPretty(ostream& out)
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;

  out << "Current Advice Messages" << endl;
  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    (*itert)->PrintPretty(out, "");
  out << endl;

  out << "Current Info Messages" << endl;
  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    (*itert)->PrintPretty(out, "");
  out << endl;

  out << "Current Define Condition Messages" << endl;
  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    (iterd->second)->PrintPretty(out, "");
  out << endl;

  out << "Current Define Directive Messages" << endl;
  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    (iterd->second)->PrintPretty(out, "");
  out << endl;

  out << "Current Define Region Messages" << endl;
  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    (iterd->second)->PrintPretty(out, "");
  out << endl;

  out << "Current Define Action Messages" << endl;
  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    (iterd->second)->PrintPretty(out, "");
  out << endl;
}

/*********************************/

bool CoachMessageRepository::addMessage(CM_Message* pMess, int timeSent, int timeReceived)
{
  //cout << "Adding message: " << *pMess << endl;
  //pMess->PrintPretty(cout, "* ");

  pMess->linkNamedElements(this);
  
  switch (pMess->getType()) {
  case CMT_Meta: {
    /* verify that this is the right message version */
    CM_MetaMessage* pMetaMess = (CM_MetaMessage*)pMess;
    vector<CM_MetaToken*>::iterator iter;
    for (iter = pMetaMess->getTokens().begin();
	 iter != pMetaMess->getTokens().end();
	 ++iter) {
      if ((*iter)->getType() == CMMT_Version) {
	float ver = ((CM_MetaToken_Version*)*iter)->getVer();
	if (fabs(ver - COACH_LANGUAGE_VERSION) > FLOAT_EPS)
	  my_error("Unsupported coach language version: got: %.2f, support: %.2f",
		   ver, COACH_LANGUAGE_VERSION);
      } else {
	my_error("Meta-Token is strange type: %d", (*iter)->getType());
      }
    }
    delete pMess;
    return true;
  } 
  case CMT_Advice: {
    CM_AdviceMessage* pAdvMess = (CM_AdviceMessage*)pMess;
    vector<CM_Token*>::iterator iter;
    for (iter = pAdvMess->getTokens().begin();
	 iter != pAdvMess->getTokens().end();
	 ++iter) {
      if ((*iter)->getType() == CMT_Clear) {
	clearAdviceQueue();
	delete *iter;
      } else if ((*iter)->getType() == CMT_Rule) {
	CM_TokRule* pRule = (CM_TokRule*)*iter;
	CMRepTokenElement* pElem =
	  new CMRepTokenElement(nextID, timeSent, timeReceived,
				pRule->getTimeToLive(), pRule);
	ostrstream ostr;
	ostr << *pElem << ends;
	LogAction3(150, "CMR: Adding advice: %.50s", ostr.str());
	lAdvTok.push_back(pElem);
	for (vector<CMRepWatcher*>::iterator i = vecWatchers.begin();
	     i != vecWatchers.end();
	     ++i)
	  (*i)->notifyAdvice(pRule);
	nextID++;
      } else {
	my_error("CMR: advice message token has bad type: %d", (*iter)->getType());
	delete *iter; //it will be forgotton below
      }
    }
    pAdvMess->forgetTokens();
    delete pMess;
    return true;
  }
  case CMT_Info: {
    CM_InfoMessage* pInfMess = (CM_InfoMessage*)pMess;
    vector<CM_Token*>::iterator iter;
    for (iter = pInfMess->getTokens().begin();
	 iter != pInfMess->getTokens().end();
	 ++iter) {
      if ((*iter)->getType() == CMT_Clear) {
	clearInfoQueue();
	delete *iter;
      } else if ((*iter)->getType() == CMT_Rule) {
	CM_TokRule* pRule = (CM_TokRule*)*iter;
	CMRepTokenElement* pElem =
	  new CMRepTokenElement(nextID, timeSent, timeReceived,
				pRule->getTimeToLive(), pRule);
	ostrstream ostr;
	ostr << *pElem << ends;
	LogAction3(150, "CMR: Adding info: %.50s", ostr.str());
	lInfTok.push_back(pElem);
	for (vector<CMRepWatcher*>::iterator i = vecWatchers.begin();
	     i != vecWatchers.end();
	     ++i)
	  (*i)->notifyInfo(pRule);
	nextID++;
      } else {
	my_error("CMR: advice message token has bad type: %d", (*iter)->getType());
	delete *iter; //it will be forgotton below
      }
    }
    pInfMess->forgetTokens();
    delete pMess;
    return true;
  }
  case CMT_Define: {
    CM_DefineMessage* pDefMess = (CM_DefineMessage*)pMess;
    vector<CM_DefineToken*>::iterator iter;
    for (iter = pDefMess->getTokens().begin();
	 iter != pDefMess->getTokens().end();
	 ++iter) {
      switch ((*iter)->getType()) {
      case CMDT_Condition: 
	if (addOrReplaceInDefQueue(cDefCon,
				   new CMRepDefineTokenElement(nextID, timeSent,
							       timeReceived, *iter)))
	  updateNamedCondition((*iter)->getName(), ((CM_DefTokCondition*)*iter)->getCondition());
	nextID++;
	break;
      case CMDT_Directive:
	if (addOrReplaceInDefQueue(cDefDir,
				  new CMRepDefineTokenElement(nextID, timeSent,
							      timeReceived, *iter)))
	  updateNamedDirective((*iter)->getName(), ((CM_DefTokDirective*)*iter)->getDirective());
	nextID++;
	break;
      case CMDT_Region:
	if (addOrReplaceInDefQueue(cDefReg,
				   new CMRepDefineTokenElement(nextID, timeSent,
							       timeReceived, *iter)))
	  updateNamedRegion((*iter)->getName(), ((CM_DefTokRegion*)*iter)->getRegion());
	nextID++;
	break;
      case CMDT_Action:
	if (addOrReplaceInDefQueue(cDefAct,
				   new CMRepDefineTokenElement(nextID, timeSent,
							       timeReceived, *iter)))
	  updateNamedAction((*iter)->getName(), ((CM_DefTokAction*)*iter)->getAction());
	nextID++;
	break;
      default:
	my_error("CMR::addMessage: Define token is bad type! %d", (*iter)->getType());
      }
    }
    pDefMess->forgetTokens();
    delete pMess;
    return true;
  }
  case CMT_Freeform:
    my_error("CoachMessageRepository does not handle freeform messages");
    break;
  default:
    my_error("CoachMessageRepository: what is message type? %d", pMess->getType());
    break;
  }
  
  delete pMess;
  return false;
}

//returns true if it had to replace
bool CoachMessageRepository::addOrReplaceInDefQueue(define_token_container& c,
						    CMRepDefineTokenElement* pElem)
{
  
  ostrstream ostr;
  ostr << *pElem << ends;
  LogAction4(150, "CMR: Adding '%s' definition: %.50s", pElem->getToken()->getName(), ostr.str());
  define_token_container::value_type v(pElem->getToken()->getName(), pElem);
  pair<define_token_container::iterator,bool> p = c.insert(v);
  if (!p.second) {
    LogAction2(160, "CMR: Replacing old definition");
    delete p.first->second;
    c.erase(p.first);
    p = c.insert(v);
    if (!p.second)
      my_error("I still couldn't insert my define token!");
    return true;
  } else {
    return false;
  }
}

int CoachMessageRepository::cleanMessageQueue(list<CMRepTokenElement*>& l,
					      int currTime, char* name)
{
  int num_removed = 0;
  list<CMRepTokenElement*>::iterator iter;
  for (iter = l.begin(); iter != l.end(); ) {
    if ((*iter)->getTimeOfDeath() <= currTime) {
      LogAction4(160, "CMR: In %s queue, message %d expired", name, (*iter)->getTokenID());
      delete (*iter);
      iter = l.erase(iter);
      num_removed++;
    } else {
      ++iter;
    }
  }
  return num_removed;
}


//removes messages which have expired  
int CoachMessageRepository::cleanMessages(int currTime)
{
  return cleanMessageQueue(lAdvTok, currTime, "advice") +
    cleanMessageQueue(lInfTok, currTime, "info");
}

/*********************************/
class eval_less_than : public unary_function<CMRepTokenElement*, bool> 
{
public:
  eval_less_than(CMeval_t min) { min_eval = min; }
  //return true for items which are less than min_eval
  bool operator()(CMRepTokenElement* p)
  { CM_Condition::getEvaluator()->resetErrorFlag();
  Region::getEvaluator()->resetErrorFlag();
  if (p->getToken()->getCondition()->eval() < min_eval)
    return true;
  if (CM_Condition::getEvaluator()->wasError() ||
      Region::getEvaluator()->wasError()) {
    //my_error("There was an error, but otherwise the conditions matched!");
    return true;
  }
  return false;
  }
  
private:
  CMeval_t min_eval;
};

//this code is taken from the definition of back_insert_iterator
class back_insert_CMRep_to_Dir_iterator {
protected:
  vector<CM_Directive*>* container;
public:
  typedef vector<CM_Directive*> container_type;
  typedef output_iterator_tag iterator_category;
  typedef void                value_type;
  typedef void                difference_type;
  typedef void                pointer;
  typedef void                reference;

  explicit back_insert_CMRep_to_Dir_iterator(vector<CM_Directive*>& x) : container(&x) {}
  back_insert_CMRep_to_Dir_iterator&
  operator=(CMRepTokenElement*& value) { 
    LogAction3(150, "rule matched: %d", value->getTokenID());
    vector<CM_Directive*> newdir = value->getToken()->getDirectives();
    copy(newdir.begin(), newdir.end(), back_inserter(*container));
    return *this;
  }
  back_insert_CMRep_to_Dir_iterator operator*() { return *this; }
  back_insert_CMRep_to_Dir_iterator operator++() { return *this; }
  back_insert_CMRep_to_Dir_iterator operator++(int) { return *this; }
};


vector<CM_Directive*> CoachMessageRepository::getCurrentMatchingAdvice(CMeval_t min_eval)
{
#ifdef OLD_CODE
  vector<CMRepTokenElement*> newvec;
  remove_copy_if(lAdvTok.begin(), lAdvTok.end(),
		 back_inserter(newvec), elt);
  cout << "Matching Advice Messages (tokens)" << endl;
  vector<CMRepTokenElement*>::iterator iter;
  for (iter = newvec.begin(); iter != newvec.end(); ++iter)
    (*iter)->PrintPretty(cout, "");
  cout << "Matching Advice Messages (directives)" << endl;
  vector<CM_Directive*>::iterator iterd;
  for (iterd = vDir.begin(); iterd != vDir.end(); ++iterd)
    (*iterd)->PrintPretty(cout, "");
#endif
  vector<CM_Directive*> vDir;
  eval_less_than elt(min_eval);
  remove_copy_if(lAdvTok.begin(), lAdvTok.end(),
		 back_insert_CMRep_to_Dir_iterator(vDir), elt);
  return vDir;
}

vector<CM_Directive*> CoachMessageRepository::getCurrentMatchingInfo(CMeval_t min_eval)
{
  vector<CM_Directive*> vDir;
  eval_less_than elt(min_eval);
  remove_copy_if(lInfTok.begin(), lInfTok.end(),
		 back_insert_CMRep_to_Dir_iterator(vDir), elt);
  return vDir;
}

void CoachMessageRepository::updateRegions()
{
  list<CMRepTokenElement*>::iterator itert;
  define_token_container::iterator iterd;
  CMI_UpdateRegion cmiter;
  
  for (itert = lAdvTok.begin(); itert != lAdvTok.end(); ++itert)
    (*itert)->getToken()->iterate(&cmiter, IT_ParentFirst);

  for (itert = lInfTok.begin(); itert != lInfTok.end(); ++itert)
    (*itert)->getToken()->iterate(&cmiter, IT_ParentFirst);

  for (iterd = cDefCon.begin(); iterd != cDefCon.end(); ++iterd)
    (iterd->second)->getToken()->iterate(&cmiter, IT_ParentFirst);

  for (iterd = cDefDir.begin(); iterd != cDefDir.end(); ++iterd)
    (iterd->second)->getToken()->iterate(&cmiter, IT_ParentFirst);

  for (iterd = cDefReg.begin(); iterd != cDefReg.end(); ++iterd)
    (iterd->second)->getToken()->iterate(&cmiter, IT_ParentFirst);

  for (iterd = cDefAct.begin(); iterd != cDefAct.end(); ++iterd)
    (iterd->second)->getToken()->iterate(&cmiter, IT_ParentFirst);
}


void CoachMessageRepository::addWatcher(CMRepWatcher* p)
{
  vecWatchers.push_back(p);
}

bool CoachMessageRepository::removeWatcher(CMRepWatcher* p)
{
  vector<CMRepWatcher*>::iterator i = vecWatchers.begin();
  vector<CMRepWatcher*>::iterator nexti = i;
  nexti++;
  for (;
       i != vecWatchers.end();
       i=nexti, nexti++) {
    if (*i == p) {
      nexti = vecWatchers.erase(i);
      return true;
    } 
  }
  return false;
}

