/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _H_COACH_MESS
#define _H_COACH_MESS

#include <stream.h>
#include <list>
//#include <set>
#include <map>

#include "coach_lang.h"
#include "utils_shared.h"

const float COACH_LANGUAGE_VERSION = 1.0;

class CMRepTokenElement {
 public:
  CMRepTokenElement()
    { tokenID = -1; timeSent = -1; timeReceived = -1; timeDeath = -1; pTok = NULL; }
  CMRepTokenElement(int tokenID, int timeSent, int timeReceived, int timeToLive,
		    CM_TokRule* pTok);
  ~CMRepTokenElement() { if (pTok) delete pTok; }

  int getTokenID() const { return tokenID; }
  int getTimeSent() const { return timeSent; }
  int getTimeReceived() const { return timeReceived; }
  int getTimeOfDeath() const { return timeDeath; }
  int getTimeToLive(int currTime) const { return timeDeath - currTime; }
  CM_TokRule* getToken() const { return pTok; }

  void forgetToken() { pTok = NULL; }
  void setToken(CM_TokRule* pTok)
    { if (this->pTok) delete this->pTok; this->pTok = pTok; }

  bool isExpired(int currTime) { return currTime >= timeDeath; }

  friend ostream& operator <<(ostream & os, const CMRepTokenElement& te);
  void PrintPretty(ostream& out, string line_header) const ;

private:
  int tokenID;
  int timeSent;
  int timeReceived;
  int timeDeath;
  CM_TokRule* pTok;
};

class CMRepDefineTokenElement {
 public:
  CMRepDefineTokenElement() { tokenID = -1; timeSent = -1; timeReceived = -1; pTok = NULL; }
  CMRepDefineTokenElement(int tokenID, int timeSent, int timeReceived,
			  CM_DefineToken* pTok);
  ~CMRepDefineTokenElement() { if (pTok) delete pTok; }

  int getTokenID() const { return tokenID; }
  int getTimeSent() const { return timeSent; }
  int getTimeReceived() const { return timeReceived; }
  CM_DefineToken* getToken() const { return pTok; }

  void forgetToken() { pTok = NULL; }
  void setToken(CM_DefineToken* pTok)
    { if (this->pTok) delete this->pTok; this->pTok = pTok; }

  friend ostream& operator <<(ostream & os, const CMRepDefineTokenElement& te);
  void PrintPretty(ostream& out, string line_header) const ;

private:
  int tokenID;
  int timeSent;
  int timeReceived;
  CM_DefineToken* pTok;
};

/* this is obsolete; it was used when the define tokens containers were sets */
struct lt_define_token 
{
  bool operator()(const CMRepDefineTokenElement* pe1, const CMRepDefineTokenElement* pe2) 
  {
    return strcmp(pe1->getToken()->getName(), pe2->getToken()->getName()) < 0;
  }
};

/************************************************************************/
class CMRepWatcher 
{
public:
  virtual void notifyAdvice(CM_TokRule*) = 0;
  virtual void notifyInfo(CM_TokRule*) = 0;
};

/************************************************************************/
class CoachMessageRepository 
{
 public:
  CoachMessageRepository() { nextID = 0; }
  ~CoachMessageRepository() { clearQueue(); }

  void clearAdviceQueue();
  void clearInfoQueue();
  void clearDefineQueue();
  void clearQueue() { clearAdviceQueue(); clearInfoQueue(); clearDefineQueue(); }

  void forgetAdviceQueue() { lAdvTok.clear(); }
  void forgetInfoQueue()   { lInfTok.clear(); }
  void forgetDefineQueue()
    { cDefCon.clear(); cDefDir.clear(); cDefReg.clear(); cDefAct.clear(); }
  void forgetQueue() { forgetAdviceQueue(); forgetInfoQueue(); forgetDefineQueue(); }

  CM_Condition* lookupCondition(char* name) const;
  CM_Directive* lookupDirective(char* name) const;
  RegionPtr     lookupRegion(char* name) const;
  CM_Action*    lookupAction(char* name) const;

  //these are like the lookup methods above except they recurse until they point to a
  // non-named object
  CM_Condition* resolveCondition(CM_Condition* pCon) const;
  CM_Directive* resolveDirective(CM_Directive* pDir) const;
  RegionPtr     resolveRegion(RegionPtr pReg) const;
  CM_Action*    resolveAction(CM_Action* pAct) const;

  friend ostream& operator <<(ostream & os, CoachMessageRepository& cmr)
    { cmr.Print(os); return os; }
  void Print(ostream& out);
  void PrintPretty(ostream& out);

  //returns whether message was succesfully put into repository
  bool addMessage(CM_Message* pMess, int timeSent, int timeReceived);
  //void addToken(CM_Token* pTok);
  //void addDefineToken(CM_DefineToken* pTok);
  
  int cleanMessages(int currTime); //removes messages which have expired

  vector<CM_Directive*> getCurrentMatchingAdvice(CMeval_t min_eval = CME_true);
  vector<CM_Directive*> getCurrentMatchingInfo(CMeval_t min_eval = CME_true);

  void updateRegions();

  void addWatcher(CMRepWatcher* p);
  bool removeWatcher(CMRepWatcher* p);
  int numWatchers() { return vecWatchers.size(); }

 private:

  typedef map<const char*, CMRepDefineTokenElement*, ltstr> define_token_container;

  //this function takes responsibility for pTok memory
  bool addOrReplaceInDefQueue(define_token_container& l,
			      CMRepDefineTokenElement* pElem);
  int cleanMessageQueue(list<CMRepTokenElement*>& l,
			int currTime, char* name);

  void clear_define_queue(define_token_container& dq);

  void updateNamedCondition(const char* name, CM_Condition* p);
  void updateNamedDirective(const char* name, CM_Directive* p);
  void updateNamedRegion   (const char* name, RegionPtr p);
  void updateNamedAction   (const char* name, CM_Action* p);
  
  list<CMRepTokenElement*>       lAdvTok;
  list<CMRepTokenElement*>       lInfTok;
  define_token_container cDefCon;
  define_token_container cDefDir;
  define_token_container cDefReg;
  define_token_container cDefAct;
  
  int nextID;

  vector<CMRepWatcher*> vecWatchers;
  
};


#endif //_H_COACH_MESS
