/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 1998,1999,2000 Patrick Riley, Peter Stone, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/* Classes dealing with statistical stuff from data points */


#include <iomanip.h>
#include <math.h>
#include "data.h"
#include "utils_shared.h"

SingleDataSummary::SingleDataSummary()
{
  Reset();
  precision = 2;
}

void SingleDataSummary::Reset() 
{
  n = 0;
  sum = 0.0;
  sum_squares = 0.0;
  min = HUGE;
  max = -HUGE;
}

void SingleDataSummary::AddPoint(double x)
{
  n++;
  sum += x;
  sum_squares += x*x;
  min = (x < min) ? x : min;
  max = (x > max) ? x : max;
}

double SingleDataSummary::GetMean()
{
  if (n==0)
    return 0.0;
  else
    return sum/n;
}
				 
double SingleDataSummary::GetVariance()
{
  if (n==1 || n==0)
    return 0;
  else
    return ((double)n*sum_squares - sum*sum) / ((double)n * (n-1));
}


double SingleDataSummary::GetStdDev()
{
  if (Round(GetVariance(), -5) < 0) {
    my_error("How is variance negative: %g", (float)GetVariance());
  }  
  return (sqrt(Round(GetVariance(),-5)));
}


int SingleDataSummary::GetNumPoints()
{
  return n;
}

void SingleDataSummary::CombineWith(SingleDataSummary* sds)
{
  n += sds->n;
  sum += sds->sum;
  sum_squares += sds->sum_squares;
}




void SingleDataSummary::WriteInfoToFile(ostream& output_file,
					bool WithNumPoints)
{
  if (WithNumPoints)
    output_file << "NumPoints: " << GetNumPoints() << "\t";
  output_file << setprecision(precision) << "Mean: " << GetMean()
	      << setprecision(precision) << "\tVariance: " << GetVariance()
	      << setprecision(precision) << "\tStdDev: " << GetStdDev()
	      << endl;
}

void SingleDataSummary::WriteCompactInfoToFile(ostream& output_file,
					       bool WithNumPoints)
{
  if (WithNumPoints)
    output_file << GetNumPoints() << "\t";
  output_file << setprecision(precision) << GetMean() << "\t"
	      << setprecision(precision) << GetVariance() << "\t"
	      << setprecision(precision) << GetStdDev();
}

void SingleDataSummary::WriteCompactMeanStdevToFile(ostream& output_file,
					       bool WithNumPoints)
{
  if (WithNumPoints)
    output_file << GetNumPoints() << "\t";
  output_file << setprecision(precision) << GetMean() << "\t"
	      << setprecision(precision) << GetStdDev();
}


Bool SingleDataSummary::ReadInfoFromFile(istream& infile, int num_points)
{
  char line[100], *pc;
  infile.getline(line, 100);
  if (!infile)
    return FALSE;
  
  pc = line;
  
  if (num_points == -1)
    if (!GetIntOffLine(&pc, &num_points))
      return FALSE;

  n = num_points;
  double mean, var, stdev;
  if (!GetDoubleOffLine(&pc, &mean) ||
      !GetDoubleOffLine(&pc, &var) ||
      !GetDoubleOffLine(&pc, &stdev))
    return FALSE;

  sum = mean * n;
  sum_squares = (var * n * (n-1) + sum*sum) / n;

  return TRUE;
}

Bool SingleDataSummary::ReadCompactInfoFromFile(istream& infile, int num_points)
{
  char line[100], *pc;
  infile.getline(line, 100);
  if (!infile)
    return FALSE;

  pc = line;

  if (num_points == -1)
    num_points = get_int(&pc);

  n = num_points;
  double mean, var, stdev;
  mean = get_double(&pc);
  var  = get_double(&pc);
  stdev = get_double(&pc);

  sum = mean * n;
  sum_squares = (var * n * (n-1) + sum*sum) / n;

  /*printf("line: %s, I got: %d %g %g %g, %g %g %g\n", line,
	 n, mean, var, stdev, 
	 GetMean(), GetVariance(), GetStdDev());*/
  
  return TRUE;
}

/***********************************************************/


//Only maintains info for max and mean
void MaxMeanSummary::Reset()
{
  sum = 0.0;
  n = 0;
  max = -HUGE;
}
  
void MaxMeanSummary::AddPoint(double x)
{
  sum += x;
  n++;
  if (x > max)
    max = x;
}

double MaxMeanSummary::GetMean()
{
  if (n == 0) {
    my_error("Tried to get mean with 0 points!");
    return 0.0;
  } 
  return sum / n;
}

/***********************************************************/

void MinMaxSummary::AddPoint(double x)
{
  if (x < min) {
    min = x;
    mindex = curr_index;
  }
  if (x > max) {
    max = x;
    maxdex = curr_index;
  }
  curr_index++;
}

