/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 1998,1999,2000 Patrick Riley, Peter Stone, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream.h>
#include <iomanip.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "geometry.h"
#include "data.h"
#include "LogAction.h"

#ifdef DEBUG_OUTPUT
#define DebugGeom(x) 
#define DebugSim(x) 
#else
#define DebugGeom(x)
#define DebugSim(x) 
#endif

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                           Vector Class                                       */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

/* reads in an optional '(', the x val, an optional ',', the y val, 
   and then an optional ')' */
istream& operator >>(istream & is, Vector &v)
{
  char ch;  
  Bool paren = FALSE;
  
  while (isspace(ch = is.get())) ;
  if (ch != '(') is.putback(ch);
  else paren = TRUE;  
  is >> v.x;
  while (isspace(ch = is.get())) ;
  if (ch != ',') is.putback(ch);
  is >> v.y;
  if (paren) {
    while (isspace(ch = is.get())) ;
    if (ch != ')') {
      my_error("Vector read: open paren, but no close");    
      is.putback(ch);
    }
  }  
  return is;  
}

void Vector::PrintToStr(char* str) const
{
  sprintf(str, "(%.2f, %.2f)", x, y);  
}


void Vector::PrintToBinStr(BinaryString* pBS, int bits, int shift_constant) const
{
  pBS->PutInt(closest_int(x+shift_constant), bits);
  pBS->PutInt(closest_int(y+shift_constant), bits);
}

bool Vector::ReadFromBinStr(BinaryString* pBS, int bits, int shift_constant)
{
  x = pBS->GetInt(bits);
  y = pBS->GetInt(bits);  
  if (x == -1 || y == -1)
    return false;
  x -= shift_constant;  
  y -= shift_constant;  
  return true;  
}



/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        Rectangle Class                                       */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

Rectangle::Rectangle()
{
  left_x = right_x = top_y = bottom_y = 0;
}

/********************************************************************************/

Rectangle::Rectangle(const float l, const float r, const float t, const float b) 
{
  if ( l >= r ) my_error("width must be positive %f %f %f %f", l, r, t, b);
  if ( t >= b ) my_error("height must be positive %f %f %f %f", l, r, t, b);

  left_x = l ; right_x = r ;
  top_y = t ; bottom_y = b ;
}

/********************************************************************************/

Rectangle::Rectangle(const Vector center,const Vector size)
{
  left_x = center.x - size.x/2.0 ;
  right_x = center.x + size.x/2.0 ;
  top_y = center.y - size.y/2.0 ;
  bottom_y = center.y + size.y/2.0 ;
}

/********************************************************************************/

Rectangle Rectangle::ShiftCenter(Vector new_center) const
{
  Vector curr_center = Center();
  return Rectangle(left_x + (new_center.x - curr_center.x),
		   right_x + (new_center.x - curr_center.x),
		   top_y + (new_center.y - curr_center.y),
		   bottom_y + (new_center.y - curr_center.y));
}

/********************************************************************************/

Bool Rectangle::IsWithin(const Vector &p, const float& buffer) const
{
  return (Bool) ((p.x >= left_x + buffer) && 
		 (p.x <= right_x - buffer) && 
		 (p.y >= top_y + buffer) && 
		 (p.y <= bottom_y - buffer) );
}

/********************************************************************************/

Vector Rectangle::AdjustToWithin(const Vector &p, const float &buffer) const
{
  // Point ends up at least buffer within the rectangle
  // buffer<0 ==> outside the rectangle

  Vector r = p;

  if ( buffer > Width()/2 || buffer > Height()/2 )
    my_error("Rectangle::AdjustToWithin -- buffer too large");
  
  if (r.y - top_y    < buffer) r.y = top_y+buffer;
  if (bottom_y - r.y < buffer) r.y = bottom_y-buffer;
  if (right_x - r.x  < buffer) r.x = right_x-buffer;
  if (r.x - left_x   < buffer) r.x = left_x+buffer;

  return r;
}

/********************************************************************************/

Vector Rectangle::AdjustToOutside(const Vector& p, const float& buffer = 0) const
{
  Vector r = p;
  Rectangle rectBuff = expand(buffer);

  if (!rectBuff.IsWithin(r))
    return r;

  Line l = rectBuff.nearestEdgeLine(r);

  return l.ProjectPoint(r);
}

/********************************************************************************/
//order: top, right, bot, left
Line Rectangle::GetEdge(int n) const
{
  switch (n % 4) {
  case 0: return TopEdge();
  case 1: return RightEdge();
  case 2: return BottomEdge();
  case 3: return LeftEdge();
  }
  my_error("Rectangle::GetEdge: how did I get here");
  return Line();
}

/********************************************************************************/
//order: TL, TR, BR, BL
Vector Rectangle::GetPoint(int n) const
{
  switch (n % 4) {
  case 0: return TopLeftCorner();
  case 1: return TopRightCorner();
  case 2: return BottomRightCorner();
  case 3: return BottomLeftCorner();
  }    
  my_error("Rectangle::GetPoint: how did I get here");
  return Vector(0,0);
}

/********************************************************************************/

Vector Rectangle::nearestHEdge(const Vector &p) const
     /* find nearest horizontal line */
{
  static Vector r ;

  r.x = Min(Max(p.x,left_x),right_x) ;
  r.y = ((p.y - top_y) < (bottom_y - p.y)) ? top_y : bottom_y ;

  return r ;
}

/********************************************************************************/

Vector Rectangle::nearestVEdge(const Vector &p) const 
     /* find nearest vertical line */
{
  static Vector r ;

  r.x = ((p.x - left_x) < (right_x - p.x)) ? left_x : right_x ;
  r.y = Min(Max(p.y,top_y),bottom_y) ;

  return r ;
}

/********************************************************************************/

Vector Rectangle::nearestEdge(const Vector &p) const 
{
  if(Min((p.x-left_x),(right_x-p.x)) < Min((p.y-top_y),(bottom_y-p.y)))
    return nearestVEdge(p) ;
  else
    return nearestHEdge(p) ;
}

/********************************************************************************/

Line Rectangle::nearestHEdgeLine(const Vector& p) const 
{
  return ((p.y - top_y) < (bottom_y - p.y)) ? TopEdge() : BottomEdge() ;
}


/********************************************************************************/

Line Rectangle::nearestVEdgeLine(const Vector& p) const 
{
  return ((p.x - left_x) < (right_x - p.x)) ? LeftEdge() : RightEdge() ;
}


/********************************************************************************/

Line Rectangle::nearestEdgeLine(const Vector& p) const 
{
  if(Min((p.x-left_x),(right_x-p.x)) < Min((p.y-top_y),(bottom_y-p.y)))
    return nearestVEdgeLine(p) ;
  else
    return nearestHEdgeLine(p) ;  
}


/********************************************************************************/

float Rectangle::DistanceToEdge(const Vector &p) const 
{
  if ( !IsWithin(p) ){
    Vector q = AdjustToWithin(p);
    return -(q.dist(p)); /* distance outside is a negative number */
  }
   
  return Min((p.x-left_x),Min((right_x-p.x),Min((p.y-top_y),(bottom_y-p.y))));
}

/********************************************************************************/

Vector Rectangle::nearestCorner(const Vector &p, float interior_dist) const 
{
  Vector corner = TopLeftCorner();
  float dist = corner.dist(p);
  float tmp;

  if ( (tmp=TopRightCorner().dist(p)) < dist ){
    corner = TopRightCorner();
    dist = tmp;
  }

  if ( (tmp=BottomRightCorner().dist(p)) < dist ){
    corner = BottomRightCorner();
    dist = tmp;
  }

  if ( (tmp=BottomLeftCorner().dist(p)) < dist ){
    corner = BottomLeftCorner();
    dist = tmp;
  }

  /* shift the point towards the center */
  if ( interior_dist > 0 )
    return PointInBetween(corner, Center(), interior_dist);
  else
    return corner;
}

/********************************************************************************/

void Rectangle::FlipCoords(bool flipX, bool flipY)
{
  if (flipX) {
    float tmp = left_x;
    left_x = -right_x;
    right_x = -tmp;
  }
  if (flipY) {
    float tmp = top_y;
    top_y = -bottom_y;
    bottom_y = -tmp;
  }
}

/********************************************************************************/

Vector Rectangle::random() const 
{
  static Vector r ;

  r.x = range_random(left_x, right_x) ;
  //r.y = range_random(bottom_y, top_y) ;
  r.y = range_random(top_y, bottom_y) ;

  return r ;
}

/********************************************************************************/

Rectangle Rectangle::expand(float val) const
{
  return Rectangle(left_x - val, right_x + val, top_y - val, bottom_y + val);
}

/********************************************************************************/

void  Rectangle::Print() const{
  printf("RECTANGLE:  x = %.1f to %.1f   y = %.1f to %.1f\n",
	 LeftX(),RightX(),TopY(),BottomY());
}

/********************************************************************************/

Vector Rectangle::RayIntersection(Ray r) const
{
  if (!IsWithin(r.origin)) {
    my_error("Rectangle/Ray intersection only handles rays inside the rectangle!");
    return 0;
  }

  Vector int_pt[2];
  int num_int = 0;
  for (int i=0; i < 4; i++) {
    Vector pt;
    if (GetEdge(i).RayIntersection(r, &pt))
      int_pt[num_int++] = pt;
  }

  if (num_int == 0) {
    my_error("Rectangle ray intersection: no edge intersection?");
    return 0;
  } else if (num_int == 1) {
    return int_pt[0]; // same slope as one pair of edges 
  } else if (num_int == 2) {
    Rectangle temp_rect = expand(FLOAT_EPS);
    for (int j = 0; j < 2; j++)
      if (temp_rect.IsWithin(int_pt[j]))
	return int_pt[j];
    my_error("Rectangle ray intersection: how did I get here");
    return 0;
  } else {
    my_error("Rectangle ray intersection: how did num_int get so big: %d", num_int);
    return 0;
  }
}


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                         Line Class                                           */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

Line::Line(float x_coef, float y_coef, float constant) 
{
  A = x_coef;
  B = y_coef;
  C = constant;
}

/********************************************************************************/

void Line::LineFromTwoPoints(Vector pt1, Vector pt2)
{
  float temp = (pt2.x - pt1.x);
  if (fabs(temp) < FLOAT_EPS) {
    if (fabs(pt2.y - pt1.y) < FLOAT_EPS)
      my_error("LineFromTwoPoints: points can not be the same!");
    A = 1;
    B = 0;
  } else {
    float m = (pt2.y - pt1.y) / temp;
    A = -m;
    B = 1;
  }  
  C = -(A * pt2.x + B * pt2.y);
}

/********************************************************************************/

void Line::LineFromRay(Ray r)   
{
  if (fabs(r.direction.y) < FLOAT_EPS && fabs(r.direction.x) < FLOAT_EPS)
    my_error("LineFromRay: dir can not be zero");
  LineFromTwoPoints(r.origin, r.origin + r.direction);
}

/********************************************************************************/

Bool Line::PointOnLine(float x, float y) const
{
  return (Bool)(fabs(A * x + B * y + C) < FLOAT_EPS);
}

/********************************************************************************/

float Line::dist(Vector pt) const     
{
  return fabs( (A*pt.x + B*pt.y + C) / sqrt(Sqr(A) + Sqr(B)) );
}

/********************************************************************************/

float Line::dist2(Vector pt) const     
{
  return fabs( Sqr(A*pt.x + B*pt.y + C) / (Sqr(A) + Sqr(B)) );
}

/********************************************************************************/

float Line::angle() const
{
  return ATan(-A / B);
}

/********************************************************************************/
Vector Line::perpendicularVectorTowards(Vector pt) const
{
  Line p = perpendicular(pt);
  Vector int_pt = intersection(p);
  Vector v = p.getVector();
  if (HalfPlaneTest(pt) != HalfPlaneTest(int_pt+v))
    v = -v;
  return v;
}

/********************************************************************************/

Vector Line::ProjectPointUsingCircle(Vector pt) const
{
  /* here's the idea:
     first get the plane equation for the ray
     then use the closest distance formula to get the closest distance
     then using the eq for that line and the eq for a circle of the dist radius
       around our curretn position, find the nearest point.
       Whew! */

  my_error("You shouldn't call ProjectPointUsingCircle");
  
  /* compute the dist */
  Vector retPt;
  float d = dist(pt); /* the min distance */

  /* intersect the circle and the line */
  float a,b,c; /* coefficent in quadratic to solve for x */
  float disc; /* discriminant in quadratic equation */
  a = 1 + Sqr(A);
  b = 2 * (A * (pt.y + C) - pt.x);
  c = Sqr(pt.x) + Sqr(pt.y + C) - Sqr(d);
  disc = Sqr(b) - 4 * a * c;
  /* the discriminant should be zero since this is the radius
     is the closest distance between the center and the line */
  if (fabs(disc) > FLOAT_EPS)
    fprintf(stderr, "GetClosestPointToBallPath: discrimannt is bad! %f\n", disc);
  retPt.x = - b / (2 * a);

  /* we compute two possible solutions for y and then see which one is on the
     line of the ray's path */
  float sol1, sol2;
  sol1 = sqrt(Sqr(d) - Sqr(retPt.x - pt.x));
  sol2 = -sol1;
  sol1 += pt.y;
  sol2 += pt.y;

  if (fabs(A * (retPt.x) + B * sol1 + C) < FLOAT_EPS ){
    /* sol1 is on line */
    retPt.y = sol1;
  } else if (fabs(A * (retPt.x) + B * sol2 + C) < FLOAT_EPS ){
    /* sol2 is on line */
    retPt.y = sol2;
  } else
    fprintf(stderr, "GetClosestPointToBallPath: neither solution works!\n");

  DebugGeom(printf("  dist: %f\t ptMod: %f\n", d,
		  sqrt(Sqr(pt.x - retPt.x) + Sqr(pt.y - retPt.y))));

  return retPt;
}

/********************************************************************************/

float Line::get_y(float x) const
{
  if ( B != 0 ) 
    return (-A*x - C)/B;
  
  my_error("can't get y"); 
  return 0;
}

/********************************************************************************/

float Line::get_x(float y) const
{
  if ( A != 0 ) 
    return (-B*y - C)/A;
  
  my_error("can't get x"); 
  return 0;
}

/********************************************************************************/

Bool Line::InBetween(Vector pt, Vector end1, Vector end2) const
{
  if (!OnLine(end1) || !OnLine(end2))
    my_error("Line::InBetween: passed in points that weren't on line");

  pt = ProjectPoint(pt);
  float dist2 = end1.dist2(end2);
  
  return (pt.dist2(end1) <= dist2 && pt.dist2(end2) <= dist2) ? TRUE : FALSE;
}

/********************************************************************************/

Vector Line::GetClosestPtInBetween(Vector pt, Vector end1, Vector end2) const
{
  if (!OnLine(end1) || !OnLine(end2))
    my_error("Line::InBetween: passed in points that weren't on line");

  if (InBetween(pt, end1, end2))
    return ProjectPoint(pt);
  if (end1.dist2(pt) < end2.dist2(pt))
    return end1;
  else
    return end2;
}

/********************************************************************************/

float Line::GetDistToClosestPtInBetween(Vector pt, Vector end1, Vector end2) const
{
  if (!OnLine(end1) || !OnLine(end2))
    my_error("Line::InBetween: passed in points that weren't on line");

  if (InBetween(pt, end1, end2))
    return dist(pt);
  float d1 = end1.dist(pt);
  float d2 = end2.dist(pt);
  if (d1 < d2)
    return d1;
  else
    return d2;
}


/********************************************************************************/

Vector Line::intersection(Line l) const
{
  Vector result = 0;
  if (SameSlope(l)) {
    //if ( B == 0 && l.B == 0 || A/B == l.A/l.B ){
    my_error("Lines have same slope");
    DebugGeom(cout << "Lines have same slope" << endl);
    return result;
  }

  if ( B == 0 ){
    result.x = -C/A;
    result.y = l.get_y(result.x);
    return result;
  }

  if ( l.B == 0){
    result.x = -l.C/l.A;
    result.y = get_y(result.y);
    return result;
  }

  result.x = (C*l.B - B*l.C)/(l.A*B - A*l.B);
  result.y = get_y(result.x);
  return result;
}  

/********************************************************************************/
Bool Line::RayIntersection(Ray r, Vector *ppt) const
{
  Line lRay(r);

  if (SameSlope(lRay))
    return FALSE;
  
  *ppt = intersection(lRay);
  return (r.InRightDir(*ppt))
      //fabs(GetNormalizeAngleDeg((*ppt - r.origin).dir() - r.direction.dir())) < 10)
    ? TRUE : FALSE;
}


/********************************************************************************/
Bool Line::IsPtCloserToPtOnLine(Vector pt1, Vector pt2, Vector targ_pt) const
{
  if (!OnLine(targ_pt)) 
    my_error("IsPtCloserToPtOnLine: targ_pt not on line");

  pt1 = ProjectPoint(pt1);
  pt2 = ProjectPoint(pt2);

  return (pt1.dist(targ_pt) < pt2.dist(targ_pt)) ? TRUE : FALSE;  
}


/********************************************************************************/

//return TRUE on top/left part of plane
Bool Line::HalfPlaneTest(Vector pt) const
{
  if (B==0)
    return (pt.x < -C/A) ? TRUE : FALSE;
  return (pt.y > get_y(pt.x)) ? TRUE : FALSE;
}

Bool Line::SameSlope(Line l) const
{
  return ( B == 0 && l.B == 0 || A/B == l.A/l.B ) ? TRUE : FALSE;    
}


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                       Ray Class                                              */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
Ray::Ray(Vector orig, Vector dir) 
{
  origin = orig;
  if (fabs(dir.y) < FLOAT_EPS && fabs(dir.x) < FLOAT_EPS) {
    my_error("Ray: dir can not be zero");
    direction = Vector (1,0);
    direction_angle = 0;
  } else {
    direction = dir;
    direction = direction.Normalize();
    direction_angle = direction.dir();
  }
}

/* This function is probably useless!
   The check is so strict that even points which have been projected to the Ray can
   fail */
Bool Ray::OnRay(Vector pt) const
{
  Vector v = pt - origin;
  AngleDeg angoff = GetNormalizeAngleDeg(v.dir() - direction_angle);
  float distoff = Sin(angoff) * v.mod();
  return (fabs(angoff) < 90) && (distoff < FLOAT_EPS) ? TRUE : FALSE;
}

Bool Ray::InRightDir(Vector pt, float angle_buffer) const
{
  return (fabs(GetNormalizeAngleDeg((pt - origin).dir() - direction_angle)) < 
	  angle_buffer)
    ? TRUE : FALSE;
}

Bool Ray::intersection(Line l, Vector *pPt) const
{
  return l.RayIntersection(*this, pPt);
}

Bool Ray::intersection(Ray r, Vector *pPt) const
{
  Line thisLine(*this), argLine(r);

  if (thisLine.SameSlope(argLine))
    return FALSE;
  *pPt = thisLine.intersection(argLine);

  /* now make sure that the intersection is the correct direction on both lines */
  return  (InRightDir(*pPt) && r.InRightDir(*pPt))
	   //	   fabs(GetNormalizeAngleDeg((*pPt - origin).dir() - direction_angle)) < 10 &&
	   //fabs(GetNormalizeAngleDeg((*pPt - r.origin).dir() - r.direction_angle)) < 10)
    ? TRUE : FALSE;
}


Vector Ray::RectangleIntersection(Rectangle R) const
{
  return R.RayIntersection(*this);
}

Vector Ray::GetClosestPoint(Vector pt) const
{
  Line l(*this);
  Vector close_pt = l.ProjectPoint(pt);
  if (InRightDir(close_pt))
    return close_pt;
  else
    return origin;
}

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                               Circle Class                                   */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

Vector Circle::GetClosestPoint(Vector pt) const
{
  if (isWithin(pt))
    return pt;
  return (pt-center).scaleTo(radius) + center;
}

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                               Ellipse Class                                   */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

vector<Circle> Ellipse::approxByCirclesOnMajor(float alpha) const 
{
  Circle c(center, 1.0);
  vector<Circle> v;
  bool xaxeslonger = (A < B);
  Vector centerincr = (xaxeslonger) ? Vector(1,0) : Vector(0,1);
  
  c.center = center;
  do {

    if (xaxeslonger)
      c.radius = YHeightAtXVal(c.center.x);
    else
      c.radius = XHeightAtYVal(c.center.y);
    
    v.push_back(c);
    v.push_back(Circle(2 * center - c.center, c.radius));
    c.center += centerincr.scaleTo(alpha * c.radius);
  } while (isWithin(c.center));

  return v;
}

//TODO: I don't know the "right" way to choose the points of the short side of the
//ellipse
vector<Vector> Ellipse::approxByTrapezoids() const
{
  vector<Vector> v;
  Vector rel;
  
  float a = Xrad();
  float b = Yrad();

  if (isXAxesLonger()) {
    //this is the positive trapezoid
    v.push_back(center + Vector(a,0));
    v.push_back(center + Vector(-a,0));
    rel.x = -a/2;
    rel.y = YHeightAtXVal(center.x + rel.x);
    v.push_back(center + rel);
    rel.x = a/2;
    v.push_back(center + rel);
    //now the negative
    v.push_back(center + Vector(a,0));
    v.push_back(center + Vector(-a,0));
    rel.x = -a/2;
    rel.y = -YHeightAtXVal(center.x + rel.x);
    v.push_back(center + rel);
    rel.x = a/2;
    v.push_back(center + rel);
  } else {
    //this is the positive trapezoid
    v.push_back(center + Vector(0,b));
    v.push_back(center + Vector(0,-b));
    rel.y = -b/2;
    rel.x = XHeightAtYVal(center.y + rel.y);
    v.push_back(center + rel);
    rel.y = b/2;
    v.push_back(center + rel);
    //now the negative
    v.push_back(center + Vector(0,b));
    v.push_back(center + Vector(0,-b));
    rel.y = -b/2;
    rel.x = -XHeightAtYVal(center.y + rel.y);
    v.push_back(center + rel);
    rel.y = b/2;
    v.push_back(center + rel);
  }
  return v;
}


istream& operator >>(istream & is, Ellipse &e)
{
  char ch;

  e.A = e.B = -1;
  
  try {
    while (isspace(ch = is.get())) ;
    if (ch != '(') throw 0;
    if ((ch = is.get()) != 'E') throw 1;
    is >> e.center;
    is >> e.A;
    is >> e.B;
    e.A = fabs(e.A);
    e.B = fabs(e.B);
    while (isspace(ch = is.get())) ;
    if (ch != ')') throw 2;
  } catch(int i) { LogAction3(200, "Could not read ellipse %d!", i); }
  
  return is;
}



/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                    RectGrid and RectGrid Multi Shared                        */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

#define RG_VAL_WIDTH 5
#define RGM_VAL_WIDTH 5

inline void SharedGetRectOfPoint(float pitch_length, float pitch_width,
				 int num_cols, int num_rows,
				 float x, float y, int* pr, int* pc)
{
  float rect_length = pitch_length / num_cols;  
  float rect_width = pitch_width / num_rows;
  /*DebugGeom(printf("addr: %p\n", this));
  DebugGeom(cout << "rows: " << num_rows
	    << "\tnum_cols: " << num_cols << endl);
  DebugGeom(cout << "rect_len: " << rect_length
	    << "\trect_wid: " << rect_width << endl);*/
  *pr = MinMax(0,
	       (int)floor( (y + pitch_width / 2) / rect_width ),
	       num_rows-1);
  *pc = MinMax(0,
	       (int)floor( (x + pitch_length / 2) / rect_length ),
	       num_cols-1);
}

static char temp_str[100];
char* int2str(int val, int wid)
{
  char fmt[10];
  sprintf(fmt, "%%%dd", wid);
  sprintf(temp_str, fmt, val);
  return temp_str;
}

char* float2str(float val, int wid, int prec = 0)
{
  char fmt[10];
  sprintf(fmt, "%%%d.%df", wid, prec);
  sprintf(temp_str, fmt, val);
  return temp_str;
}

inline float distance2(int r1, int c1, int r2, int c2)
{
  return Sqr(r2-r1) + Sqr(c2-c1);
}

inline float distance(int r1, int c1, int r2, int c2)
{
  return sqrt(distance2(r1, c1, r2, c2));  
}


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        RectGrid Class                                        */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

const char rgHeaderLine[] = "Writing grid with (r, c) = ";
const char rgHeaderLineEmpty[] = "<NONE>";

RectGrid::RectGrid()
{
  vals = NULL;
  pitch_length = pitch_width = 0.0;
  num_rows = num_cols = 0;
}

void RectGrid::Initialize(float pitch_length, float pitch_width,
			  int num_rows, int num_cols)
{
  this->pitch_length = pitch_length;
  this->pitch_width  = pitch_width;
  this->num_rows     = num_rows;
  this->num_cols     = num_cols;

  DebugGeom(printf("addr: %p\n", this));
  DebugGeom(cout << "RectGrid Init: rows: " << num_rows
	    << "\tnum_cols: " << num_cols << endl);
  vals = new (float*)[num_rows];
  for (int row = 0; row < num_rows; row++) {
    vals[row] = new float[num_cols];
  }
  Zero();
}


RectGrid::~RectGrid()
{
  for (int row = 0; row < num_rows; row++) {
    delete [] vals[row];
  }
  delete [] vals;
}


void RectGrid::Zero()
{
  DebugGeom(printf("Zeroing RectGrid at %p\n", this));
  for (int i = 0; i < num_rows; i++) {
    for (int j = 0; j < num_cols; j++) {
      vals[i][j] = 0;
    }    
  }  
}


void RectGrid::GetRectOfPoint(float x, float y, int* pr, int* pc)
{
  SharedGetRectOfPoint(pitch_length, pitch_width, num_cols, num_rows,
		       x, y, pr, pc);
}
void RectGrid::GetRectOfPoint(Vector pos, int* pr, int* pc)
{
  GetRectOfPoint(pos.x, pos.y, pr, pc);
}


float RectGrid::IncrementRect(int r, int c, float by, Bool invert)
{
  if (r < 0 || r >= num_rows ||
      c < 0 || c >= num_cols) {
    my_error("RectGrid, tried to incrment bad rect: (r,c) = (%d, %d)\n", r, c);
    return -1.0;
  }
  DebugGeom(printf("Incrmenting %d %d\t first level: %p\tsec level: %p\n", r, c, vals, vals[r]));

  if (invert) {
    r = (num_rows-1) - r;
    c = (num_cols-1) - c;
  }
  
  return (vals[r][c] += by); 
}

float RectGrid::IncrementRectForPoint(float x, float y, float by, Bool invert)
{
  int r,c;
  if (invert)
    GetRectOfPoint(-x, -y, &r, &c);
  else 
    GetRectOfPoint(x, y, &r, &c);
  return IncrementRect(r, c, by, FALSE); //we already inverted it
}
float RectGrid::IncrementRectForPoint(Vector pos, float by, Bool invert)
{
  return IncrementRectForPoint(pos.x, pos.y, by, invert);
}

//This is Bresenham's scan conversion algorithm
void RectGrid::GetInfoOfTrajectory(Vector start, Vector end, 
				   float* pAvg, float* pMax)
{
  MaxMeanSummary summ;
  
  int x0, y0, x1, y1;
  //These are in row/column, not x/y!
  GetRectOfPoint(start, &y0, &x0);
  GetRectOfPoint(end, &y1, &x1);
  x0 = MinMax(0, x0, num_cols-1);
  x1 = MinMax(0, x1, num_cols-1);
  y0 = MinMax(0, y0, num_rows-1);
  y1 = MinMax(0, y1, num_rows-1);

  /*
  cout << "  int path from (" << x0 << ", " << y0 << ") to ("
       << x1 << ", " << y1 << ")" << endl;
  */

  if ( fabs(x1-x0) < fabs(y1-y0) ) {
    //incr y
    if (y1 < y0) { //swap start,end
      int tmp;
      tmp = x0; x0 = x1; x1 = tmp;
      tmp = y0; y0 = y1; y1 = tmp;
    }
    int dx,dy,dxdy,d,incr;
    if (x1 >= x0) {
      dx = 2*(x1-x0);
      dy = 2*(y1-y0);
      incr = 1;
    } else {
      dx = -2*(x1-x0);
      dy = 2*(y1-y0);
      incr = -1;
    }
    dxdy = dx - dy;
    d = dx - dy/2;
    
    int s,t;

    //for (s=x0, t=y0; s<=x1; s++) {
    for (t=y0, s=x0; t<=y1; t++) {
      summ.AddPoint(vals[t][s]);
      
      if (d<0) {
	d += dx;
      } else {
	s += incr;
	d += dxdy;
      }
    }
  } else {
    if (x1 < x0) { //swap start,end
      int tmp;
      tmp = x0; x0 = x1; x1 = tmp;
      tmp = y0; y0 = y1; y1 = tmp;
    }
    int dx,dy,dydx,d,incr;
    if (y1 >= y0) {
      dx = 2*(x1-x0);
      dy = 2*(y1-y0);
      incr = 1;
    } else {
      dx = 2*(x1-x0);
      dy = -2*(y1-y0);
      incr = -1;
    }
    dydx = dy - dx;
    d = dy - dx/2;
    
    int s,t;

    for (s=x0, t=y0; s<=x1; s++) {
      summ.AddPoint(vals[t][s]);
      if (d<0) {
	d += dy;
      } else {
	t += incr;
	d += dydx;
      }
    }
  }

  *pAvg = summ.GetMean();
  *pMax = summ.GetMax();
}

float RectGrid::Sum()
{
  float sum = 0;
  for (int r = 0; r < num_rows; r++) {
    for (int c = 0; c < num_cols; c++) {
      sum += vals[r][c];
    }    
  }
  DebugGeom(cout << "Returning a sum of: " << sum << endl);
  return sum;
}

float RectGrid::AverageX()
{
  float rect_length = pitch_length / num_cols;  

  float sum = 0;
  float num_points = 0;  
  for (int r = 0; r < num_rows; r++) {
    for (int c = 0; c < num_cols; c++) {
      num_points += vals[r][c];
      sum += vals[r][c]*c;      
    }
  }
  
  DebugGeom(cout << "RectGrid::AverageX: num_points= " << num_points << endl);
  DebugGeom(cout << "RectGrid::AverageX: sum= " << sum << endl);

  sum = sum*rect_length + 
    num_points*(rect_length/2) - num_points*(pitch_length/2);  

  if (num_points == 0.0)
    return 0.0;
  else 
    return sum/num_points;
}


float RectGrid::AverageY()
{
  float rect_width = pitch_width / num_rows;  

  float sum = 0;
  float num_points = 0;  
  for (int r = 0; r < num_rows; r++) {
    for (int c = 0; c < num_cols; c++) {
      num_points += vals[r][c];
      sum += vals[r][c]*r;      
    }
  }
  
  DebugGeom(cout << "RectGrid::AverageY: num_points= " << num_points << endl);
  DebugGeom(cout << "RectGrid::AverageY: sum= " << sum << endl);

  sum = sum*rect_width + 
    num_points*(rect_width/2) - num_points*(pitch_width/2);  
  
  if (num_points == 0.0)
    return 0.0;
  else 
    return sum/num_points;
}



void RectGrid::WriteToFile(ostream& outfile, int prec)
{
  int non_zero = 0;
  int row, col;
  for (col=0; col < num_cols && !non_zero; col++) {
    for (row=0; row < num_rows && !non_zero; row++) {
      if (vals[row][col] != 0) {
	non_zero = 1;
      }      
    }
  }  

  if (non_zero) {
    outfile << rgHeaderLine
	    << "(" <<num_rows << ", " << num_cols << ")" << endl;
    for (row=0; row < num_rows; row++) {
      for (col=0; col < num_cols; col++) {
	outfile << ' ' << float2str(vals[row][col], RG_VAL_WIDTH, prec);
	//outfile << vals[row][col] << " ";
      }
      outfile << endl;
    }
  } else {
    outfile << rgHeaderLineEmpty << endl;
  }
}

Bool RectGrid::ReadFromFile(istream& infile)
{
  const int linesize = 100;
  char line[linesize];
  char ch;
  int rows, cols;

  /* get the first header line */
  infile.getline(line, linesize);
  if (!infile) return FALSE;
  while (strlen(line) == 0) { /* skip blank lines */
    infile >> ch; /* get the newline off */
    infile.getline(line, linesize);
    if (!infile) return FALSE;
  } 
  
  if (strncmp(line, rgHeaderLineEmpty, strlen(rgHeaderLineEmpty)) == 0)
    return TRUE; /* succesfully read in an empty RectGrid */

  else if (strncmp(line, rgHeaderLine, strlen(rgHeaderLine)) != 0)
    return FALSE; /* the header line is invalid */

  else {
    /* read in the rows/cols and verify they're equal */
    if (sscanf(line + strlen(rgHeaderLine), "(%d, %d)", &rows, &cols) != 2) {
      my_error("RectGrid::ReadFromFile: sscanf of row/col failed\n");
      return FALSE;
    }
    
    if (rows != num_rows || cols != num_cols) {
      my_error("RectGrid::ReadFromFile: diff size (%d, %d) vs. (%d, %d)",
	       rows, cols, num_rows, num_cols);
      return FALSE;
    }
    
    /* now we can read in all the values */
    for (int r=0; r<num_rows; r++)
      for (int c=0; c<num_cols; c++) {
	infile >> vals[r][c];
	if (!infile) return FALSE;
      }
  }
  
  return TRUE;
}

/* the prg values are multiplied by scale */
float RectGrid::GravSimFunc(RectGrid* prg, float scale, float M)
{
  /* compute the RectGrid of difference */
  RectGrid* prgDiff = new RectGrid;
  prgDiff->Initialize(pitch_length, pitch_width, 
		      num_rows, num_cols);

  int num_peaks=0, num_valleys=0;
  int r,c;
  struct VALLEYINFO {
    int r,c;
  } arrVall[500];
  
  for (r=0; r < num_rows; r++)
    for (c=0; c < num_cols; c++) {
      prgDiff->vals[r][c] = vals[r][c] - scale*prg->vals[r][c];
      if (prgDiff->vals[r][c] < 0) {
	arrVall[num_valleys].r = r;
	arrVall[num_valleys].c = c;
	num_valleys++;
      } else if (prgDiff->vals[r][c] > 0)
	num_peaks++;
    }

  DebugSim(printf(" RGSim: num_peaks: %d\tnum_valleys: %d\n", 
		  num_peaks, num_valleys));

  float peak_sum = 0.0;
  float difference_sum = 0.0;

  /* now go through every peak and squish it into all the valleys */
  for (r=0; r < num_rows; r++) {
    for (c=0; c < num_cols; c++) {
      if (prgDiff->vals[r][c] > 0) {
	/* this is a peak to squish */
	float valley_sum = 0.0; /* holds the sum of the size/d^2 */
	float peak_term = 0.0;	
	int vall;
	for (vall=0; vall < num_valleys; vall++) {
	  peak_term += -prgDiff->vals[arrVall[vall].r][arrVall[vall].c] /
	    distance(r,c,arrVall[vall].r,arrVall[vall].c);
	  valley_sum += -prgDiff->vals[arrVall[vall].r][arrVall[vall].c] /
	    distance2(r,c,arrVall[vall].r,arrVall[vall].c);
	} /* valley loop */
	difference_sum += peak_term * prgDiff->vals[r][c] / valley_sum;
	peak_sum += prgDiff->vals[r][c];	
      }
    }
  }

  DebugSim(printf(" RGSim: peak_sum: %f\tdiff_sum: %f\n", 
		  peak_sum, difference_sum));

  /* every peak is squished - normalize it */
  if (peak_sum > 0.0)
    difference_sum /= (peak_sum * M);
  
  delete prgDiff;

  return (1/(difference_sum+1.0));  /* add 1.0 so that max is 1.0 */
}


/* returns valus in range [0,1] */
/* M is an additional squahsing constant, the bigger it is, the more similar
   things will appear */
const int MAX_VALLEYS = 100;
/* SMURF: describe algorithm */
float RectGrid::SimilarityTo(RectGrid* prg, float M)
{
  float scale;

  /* SMURF: might want to check that sizes are the same */
  
  scale = ((float)Sum()) / prg->Sum();

  /* divide by 2.0 so that we have a value in [0,1] */
  return (GravSimFunc(prg, scale, M) + prg->GravSimFunc(this, 1/scale, M)) / 2.0;  
}

void RectGrid::CombineWith(RectGrid* prg, float weight)
{
  /* SMURF: might want to check that sizes are the same */

  for (int i = 0; i < num_rows; i++) {
    for (int j = 0; j < num_cols; j++) {
      vals[i][j] = (vals[i][j] + weight*prg->vals[i][j]) / (1.0 + weight);
    }    
  }  
  
}

void RectGrid::GenerateRandom(int num_points)
{
  Zero();
  for (int i=0; i<num_points; i++) {
    int r = int_random(num_rows);
    int c = int_random(num_cols);
    IncrementRect(r,c);
  }
  
}

/* does a random walk though the RectGrid, rather than every step being random */
void RectGrid::GenerateConnectedRandom(int num_points) 
{
  Zero();
  int r = int_random(num_rows);
  int c = int_random(num_cols);
  for (int i=0; i<num_points; i++) {
    if (range_random(0,1) < .5) {
      int r_chg = int_random(1 + (r==0?0:1) + (r==num_rows-1?0:1));
      r_chg -= (r==0?0:1);
      int c_chg = int_random(1 + (c==0?0:1) + (c==num_cols-1?0:1));
      c_chg -= (c==0?0:1);
      r += r_chg;
      c += c_chg;
    }
    IncrementRect(r,c);
  }  
}




/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        RectGridMulti Class                                        */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

const char rgmHeaderLine[] = "Writing multigrid with (r, c, v) = ";
const char rgmHeaderLineEmpty[] = "<NONE>";

RectGridMulti::RectGridMulti()
{
  vals = NULL;
  pitch_length = pitch_width = 0.0;
  num_rows = num_cols = num_vals = 0;
}

void RectGridMulti::Initialize(float pitch_length, float pitch_width,
			       int num_rows, int num_cols, int num_vals)
{
  this->pitch_length = pitch_length;
  this->pitch_width  = pitch_width;
  this->num_rows     = num_rows;
  this->num_cols     = num_cols;
  this->num_vals     = num_vals;

  DebugGeom(printf("addr: %p\n", this));
  DebugGeom(cout << "RectGridMulti Init: rows: " << num_rows
	    << "\tnum_cols: " << num_cols << endl);
  vals = new (float)[num_rows*num_cols*num_vals];
  Zero();
}



RectGridMulti::~RectGridMulti()
{
  delete [] vals;
}


void RectGridMulti::Zero()
{
  DebugGeom(printf("Zeroing RectGridMulti at %p\n", this));
  for (int i = 0; i < num_rows*num_cols*num_vals; i++) {
    vals[i] = 0;
  }  
}

void RectGridMulti::GetRectOfPoint(float x, float y, int* pr, int* pc)
{
  SharedGetRectOfPoint(pitch_length, pitch_width, num_cols, num_rows,
		       x, y, pr, pc);
}
void RectGridMulti::GetRectOfPoint(Vector pos, int* pr, int* pc)
{
  GetRectOfPoint(pos.x, pos.y, pr, pc);
}


float RectGridMulti::IncrementRect(int r, int c, int which, float by, Bool invert)
{
  if (r < 0 || r >= num_rows ||
      c < 0 || c >= num_cols) {
    my_error("RectGridMulti, tried to incrment bad rect: (r,c) = (%d, %d)\n", r, c);
    return 0;
  }
  
  if (invert) {
    r = (num_rows-1) - r;
    c = (num_cols-1) - c;
    which = (num_vals-1) - which;
  }
  
  return (*Ref(r,c, which) += by); 
}

float RectGridMulti::IncrementRectForPoint(float x, float y, int which,
					   float by, Bool invert)
{
  int r,c;
  if (invert)
    GetRectOfPoint(-x, -y, &r, &c);
  else
    GetRectOfPoint(x, y, &r, &c);
  return IncrementRect(r, c, (num_vals-1) - which, by, FALSE); //allready inverted it
}
float RectGridMulti::IncrementRectForPoint(Vector pos, int which,
					   float by, Bool invert)
{
  return IncrementRectForPoint(pos.x, pos.y, which, by, invert);
}

/* this is basically the same as scan conversion in computer graphics
   it's a little more complicated because the RectGridMulti has a more
   complicated relation from Vector coords to RectGrid coords */
void RectGridMulti::RecordTrajectory(Vector start, Vector end, Bool invert) 
{
  float dy = end.y - start.y;
  float dx = end.x - start.x;
  float m =  dy / dx;
  /* ang is actually the angle + 45 degrees */
  int ang = (((int)(end - start).dir()) + 45 + 360) % 360;
  /* the +45 is so that the case where num_vals is 4 makes more sense graphically */
  int which = ang / ((int)(360 / num_vals));
  DebugGeom(printf("  ang+45: %d\twhich: %d\n", ang, which));
  
  if (-1.0 <= m && m <= 1.0) {
    /* slope is such that we do one increment of x each iteration */
    double xinc = Sign((end - start).x) * pitch_length / num_cols;
    double x = start.x;
    double y = start.y;
    for ( ; Sign(xinc)*x <= Sign(xinc)*end.x; x += xinc, y += m*xinc ) {
      IncrementRectForPoint(x, y, which, invert);
    }
  } else {
    /* slope is such that we do one increment of y each iteration */
    double yinc = Sign((end - start).y) * pitch_width / num_rows;
    double x = start.x;
    double y = start.y;
    for ( ; Sign(yinc)*y <= Sign(yinc)*end.y; x += yinc / m, y += yinc ) {
      IncrementRectForPoint(x, y, which, invert);
    }
  }
}



float RectGridMulti::Sum()
{
  float sum = 0;
  for (int i = 0; i < num_rows*num_cols*num_vals; i++) {
    sum += vals[i];
  }
  DebugGeom(cout << "Returning a sum of: " << sum << endl);
  return sum;
}


void RectGridMulti::WriteToFile(ostream& outfile, int prec)
{
  int non_zero = 0;
  int row, col, which;
  for (col=0; col < num_cols && !non_zero; col++) {
    for (row=0; row < num_rows && !non_zero; row++) {
      for (which=0; which < num_vals && !non_zero; which++) {
	if (*Ref(row, col, which) != 0) {
	  non_zero = 1;
	}
      }
    }
  }  

  if (non_zero) {
    outfile << rgmHeaderLine << "(" 
	    << num_rows << ", " << num_cols << ", " << num_vals << ")" << endl;
    if (num_vals == 4) {
      /* there's prettier writing for num vals of 4 */
      for (row=0; row < num_rows; row++) {
	/* we'll amke three passes to get the upper number, then the
	   left and right ones, then the lower one */
	for (col=0; col < num_cols; col++) 
	  outfile << "   " << float2str(*Ref(row,col,1), RGM_VAL_WIDTH, prec) << "   ";
	outfile << endl; 
	for (col=0; col < num_cols; col++) {
 	  outfile << " " << float2str(*Ref(row,col,2), RGM_VAL_WIDTH, prec) << "##";
	  outfile << float2str(*Ref(row,col,0), 2, prec) << " ";
	}
	outfile << endl;
	for (col=0; col < num_cols; col++) 	
	  outfile << "   " << float2str(*Ref(row,col,3), RGM_VAL_WIDTH, prec) << "   ";
	outfile << endl;  
      }
    } else {
      for (row=0; row < num_rows; row++) {
	for (col=0; col < num_cols; col++) {
	  outfile << "[";
	  for (which=0; which < num_vals; which++) {
	    outfile << (which==0?"":",")
		    << float2str(*Ref(row,col,which), RGM_VAL_WIDTH, prec);
	  //outfile << vals[row][col] << " ";
	  }
	  outfile << "]";
	}
	outfile << endl;
      }
    }
  } else {
    outfile << rgmHeaderLineEmpty << endl;
  }
}

Bool RectGridMulti::ReadFromFile(istream& infile)
{
  const int linesize = 100;
  char line[linesize];
  int rows, cols, vals;

  /* get the first header line */
  if (!skip_white_space(infile)) return FALSE;  
  infile.getline(line, linesize);
  if (!infile) return FALSE;
  
  if (strncmp(line, rgmHeaderLineEmpty, strlen(rgmHeaderLineEmpty)) == 0)
    return TRUE; /* succesfully read in an empty RectGrid */

  else if (strncmp(line, rgmHeaderLine, strlen(rgmHeaderLine)) != 0)
    return FALSE; /* the header line is invalid */

  else {
    /* read in the rows/cols and verify they're equal */
    if (sscanf(line + strlen(rgmHeaderLine), "(%d, %d, %d)", 
	       &rows, &cols, &vals) != 3) {
      my_error("RectGridMulti::ReadFromFile: sscanf of row/col/val failed\n");
      return FALSE;
    }
    
    if (rows != num_rows || cols != num_cols || vals != num_vals) {
      my_error("RectGridMulti::ReadFromFile: diff size (%d, %d, %d) vs. (%d, %d, %d)",
	       rows, cols, vals, num_rows, num_cols, num_vals);
      return FALSE;
    }
    
    /* now we can read in all the values */
    if (num_vals == 4) {
      char c;
      /* there's prettier writing for num vals of 4 */
      /* we'll amke three passes to get the upper number, then the
	 left and right ones, then the lower one */
      for (int row=0; row < num_rows; row++) {
	int col;	
	for (col=0; col < num_cols; col++) 
	  infile >> *Ref(row,col,1);
	for (col=0; col < num_cols; col++) {
	  infile >> *Ref(row,col,2);
	  skip_white_space(infile);
	  infile >> c >> c; /* skip past the ## */
	  infile >> *Ref(row,col,0);
	}
	for (col=0; col < num_cols; col++) 	
	  infile >> *Ref(row,col,3);
      }
    } else {
      char c;      
      for (int row=0; row < num_rows; row++) {
	for (int col=0; col < num_cols; col++) {
	  if (!skip_white_space(infile)) return FALSE;
	  infile >> c;	  
	  for (int which=0; which < num_vals; which++) {
	    infile >> *Ref(row,col,which) >> c;
	    if (!infile) return FALSE;
	  }
	}
      }
    }
    
    return TRUE;
  }
     
}


/* returns valus in range [0,1] */
/* M is an additional squahsing constant, the bigger it is, the more similar
   things will appear */
/* this just squashes the RGM into a RG, then calls that sim */
float RectGridMulti::SimilarityTo(RectGridMulti* prgm, float M)
{
  float sim;
  RectGrid* prgThis = SquashToRectGrid();
  RectGrid* prgComp = prgm->SquashToRectGrid();

  sim = prgThis->SimilarityTo(prgComp, M);

  delete prgThis;
  delete prgComp;

  return sim;  
}


float* RectGridMulti::Ref(int r, int c, int which)
{
  if (r < 0 || r >= num_rows ||
      c < 0 || c >= num_cols ||
      which < 0 || which >= num_vals)
    my_error("RectGridMulti: Ref out of range: (%d, %d, %d)\n", r, c, which);
  return &(vals[r*(num_cols*num_vals) + c*(num_vals) + which]);
}

void RectGridMulti::CombineWith(RectGridMulti* prgm, float weight)
{
  /* SMURF: might want to check that sizes are the same */

  for (int i = 0; i < num_rows*num_cols*num_vals; i++) {
    vals[i] = (vals[i] + weight*prgm->vals[i]) / (1.0 + weight);
  }
}

RectGrid* RectGridMulti::SquashToRectGrid()
{
  RectGrid* prg;

  prg = new RectGrid;
  prg->Initialize(pitch_length, pitch_width, num_rows, num_cols);

  for (int r = 0; r < num_rows; r++)
    for (int c = 0; c < num_cols; c++)
      for (int w = 0; w < num_vals; w++)
	prg->vals[r][c] += *Ref(r,c,w);
	
  return prg;
}

void RectGridMulti::GenerateRandom(int num_points)
{
  Zero();
  for (int i=0; i<num_points; i++) {
    int r = int_random(num_rows);
    int c = int_random(num_cols);
    int w = int_random(num_vals);
    IncrementRect(r,c,w);
  }
}

void RectGridMulti::GenerateConnectedRandom(int num_points)
{
  Zero();
  for (int i=0; i<num_points; i++) {
    float x1 = range_random(-pitch_length/2, pitch_length/2);
    float y1 = range_random(-pitch_width/2, pitch_width/2);
    float x2 = range_random(-pitch_length/2, pitch_length/2);
    float y2 = range_random(-pitch_width/2, pitch_width/2);
    RecordTrajectory(Vector(x1,y1), Vector(x2,y2));
  }
}




/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                  Miscellaneous Geometry Functions                            */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/


/********************************************************************************/

int QuadraticFormula(float a, float b, float c, float* psol1, float* psol2)
{
  float d = Sqr(b) - 4*a*c;
  if (fabs(d) < FLOAT_EPS) {
    *psol1 = -b / (2 * a);
    return 1;
  } else if (d < 0) {
    return 0;
  } else {
    d = sqrt(d);
    *psol1 = (-b + d ) / (2 * a);
    *psol2 = (-b - d ) / (2 * a);
    return 2;
  }
}

/********************************************************************************/

/* some test code
   Vector sol1, sol2;
  int num;
  num = LineCircleIntersect(LineFromTwoPoints(Vector(-10, 2), Vector(10, 2)),
					      1, Vector(1,2), &sol1, &sol2);
  cout << "num: " << num << "\tsol1: " << sol1 << "\tsol2: " << sol2 << endl;
  num = LineCircleIntersect(LineFromTwoPoints(Vector(2, 10), Vector(2, -10)),
					      1, Vector(2,1), &sol1, &sol2);
  cout << "num: " << num << "\tsol1: " << sol1 << "\tsol2: " << sol2 << endl;
  num = LineCircleIntersect(LineFromTwoPoints(Vector(-10, -10), Vector(10, 10)),
					      sqrt(2), Vector(1,1), &sol1, &sol2);
  cout << "num: " << num << "\tsol1: " << sol1 << "\tsol2: " << sol2 << endl;
  exit(1);
  */
int LineCircleIntersect(Line l, Circle circ,
			Vector* psol1, Vector* psol2)
{
  *psol1 = *psol2 = Vector(0,0);
  if (fabs(l.getA()) > FLOAT_EPS) {
    float a,b,c;
    a = 1 + Sqr(l.getB()/l.getA());
    b = 2*(-circ.center.y + (l.getC()/l.getA() + circ.center.x)*(l.getB()/l.getA()));
    c = -Sqr(circ.radius) + Sqr(l.getC()/l.getA() + circ.center.x) + Sqr(circ.center.y);
    int numSol = QuadraticFormula(a, b, c, &(psol1->y), &(psol2->y));
    psol1->x = - ( (l.getB() * psol1->y + l.getC()) / l.getA() );
    psol2->x = - ( (l.getB() * psol2->y + l.getC()) / l.getA() );
    return numSol;
  } else {
    int numSol = QuadraticFormula(1, -2*circ.center.x,
				  Sqr(circ.center.x) + Sqr(l.getC()/l.getB() + circ.center.y) - Sqr(circ.radius),
				  &(psol1->x), &(psol2->x));
    psol1->y = psol2->y = - l.getC() / l.getB();
    return numSol;
  }
}


/********************************************************************************/

/* loop over rectangle edges. See if we're outside of the edge.
   If so, try to intersect with edge to move it inside */
Vector AdjustPtToRectOnLine(Vector pt, Rectangle r, Line l)
{
  DebugGeom(cout << "Adjust: " << pt << "\t" << r << "\t" << l << endl);
  Vector c = r.Center();
  for (int i = 0; i < 4; i++) {
    Line edge = r.GetEdge(i);
    if ( edge.SameSlope(l) )
      continue;
    if (edge.HalfPlaneTest(pt) != edge.HalfPlaneTest(c)) {
      // pt is outside this edge
      DebugGeom(printf("AdjustPtToRectOnLine: HalfPlaneTest failed for %d\n", i));
      Vector newPt = edge.intersection(l);
      if (edge.InBetween(newPt, r.GetPoint(i), r.GetPoint(i+1))) {
	DebugGeom(printf("AdjustPtToRectOnLine: Intersection okay for %d\n", i));	
	return newPt;
      }      
    }    
  }
  return pt;
}

/********************************************************************************/

Bool InBetween(Vector pt, Vector end1, Vector end2)
{
  pt = LineFromTwoPoints(end1, end2).ProjectPoint(pt);
  float dist2 = end1.dist2(end2);
  
  return (pt.dist2(end1) <= dist2 && pt.dist2(end2) <= dist2) ? TRUE : FALSE;
}

/********************************************************************************/

Vector PointInBetween(Vector pt1, Vector pt2, float pt1dist)
{
  if ( pt1dist < 0 ) my_error ("no neg dists");
  Vector targ = pt2 - pt1;
  targ *= pt1dist / targ.mod();
  return(pt1 + targ);
}

/********************************************************************************/

Vector PointInFracBetween(Vector pt1, Vector pt2, float pt1frac=.5)
{
  Vector targ = pt2 - pt1;
  targ *= pt1frac;
  return(pt1 + targ);
}

/********************************************************************************/

AngleDeg GetAngle(Vector pt1, Vector vertex, Vector pt2)
{
  AngleDeg dir1 = (pt1-vertex).dir();
  AngleDeg dir2 = (pt2-vertex).dir();
  
  return fabs(GetNormalizeAngleDeg(dir1 - dir2));
}


/********************************************************************************/

AngleDeg AngleBisect(AngleDeg a1, AngleDeg a2)
{
  if ( fabs(a1-a2) > 180 )
    return GetNormalizeAngleDeg((Min(a1,a2)+360+Max(a1,a2))/2);
  
  return GetNormalizeAngleDeg((a1+a2)/2);
}

/********************************************************************************/

Vector GetClosestPtInBetween(Vector pt, Vector end1, Vector end2)
{
  Line l;
  l.LineFromTwoPoints(end1, end2);
  return l.GetClosestPtInBetween(pt, end1, end2);
}

/********************************************************************************/

float GetDistToClosestPtInBetween(Vector pt, Vector end1, Vector end2)
{
  Line l;
  l.LineFromTwoPoints(end1, end2);
  return l.GetDistToClosestPtInBetween(pt, end1, end2);
}

/********************************************************************************/

Bool IsPointInCone(Vector pt, float wid_dist_ratio, Vector end, Vector vert)
{
  Line l = LineFromTwoPoints(vert, end);
  Vector proj_pt = l.ProjectPoint(pt);
  return  ((proj_pt.dist2(pt) < proj_pt.dist2(vert)*wid_dist_ratio*wid_dist_ratio &&
	   l.InBetween(proj_pt, vert, end)))
    ? TRUE : FALSE;
}

/********************************************************************************/

//top_y is negative, bot_y is positive
float GetDistToVerticalSegment(float x_val, float top_y, float bot_y, Vector pt)
{
  if (top_y > bot_y) {
    float tmp = top_y;
    top_y = bot_y;
    bot_y = tmp;
  }
  if (pt.y < top_y) {
    return pt.dist(Vector(x_val, top_y));
  } else if (pt.y > bot_y) {
    return pt.dist(Vector(x_val, bot_y));
  } else {
    return fabs(pt.x - x_val);
  }
}

/********************************************************************************/
bool SolveForVector(Ray r, Circle c, Vector* pReturn)
{
  int num_sol;
  Vector s1, s2;
  
  num_sol = r.CircleIntersect(c, &s1, &s2);
  
  if (num_sol == 0)
    return FALSE;
  *pReturn = s1 - r.GetOrigin(); //if there are two solutions, s1 is closer
  return TRUE;
}

/********************************************************************************/
/* intersects a ray and a cricle */
/* return the number of solutions */
/* psol1 1 is not as afar along the ray as psol2 */
int RayCircleIntersect(Ray r, Circle circ, Vector* psol1, Vector* psol2)
{
  DebugGeom(cout << "RCI: origin: " << r.GetOrigin() << "\tdirection: " << r.GetDirection() << endl
	    << "rad: " << circ.radius << "\tcenter: " << circ.center << endl);
  float a,b,c,disc;
  float t1, t2;
  a = Sqr(r.GetDirection().x) + Sqr(r.GetDirection().y);
  b = 2.0 * ((r.GetOrigin().x-circ.center.x) * r.GetDirection().x + (r.GetOrigin().y-circ.center.y) * r.GetDirection().y);
  c = Sqr(r.GetOrigin().x-circ.center.x) + Sqr(r.GetOrigin().y-circ.center.y) - Sqr(circ.radius);
  DebugGeom(printf(" RCI: a: %f\tb: %f\t c: %f\n", a,b,c));
  
  disc = Sqr(b) - 4 * a * c;
  if (disc < 0) {
    DebugGeom(printf(" RCI disc < 0: %f\n", disc));
    return 0;
  }
  
  disc = sqrt(disc);
  t1 = (-b + disc) / (2.0 * a);
  t2 = (-b - disc) / (2.0 * a);
  DebugGeom(printf(" RCI: t1: %f\tt2: %f\n", t1, t2));
  
  if (t1 > t2) {
    DebugGeom(printf(" RCI: reversing t1, t2\n"));
    float temp = t1;
    t1 = t2;
    t2 = temp;
  }
 
  if (t1 > 0.0) {
    if (t2 > 0.0) {
      *psol1 = r.GetOrigin() + r.GetDirection() * t1;
      *psol2 = r.GetOrigin() + r.GetDirection() * t2;
      DebugGeom(printf(" RCI:two sols\n"));
      return 2;
    } else {
      my_error("RayCircleIntersect: weird roots");
      return 0;
    }
  } else if (t2 > 0.0) {
    *psol1 = r.GetOrigin() + r.GetDirection() * t2;
    DebugGeom(printf(" RCI:t2 only sol\n"));
    return 1;
  } else
    return 0;

  return 0;
}

/********************************************************************************/
/* intersects a line and a rectangle */
/* return the number of solutions */
int LineRectangleIntersect(Line l, Rectangle r, Vector* psol1, Vector* psol2)
{
  LogAction2(150, "Calculating intersection points of a line with a rectangle");
  int numSol = 0;

  Vector corners[4];

  for(int i = 0; i < 4; i++) 
    corners[i] = r.GetPoint(i);

  for(int i = 0; i < 4; i++) {
    Vector pt1 = corners[i];
    Vector pt2;

    if(i == 3)
      pt2 = corners[0];
    else
      pt2 = corners[i+1];

    
    LogAction6(150, "Edge (%f, %f) - (%f, %f)", pt1.x, pt1.y, pt2.x, pt2.y);

    Line edge = LineFromTwoPoints(pt1, pt2);
    
    Vector intersectPt = l.intersection(edge);

    if(edge.InBetween(intersectPt, pt1, pt2)) {
      numSol++;
      if(numSol == 1)
	*psol1 = intersectPt;
      else {
	*psol1 = intersectPt;
	if(*psol1 == *psol2) // points are the same solution
	  numSol--;
	break;
      }
    }
  }

  return numSol;
}

/********************************************************************************/
/* intersects a line and a rectangle */
/* return the number of solutions */

int LineSegmentCircleIntersect(Vector pt1, Vector pt2, Circle c,
                               Vector* psol1, Vector* psol2)
{
  Line l = LineFromTwoPoints(pt1, pt2);

  int num_sol = l.CircleIntersect(c, psol1, psol2);

  if (num_sol == 0)
    return 0;
  if (num_sol == 1) {
    return (l.InBetween(*psol1, pt1, pt2)) ? 1 : 0;
  }
  if (num_sol == 2) {
    num_sol = 2;
    if (!l.InBetween(*psol1, pt1, pt2)) {
      *psol1 = *psol2;
      num_sol--;
    }
    return (l.InBetween(num_sol == 2 ? *psol2 : *psol1, pt1, pt2))
      ? num_sol : num_sol-1;
  }
  
  my_error("How did I get here?");
  return -1;
}

