/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 1998,1999,2000 Patrick Riley, Peter Stone, Manuela Veloso

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef _GEOMETRY_
#define _GEOMETRY_

#include <vector>
#include <strstream.h>
#include <fstream.h>
#include <math.h>
#include "BinaryString.h"
#include "utils_shared.h"

class Vector;
class Ray;
class Line;
class Rectangle;
class RectGrid;
class RectGridMulti;
class Circle;

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                  Miscellaneous Geometry Functions                            */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/* more of this at the bottom of the file */

int QuadraticFormula(float a, float b, float c, float* psol1, float* psol2);

int RayCircleIntersect(Ray r, Circle c,
		       Vector* psol1, Vector* psol2);

int LineCircleIntersect(Line l, Circle c,
			Vector* psol1, Vector* psol2);

int LineRectangleIntersect(Line l, Rectangle r,
			Vector* psol1, Vector* psol2);

Vector AdjustPtToRectOnLine(Vector pt, Rectangle r, Line l);

int LineSegmentCircleIntersect(Vector pt1, Vector pt2, Circle c, Vector* psol1, Vector* psol2);

bool InBetween(Vector pt, Vector end1, Vector end2);

Vector PointInBetween(Vector pt1, Vector pt2, float pt1dist=0);

Vector PointInFracBetween(Vector pt1, Vector pt2, float pt1frac=.5);

AngleDeg GetAngle(Vector pt1, Vector vertex, Vector pt2);

AngleDeg AngleBisect(AngleDeg a1, AngleDeg a2);

Vector GetClosestPtInBetween(Vector pt, Vector end1, Vector end2);
float  GetDistToClosestPtInBetween(Vector pt, Vector end1, Vector end2);

bool IsPointInCone(Vector pt, float wid_dist_ratio, Vector end, Vector vert);

float GetDistToVerticalSegment(float x_val, float top_y, float bot_y, Vector pt);

//Gives the vector which lines up with r and intersects the closest point of the circle
//Returns whether a solution exists
bool SolveForVector(Ray r, Circle c, Vector* pReturn);


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        Vector Class                                          */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/


class Vector {
public:
  float x;
  float y;

  Vector(const Vector& v) { x = v.x; y = v.y; }
  Vector(float vx = 0, float vy = 0) { x = vx; y = vy; }
  Vector operator -() const { return Vector(-x, -y); }
  Vector operator +(const Vector &a) const { return Vector(x + a.x, y + a.y); } 
  Vector operator -(const Vector &a) const { return Vector(x - a.x, y - a.y); } 
  Vector operator *(const float &a) const { return Vector(x * a, y * a); }
  Vector operator *(const Vector &a) const { return Vector(x * a.x, y * a.y); }
  Vector operator /(const float &a) const { return Vector(x / a, y / a); }
  Vector operator /(const Vector &a) const { return Vector(x / a.x, y / a.y); }
  Vector& operator =(const float &a) { x = a; y = a; return *this; }
  Vector& operator +=(const Vector &a) { x += a.x;  y += a.y; return *this; }
  Vector& operator +=(const float &a) { x += a;  y += a; return *this; }
  Vector& operator -=(const Vector &a) { x -= a.x;  y -= a.y; return *this; }
  Vector& operator -=(const float &a) { x -= a;  y -= a; return *this; }
  Vector& operator *=(const float &a) { x *= a;  y *= a; return *this; }
  Vector& operator /=(const float &a) { x /= a;  y /= a; return *this; }
  bool operator !=(const Vector &a) const { return (x != a.x) || (y != a.y); }
  bool operator !=(const float &a) const { return (x != a) || (y != a); } 
  bool operator ==(const Vector &a) const { return (x == a.x) && (y == a.y); } 

  friend ostream& operator <<(ostream & os, Vector v) { return os << "("<<v.x<<", "<<v.y<<")"; }
  friend istream& operator >>(istream & is, Vector &v);  
  void Print() const { cout << *this << endl; } 
  void PrintToStr(char* str) const;  

  void PrintToBinStr(BinaryString* pBS, const int bits, const int shift_constant) const;
  bool ReadFromBinStr(BinaryString* pBS, const int bits, const int shift_constant);  

  bool IsInf() const { return isinf(x) || isinf(y); }
  bool IsNan() const { return isnan(x) || isnan(y); }
  
  Vector FlipCoords(bool flipX=TRUE, bool flipY=TRUE) const
  { return Vector(flipX ? -x : x, flipY ? -y : y); }
  Vector FlipX() const { return FlipCoords(TRUE, FALSE); }
  Vector FlipY() const { return FlipCoords(FALSE, TRUE); }

  void ins(float vx = 0, float vy = 0) { x = vx; y = vy; }
  float mod() const { return sqrt(x*x + y*y); } 
  float mod2() const { return x*x + y*y; } 
  float dist(const Vector &a) const { return (*this-a).mod(); } 
  float dist2(const Vector &a) const { return (*this-a).mod2(); } 
  float dotprod(const Vector &a) const { return (this->x * a.x) + (this->y * a.y); }

  Vector Normalize() const { return scaleTo(1.0); }
  Vector scaleTo(float len) const { return (*this) * (len / mod()); }

  AngleDeg dir() const { return ATan2(y,x); } 
  Vector rotate(AngleDeg angle) const
  { return Vector(this->mod()*Cos(this->dir()+angle), this->mod()*Sin(this->dir()+angle)); }
  Vector rotateNeg90() const { return Vector(-y, x); }
  Vector rotate90() const    { return Vector(y, -x); }
    
  Vector Global2Relative(Vector orig, float ang) const;
  Vector Relative2Global(Vector orig, float ang) const;
  bool InFrontOf(const Vector &a) const { return x > a.x ? TRUE : FALSE; }
  bool InFrontOf(const float &a) const { return x > a ? TRUE : FALSE; } 
  bool Behind(const Vector &a) const { return x < a.x ? TRUE : FALSE; }
  bool Behind(const float &a) const { return x < a ? TRUE : FALSE; } 

  bool ApproxEqual(const Vector &a) const
    { return (fabs(x-a.x) < FLOAT_EPS && fabs(y-a.y) < FLOAT_EPS) ? TRUE : FALSE; }

  Vector WeightedAverage(const Vector &a, float w) const
    { return Vector((1.0-w)*x + w * a.x, (1.0-w)*y + w * a.y); }
};

inline Vector Polar2Vector(float mod, AngleDeg ang) {
  return Vector(mod * Cos(ang), mod * Sin(ang));
}
  
inline Vector Vector::Global2Relative(Vector orig, float ang) const
{
  return (*this - orig).rotate(-ang);
}

inline Vector Vector::Relative2Global(Vector orig, float ang) const
{
  return (rotate(ang) + orig);
}

inline Vector operator *(const float &a, const Vector &v) 
{ return Vector(v.x * a, v.y * a); }

inline float dot_product(Vector v1, Vector v2){
  return v1.x * v2.x + v1.y *v2.y;
}

inline float cross_product(Vector pt, Vector dir){
  return pt.x * dir.y - pt.y * dir.x;
}

inline Vector midpoint(Vector a, Vector b) 
{ return Vector((a.x + b.x) / 2.0, (a.y + b.y) / 2.0); }
  

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                       Ray Class                                              */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/* stores origin and direction (which is normalized) */
class Ray 
{
public:
  Ray() 
    { origin = Vector(0,0); direction = Vector(1,0); direction_angle = 0; }
  Ray(Vector orig, Vector dir);
  Ray(Vector orig, float ang)
    { origin = orig; direction = Polar2Vector(1.0, ang); direction_angle = ang; }

  void SetFromToPts(Vector orig, Vector dest) { *this = Ray(orig, dest-orig); }

  /* This function is probably useless!
     The check is so strict that even points which have been projected to the Ray can
     fail */
  bool OnRay(Vector pt) const;

  bool InRightDir(Vector pt, float angle_buffer = 10) const; // more lenient than above about distance off ray 
  
  bool intersection(Ray r, Vector *pPt) const;
  bool intersection(Line l, Vector *pPt) const;

  int CircleIntersect(Circle c, Vector* psol1, Vector* psol2) const;

  Vector GetClosestPoint(Vector pt) const;
  float  GetDist(Vector pt) const { return pt.dist(GetClosestPoint(pt)); }

  Vector RectangleIntersection(Rectangle R) const;

  Vector GetOrigin() const { return origin; }
  Vector GetDirection() const { return direction; }
  AngleDeg GetAngle() const { return direction_angle; }
  
protected:
  friend Line;
  friend Rectangle;
  Vector origin;
  Vector direction;
  float  direction_angle;
};

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                      Line Class                                              */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
class Line {
 public:
  Line(float x_coef=0.0, float y_coef=0.0, float constant=0.0);
  Line(Ray r) 
    { LineFromRay(r); }

  void LineFromTwoPoints(Vector pt1, Vector pt2);
  void LineFromRay(Ray r);
  void LineFromRay(Vector orig, Vector dir)
    { LineFromRay(Ray(orig, dir)); }
  void LineFromRay(Vector orig, AngleDeg dir)
    { LineFromRay(orig, Polar2Vector(1.0, dir)); }

  void MakeVert(float x) { A = 1.0; B = 0.0; C = -x; }
  void MakeHoriz(float y) { A = 0.0; B = 1.0; C = -y; }
  
  bool PointOnLine(float x, float y) const;
  bool PointOnLine(Vector pt) const { return PointOnLine(pt.x, pt.y); }

  float dist(Vector pt) const;
  float dist2(Vector pt) const;

  float angle() const; /* returns the angle of the line in [-90, 90] */
  /* gets a normalized vector in the direction of the line.
     It's not defined which way the vector  goes
     It's surprising to me that it's -B	*/
  Vector getVector() const { return Vector(-B, A).Normalize(); }

  Vector perpendicularVectorTowards(Vector pt) const;
  
  bool InBetween(Vector pt, Vector end1, Vector end2) const;
  Vector GetClosestPtInBetween(Vector pt, Vector end1, Vector end2) const;
  float  GetDistToClosestPtInBetween(Vector pt, Vector end1, Vector end2) const;

  /* the buffer should really be linked to an option or something,
     but 1.0 is okay */
  bool OnLine(Vector pt, float buffer = 1.0) const
    { return (dist2(pt) < Sqr(buffer)) ? TRUE : FALSE; }
  
  Vector ProjectPointUsingCircle(Vector pt) const;
  Vector ProjectPoint(Vector pt) const { return intersection(perpendicular(pt)); }

  float get_y(float x) const;
  float get_x(float y) const;
  Vector intersection(Line l) const;
  Line perpendicular(Vector pt) const { return Line(B,-A,A*pt.y - B*pt.x); }

  Line parallelThrough(Vector pt) const
  { return Line(A, B, C - eval(pt)); }
  
  bool RayIntersection(Ray r, Vector *ppt) const;
  int  CircleIntersect(Circle c, Vector *psol1, Vector* psol2) const;
  int  RectangleIntersect(Rectangle r, Vector *psol1, Vector* psol2) const;
  
  Line shift_y(float val) const
    { return Line(A, B, C - val * B); }
  Line shift_x(float val) const
    { return Line(A, B, C - val * A); }  
  Line shift_by_vector(Vector v) const
  { return shift_x(v.x).shift_y(v.y); }
  
  /* returns whether the projection of pt1 is closer to targ_pt than the
     projection of pt2 */
  bool IsPtCloserToPtOnLine(Vector pt1, Vector pt2, Vector targ_pt) const;
  
  bool HalfPlaneTest(Vector pt) const; //return TRUE on top/left part of plane
  bool SameSlope(Line l) const;

  float eval(Vector pt) const { return A*pt.x + B*pt.y + C; }
  Line  negateCoeff() const { return Line(-A, -B, -C); }
  
  friend ostream& operator <<(ostream & os, Line l)
    { return os << "#L("<<l.A<<", "<<l.B<<", "<<l.C<<")"; } 
  void Print(void) const { cout << *this << endl; }

  float getA() const { return A; }
  float getB() const { return B; }
  float getC() const { return C; }

  float getXCoeff() const { return A; }
  float getYCoeff() const { return B; }
  float getConst() const { return C; }
  
private:
  float A,B,C; /* the three coeffs in the line equation */
  /* Ax + By + C = 0 */
} ;

inline Line LineFromTwoPoints(Vector pt1, Vector pt2) { Line l; l.LineFromTwoPoints(pt1,pt2); return l; }
inline Line LineFromRay      (Ray r)
{ Line l; l.LineFromRay     (r); return l; }
inline Line LineFromRay      (Vector orig, Vector dir)
{ Line l; l.LineFromRay     (orig,dir); return l; }
inline Line LineFromRay      (Vector orig, AngleDeg dir)
{ Line l; l.LineFromRay     (orig,dir); return l; }

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        Rectangle Class                                       */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

class Rectangle {
public:

  Rectangle() ;
  Rectangle(const float left, const float right, 
	    const float top, const float bottom) ;
  Rectangle(const Vector center, const Vector size) ;

  bool operator ==(const Rectangle &a) 
  { return (left_x == a.left_x) && (right_x == a.right_x) && 
      (top_y == a.top_y) && (bottom_y == a.bottom_y); } 

  float TopY()     const { return top_y; }
  float BottomY()  const { return bottom_y; }
  float RightX()   const { return right_x; }
  float LeftX()    const { return left_x; }

  Vector TopLeftCorner() const { return Vector(left_x, top_y); }
  Vector TopRightCorner() const { return Vector(right_x, top_y); }
  Vector BottomLeftCorner() const { return Vector(left_x, bottom_y); }
  Vector BottomRightCorner() const { return Vector(right_x, bottom_y); }        

  Line LeftEdge() const {return LineFromTwoPoints(TopLeftCorner(), BottomLeftCorner());}
  Line RightEdge() const {return LineFromTwoPoints(TopRightCorner(), BottomRightCorner());}
  Line TopEdge() const {return LineFromTwoPoints(TopLeftCorner(), TopRightCorner());}
  Line BottomEdge() const {return LineFromTwoPoints(BottomLeftCorner(), BottomRightCorner());}

  Line GetEdge(int n) const;  //order: top, right, bot, left; starting at 0
  Vector GetPoint(int n) const; //order: TL, TR, BR, BL; starting at 0
  
  float Width() const    { return right_x - left_x; }   
  float Height() const   { return bottom_y - top_y; }  

  Vector Center() const  { return Vector( LeftX() + Width()/2, TopY() + Height()/2 ); }

  Rectangle ShiftCenter(Vector center) const;
  
  float Area() const { return Width() * Height(); }
  
  bool   IsWithin(const Vector& p, const float& buffer = 0) const ;
  Vector nearestHEdge(const Vector& p) const ;
  Vector nearestVEdge(const Vector& p) const ;
  Vector nearestEdge(const Vector& p) const ;

  float  DistanceToEdge(const Vector& p) const;
  Vector AdjustToWithin(const Vector& p, const float& buffer = 0) const;
  Vector AdjustToOutside(const Vector& p, const float& buffer = 0) const;

  Line nearestHEdgeLine(const Vector& p) const ;
  Line nearestVEdgeLine(const Vector& p) const ;
  Line nearestEdgeLine(const Vector& p) const ;
  Vector nearestCorner(const Vector& p, float interior_dist = 0) const ;
  
  Vector random() const ;

  void FlipCoords(bool flipX, bool flipY);
  void FlipX() { FlipCoords(TRUE, FALSE); }
  void FlipY() { FlipCoords(FALSE, TRUE); }

  Rectangle shrink(float val) const
    { return expand(-val); }
  Rectangle expand(float val) const;

  Rectangle expandLeft(float val) const
    {return Rectangle(left_x - val, right_x, top_y, bottom_y);}
  Rectangle expandRight(float val) const
    {return Rectangle(left_x, right_x + val, top_y, bottom_y);}
  Rectangle expandTop(float val) const
    {return Rectangle(left_x, right_x, top_y - val, bottom_y);}
  Rectangle expandBottom(float val) const
    {return Rectangle(left_x, right_x, top_y, bottom_y + val);}
  Rectangle shrinkLeft(float val) const
    { return expandLeft(-val); }
  Rectangle shrinkRight(float val) const
    { return expandRight(-val); }
  Rectangle shrinkTop(float val) const
    { return expandTop(-val); }
  Rectangle shrinkBottom(float val) const
    { return expandBottom(-val); }
  
  friend ostream& operator <<(ostream & os, Rectangle r)
    { return os << "RECTANGLE:  x = " << r.LeftX() << " to " << r.RightX()
		<< "   y = " << r.TopY() << " to " << r.BottomY(); }
  void Print() const;

  Vector RayIntersection(Ray r) const;

private:
  float left_x ;
  float right_x ;
  float top_y ;
  float bottom_y ;

} ;

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                               Circle Class                                   */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

class Circle {
public:
  Circle(Vector c, float r) : center(c), radius(r) { }
  bool isWithin(Vector u) const {return (u.dist2(center) < radius * radius)
				   ? true : false; }

  int RayIntersect(Ray r, Vector* psol1, Vector* psol2) const;
  int LineIntersect(Line l, Vector *psol1, Vector* psol2) const
    { return LineCircleIntersect(l, *this, psol1, psol2); }
  int LineSegmentIntersect(Vector pt1, Vector pt2, Circle c, Vector* psol1, Vector* psol2);
  
  Vector GetClosestPoint(Vector pt) const;

  friend ostream& operator <<(ostream & os, Circle c)
  { return os << "C[c=" << c.center << ", r=" << c.radius << "]"; }

  Circle shrink(float val) { return expand(-val); }
  Circle expand(float val) { return Circle(center, radius+val); }
  
  Vector center;
  float radius;
};

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                               Ellipse Class                                   */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

class Ellipse {
public:
  Ellipse() : center(0,0), A(1), B(1) {}
  Ellipse(Vector c, float xrad, float yrad) : center(c), A(1.0/Sqr(xrad)), B(1.0/Sqr(yrad)) { }
  //Ellipse(Vector center1, Vector center2);

  bool valid() const { return A>=0.0 && B>=0.0; }
  
  float Xrad() const { return sqrt(1.0/A); }
  float Yrad() const { return sqrt(1.0/B); }

  bool isXAxesLonger() const { return (A < B); }
  
  //alpha is the ratio of the previous radius to the next starting point
  //small alpha give more circles and better coverage
  //large alpha gives less circle and worse coverage
  vector<Circle> approxByCirclesOnMajor(float alpha = 1.0) const;
  //vector<Circle> approxByCirclesWithPotential() const;
  vector<Vector> approxByTrapezoids() const;

  float YHeightAtXVal(float x) const { return sqrt((1 - A*Sqr(x - center.x) ) / B); }
  float XHeightAtYVal(float y) const { return sqrt((1 - B*Sqr(y - center.y) ) / A); }
  
  bool isWithin(Vector pt) const { return evalAt(pt) <= 0; }
  float evalAt(Vector pt) const { return A*Sqr(pt.x-center.x) + B*Sqr(pt.y-center.y) - 1; }
  
  friend ostream& operator <<(ostream & os, Ellipse e)
  { return os << "(E "<<e.center<<" "<<e.A<<" "<<e.B<<")"; }
  bool ReadFromStr(char* str)
  { istrstream i(str); A=B=-1; i >> *this; return valid(); }
  friend istream& operator >>(istream & is, Ellipse &e);  
  void Print() const { cout << *this << endl; } 

  //A x^2 + B y^2 = 1
  Vector center;
  float A,B;
};

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        RectGrid Class                                        */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

class RectGrid {
public:
  RectGrid();
  void Initialize(float pitch_length, float pitch_width,
		  int num_rows, int num_cols);

  ~RectGrid();

  void Zero();

  void GetRectOfPoint(float x, float y, int* pr, int* pc);
  void GetRectOfPoint(Vector pos, int* pr, int* pc);
  
  float IncrementRect(int r, int c, float by=1.0, bool invert = FALSE);
  float IncrementRectForPoint(float x, float y, float by=1.0, bool invert = FALSE);
  float IncrementRectForPoint(Vector pos, float by=1.0, bool invert = FALSE);
  float IncrementRect(int r, int c, bool invert)
    { return IncrementRect(r, c, 1.0, invert); }
  float IncrementRectForPoint(float x, float y, bool invert)
    { return IncrementRectForPoint(x, y, 1.0, invert); }
  float IncrementRectForPoint(Vector pos, bool invert)
    { return IncrementRectForPoint(pos, 1.0, invert); }

  void GetInfoOfTrajectory(Vector start, Vector end, float* pAvg, float* pMax);

  float Sum();
  float AverageX();
  float AverageY();
  
  void WriteToFile(ostream& outfile, int prec = 0);
  friend ostream& operator <<(ostream & os, RectGrid& rg)
    { rg.WriteToFile(os); return os; }

  bool ReadFromFile(istream& infile);  

  /* returns valus in range [0,1] */
  /* M is an additional squahsing constant, the bigger it is, the more similar
     things will appear */
  static const float default_M = 1.0;
  float SimilarityTo(RectGrid* prg, float M = default_M);  
  //float SimilarityTo(RectGrid& rg, float M = 1.0)
  //{ return SimilarityTo(&rg, M); }

  /* the weight is for the argument object; the this object always has weight 1.0 */
  void CombineWith(RectGrid* prg, float weight = 1.0);

  void GenerateRandom(int num_points);
  void GenerateConnectedRandom(int num_points);
  
  float pitch_length;
  float pitch_width;
  int num_rows;
  int num_cols;

  friend RectGridMulti;
  
private:
  float** vals;

  float GravSimFunc(RectGrid* prg, float scale, float M);  
} ;

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                      RectGridMulti Class                                     */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

/* like RectGrid, but stores multiple values at each point */

class RectGridMulti {
public:
  RectGridMulti();

  void Initialize(float pitch_length, float pitch_width,
		  int num_rows, int num_cols, int num_vals);

  ~RectGridMulti();

  void Zero();

  void GetRectOfPoint(float x, float y, int* pr, int* pc);
  void GetRectOfPoint(Vector pos, int* pr, int* pc);
  
  float IncrementRect(int r, int c, int which, float by=1.0, bool invert = FALSE);
  float IncrementRectForPoint(float x, float y, int which, float by=1.0, bool invert = FALSE);
  float IncrementRectForPoint(Vector pos, int which, float by=1.0, bool invert = FALSE);
  float IncrementRect(int r, int c, int which, bool invert)
    { return IncrementRect(r, c, which, 1.0, invert); }
  float IncrementRectForPoint(float x, float y, int which, bool invert)
    { return IncrementRectForPoint(x, y, which, 1.0, invert); }
  float IncrementRectForPoint(Vector pos, int which, bool invert)
    { return IncrementRectForPoint(pos, which, 1.0, invert); }

  void RecordTrajectory(Vector start, Vector end, bool invert = FALSE);

  float Sum();
  
  void WriteToFile(ostream& outfile, int prec = 0);
  friend ostream& operator <<(ostream & os, RectGridMulti rgm)
    { rgm.WriteToFile(os); return os; }

  bool ReadFromFile(istream& infile);  

  /* returns valus in range [0,1] */
  /* M is an additional squashing constant, the bigger it is, the more similar
     things will appear */
  static const float default_M = 1.0;
  float SimilarityTo(RectGridMulti* prgm, float M = default_M);  
  float SimilarityTo(RectGridMulti& rgm, float M = default_M)
    { return SimilarityTo(&rgm, M); }

  /* the weight is for the argument object; the this object always has weight 1.0 */
  void CombineWith(RectGridMulti* prgm, float weight = 1.0);

  void GenerateRandom(int num_points);
  void GenerateConnectedRandom(int num_points);
  
  RectGrid* SquashToRectGrid();

  float pitch_length;
  float pitch_width;
  int num_rows;
  int num_cols;
  int num_vals; /* the number of values at each square */

private:

  /* the storage layout is one long array of ints. The Ref function makes it
     easier to think about it being a three dimensional array */
  float* Ref(int r, int c, int which);
  float* vals;

} ;


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                  Miscellaneous Geometry Functions                            */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

inline float GetDistToLineSegment(Vector pt, Vector end1, Vector end2) 
{ Line l = LineFromTwoPoints(end1, end2); return l.GetDistToClosestPtInBetween(pt, end1, end2); }

/* These definitions have to go down here because otherwise we get incomplete
   type errors */
inline int Ray::CircleIntersect(Circle c, Vector* psol1, Vector* psol2) const
{ return RayCircleIntersect(*this, c, psol1, psol2); }

inline int Circle::RayIntersect(Ray r, Vector* psol1, Vector* psol2) const
{ return r.CircleIntersect(*this, psol1, psol2); }

inline int  Line::CircleIntersect(Circle c, Vector *psol1, Vector* psol2) const
{ return LineCircleIntersect(*this, c, psol1, psol2); }

inline int  Line::RectangleIntersect(Rectangle r, Vector *psol1, Vector* psol2) const
{ return LineRectangleIntersect(*this, r, psol1, psol2); }

#endif

