/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Itsuki Noda,	Mikhail Prokopenko, Jan Wendler 
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* This is a modified version of a file distributed with the SoccerServer system
   http://sserver.sourceforge.net/
   Patrick Riley (pfr@cs.cmu.edu) 08/2001
*/


/* This files defines a class region to specify regions for the coach messages */

#include <iomanip.h>

#include "region.h"

#include "coach_mess.h"
#include "data.h"

#ifdef WITH_DRAW
#include "FieldImage.h"
using namespace Magick;
#endif

const float REGION_BORDER_STROKE_WIDTH = .4;

/****** utility function *******/
RegionPtr convertEllipseToRegion(Ellipse e)
{
  vector<Circle> vC = e.approxByCirclesOnMajor();
  vector<RegionPtr> vReg;
  //if (v.size > 5) {
  /* pfr 6/7/01: I think I like the trapezoid approximations better in general,
     so we'll use those */
  if (vC.size() > 1) {
    //This ellipse is really flat. Let's approximate by quadrangle instead
    vector<Vector> vV = e.approxByTrapezoids();
    vector<Vector>::iterator i = vV.begin();
    RegPoint *p1, *p2, *p3, *p4;
    p1 = new RegPointSimple(*i); i++;
    p2 = new RegPointSimple(*i); i++;
    p3 = new RegPointSimple(*i); i++;
    p4 = new RegPointSimple(*i); i++;
    vReg.push_back(RegionPtr(new RegQuad(p1, p2, p3, p4)));
    p1 = new RegPointSimple(*i); i++;
    p2 = new RegPointSimple(*i); i++;
    p3 = new RegPointSimple(*i); i++;
    p4 = new RegPointSimple(*i); i++;
    vReg.push_back(RegionPtr(new RegQuad(p1, p2, p3, p4)));
  } else {
    for (vector<Circle>::iterator i = vC.begin();
	 i != vC.end();
	 ++i) {
      vReg.push_back(RegionPtr(new RegArc(*i)));
    }
  }
  return RegionPtr(new RegUnion(vReg));
  return RegionPtr(NULL); //gcc 2.95.3 gives a spurios warning about no return here
}

void printIsPointIn(ostream& out, Region* pReg,
		    double xmin, double xmax, double xstep,
		    double ymin, double ymax, double ystep)
{
  for (double y=ymin; y<=ymax; y+=ystep) {
    for (double x=xmin; x<=xmax; x+=xstep) {
      if (pReg->isPointIn(Vector(x,y)))
	out << 'X';
      else
	out << '.';
    }
    out << endl;
  }
}

void printIsPointInFuzzy(ostream& out, Region* pReg, double max_dist,
			 double xmin, double xmax, double xstep,
			 double ymin, double ymax, double ystep)
{
  for (double y=ymin; y<=ymax; y+=ystep) {
    for (double x=xmin; x<=xmax; x+=xstep) {
      out << setprecision(2) << setw(5)
	  << pReg->isPointInFuzzy(Vector(x,y), max_dist);
    }
    out << endl;
  }
}

void printPointsOf(ostream& out, Region* pReg, double dist_pt)
{
  RegionIterator* pRI;
  for (pRI = pReg->newIterator(dist_pt);
       pRI->valid();
       (*pRI)++) {
    //out << pRI->getPt() << endl;
    Vector v = pRI->getPt();
    out << v.x << " " << v.y << endl;
  }
  
  delete pRI;
}

void printPointsOfReverse(ostream& out, Region* pReg, double dist_pt)
{
  RegionIterator* pRI = pReg->newIterator(dist_pt);;
  for (pRI->resetToEnd();
       pRI->valid();
       (*pRI)--) {
    //out << pRI->getPt() << endl;
    Vector v = pRI->getPt();
    out << v.x << " " << v.y << endl;
  }
  
  delete pRI;
}


/* given a dist from the region and the max dist, returns a 0.0 - 1.0 number
   for how close to the region is */
/* right now this is a linear interpolation, but we coudl do something better */
double fuzzy_distance_function(double dist, double max_dist)
{
  return MinMax(0.0, 1.0 - dist/max_dist, 1.0);
}


/********************** Region **********************/
/* static member element */
CoachMessageEvaluator* Region::pEval = NULL;
void Region::setEvaluator(CoachMessageEvaluator* pe)
{ if (pEval) delete pEval; pEval = pe; }
CoachMessageEvaluator* Region::getEvaluator() { return pEval; }
void Region::forgetEvaluator() { pEval = NULL; }

double Region::isPointInFuzzy(Vector pt, double max_dist)
{
  float d;
  getClosestPointTo(pt, &d);
  return fuzzy_distance_function(d, max_dist);
}

bool Region::check()
{
  pEval->resetErrorFlag();
  //this is actually guarunteed to work because RegUnion does short-circuit eval
  isPointIn(Vector(HUGE_VAL,HUGE_VAL));
  return !pEval->wasError();
}



/********************** RegQuad **********************/
RegQuad::RegQuad(const Rectangle& r) 
{
  pts[0] = new RegPointSimple(r.TopLeftCorner());
  pts[1] = new RegPointSimple(r.TopRightCorner());
  pts[2] = new RegPointSimple(r.BottomRightCorner());
  pts[3] = new RegPointSimple(r.BottomLeftCorner());
  updateStoredValues();
}


RegQuad::~RegQuad()
{
  if (pts[0]) delete pts[0];
  if (pts[1]) delete pts[1];
  if (pts[2]) delete pts[2];
  if (pts[3]) delete pts[3];
}

bool RegQuad::setPt(int i, RegPoint* pt)
{
  if (i < 0 || i > 3)
    return false;
  pts[i] = pt;
  correctPointOrder();
  return true;
}

void RegQuad::setAllPts(RegPoint* pt0, RegPoint* pt1, RegPoint* pt2, RegPoint* pt3)
{
  if (pts[0]) delete pts[0];
  if (pts[1]) delete pts[1];
  if (pts[2]) delete pts[2];
  if (pts[3]) delete pts[3];
  pts[0] = pt0;
  pts[1] = pt1;
  pts[2] = pt2;
  pts[3] = pt3;
  correctPointOrder();
}

/* Here we have to check the two triangles 012 and 023 */
bool RegQuad::isPointIn(Vector pt)
{
  /*updateStoredValues();*/

  bool inside = true; 
  if (l02.eval(pt) < 0) {
    //check the 012 triangle
    if (inside && l01.eval(pt) > 0) inside = false;
    if (inside && l12.eval(pt) > 0) inside = false;
  } else {
    //check the 023 triangle
    if (inside && l30.eval(pt) > 0) inside = false;
    if (inside && l23.eval(pt) > 0) inside = false;
  }
  return inside;
}

Vector RegQuad::getSomePointIn()
{
  /*updateStoredValues();*/
  return midpoint(pts[0]->toVector(), pts[2]->toVector());
}

Vector RegQuad::getClosestPointTo(Vector pt, float *pdist)
{
  /*updateStoredValues();*/

  if (isPointIn(pt)) {
    if (pdist) *pdist = 0;
    return 1.0;
  }
  
  Vector closepts[4];
  closepts[0] = l01.GetClosestPtInBetween(pt, pts[0]->toVector(), pts[1]->toVector());
  closepts[1] = l12.GetClosestPtInBetween(pt, pts[1]->toVector(), pts[2]->toVector());
  closepts[2] = l23.GetClosestPtInBetween(pt, pts[2]->toVector(), pts[3]->toVector());
  closepts[3] = l30.GetClosestPtInBetween(pt, pts[3]->toVector(), pts[0]->toVector());

  MinMaxSummary dists;
  for(int i=0; i<4; i++)
    dists.AddPoint(closepts[i].dist(pt));
  
  if (pdist)
    *pdist = dists.GetMin();

  return closepts[dists.GetMinIndex()];
}

bool RegQuad::entirelyXLess(float v) 
{
  return (pts[0]->entirelyXLess(v) &&
	  pts[1]->entirelyXLess(v) &&
	  pts[2]->entirelyXLess(v) &&
	  pts[3]->entirelyXLess(v));
}

bool RegQuad::entirelyXGreater(float v) 
{
  return (pts[0]->entirelyXGreater(v) &&
	  pts[1]->entirelyXGreater(v) &&
	  pts[2]->entirelyXGreater(v) &&
	  pts[3]->entirelyXGreater(v));
}

bool RegQuad::entirelyYLess(float v) 
{
  return (pts[0]->entirelyYLess(v) &&
	  pts[1]->entirelyYLess(v) &&
	  pts[2]->entirelyYLess(v) &&
	  pts[3]->entirelyYLess(v));
}

bool RegQuad::entirelyYGreater(float v) 
{
  return (pts[0]->entirelyYGreater(v) &&
	  pts[1]->entirelyYGreater(v) &&
	  pts[2]->entirelyYGreater(v) &&
	  pts[3]->entirelyYGreater(v));
}

void RegQuad::Print(ostream& out) const 
{
  out << "(quad";
  for (int i=0; i<4; i++)
    out << " " << *pts[i];
  out << ")";
}

void RegQuad::PrintPretty(ostream& out, string line_header) const 
{
  out << line_header
      << "Quadrangle: ";
  for (int i=0; i<4; i++) {
    pts[i]->PrintSimple(out);
    out << " ";
  }
  out << endl;
}

void RegQuad::iterate(CoachMessageIterator* pIter, IteratorType it)
{
  if (it == IT_ParentFirst) pIter->iterRegion(this);
  for (int i=0; i<4; i++)
    if (pts[i]) pts[i]->iterate(pIter, it);
  if (it == IT_ParentLast) pIter->iterRegion(this);
}

bool RegQuad::correctPointOrder()
{
  Line l;
  l.LineFromTwoPoints(pts[0]->toVector(), pts[2]->toVector());
  if (l.HalfPlaneTest(pts[1]->toVector()) == l.HalfPlaneTest(pts[3]->toVector())) {
    //we have to rotate the points
    RegPoint* temp = pts[0];
    for (int i=0; i<=2; i++)
      pts[i] = pts[i+1];
    pts[3] = temp;
  }
  l02.LineFromTwoPoints(pts[0]->toVector(), pts[2]->toVector());
  if (l02.eval(pts[1]->toVector()) > 0) l02 = l02.negateCoeff();
  l01.LineFromTwoPoints(pts[0]->toVector(), pts[1]->toVector());
  if (l01.eval(pts[2]->toVector()) > 0) l01 = l01.negateCoeff();
  l12.LineFromTwoPoints(pts[1]->toVector(), pts[2]->toVector());
  if (l12.eval(pts[0]->toVector()) > 0) l12 = l12.negateCoeff();
  l23.LineFromTwoPoints(pts[2]->toVector(), pts[3]->toVector());
  if (l23.eval(pts[0]->toVector()) > 0) l23 = l23.negateCoeff();
  l30.LineFromTwoPoints(pts[3]->toVector(), pts[0]->toVector());
  if (l30.eval(pts[2]->toVector()) > 0) l30 = l30.negateCoeff();
  return false;
}


#ifdef WITH_DRAW
void RegQuad::draw(FieldImage* pfi, const Magick::Color* pc)
{
  std::list<Coordinate> lcoord;
  for (int i=0; i<4; i++) {
    pts[i]->draw(pfi, pc);
    lcoord.push_back(Vector2Coord(pts[i]->toVector()));
  }
  std::list<Drawable> drawList;
  drawList.push_back(DrawableStrokeWidth(REGION_BORDER_STROKE_WIDTH));
  drawList.push_back(DrawableStrokeColor(*pc));
  drawList.push_back(DrawableFillColor(*pc));
  drawList.push_back(DrawableFillOpacity((double)80));
  drawList.push_back(DrawablePolygon(lcoord));
  pfi->addDrawableObjects(drawList);  
}
#endif

RegionPtr RegQuad::clone() const
{
  RegionPtr p0 = pts[0]->clone();
  RegionPtr p1 = pts[1]->clone();
  RegionPtr p2 = pts[2]->clone();
  RegionPtr p3 = pts[3]->clone();
  return RegionPtr(new RegQuad((RegPoint*)p0.forget(), (RegPoint*)p1.forget(),
			       (RegPoint*)p2.forget(), (RegPoint*)p3.forget()));
}

void RegQuad::rotate180()
{
  for (int i=0; i<=3; i++)
    pts[0]->rotate180();
}


/* RegQuadIterator */

RegQuad::RegQuadIterator::RegQuadIterator(float d, RegQuad* p)
  : RegionIterator(d), pReg(p) 
{
  resetToBegin();
}

void RegQuad::RegQuadIterator::operator++()
{
  curr_pt += pt_step;
  if (!pReg->isPointIn(curr_pt)) {
    //need to advance the line
    curr_line = curr_line.shift_by_vector(line_step);
    //first decide which pair of lines to use
    //note that curr_pt + line_step is just some point on curr_line
    if (pReg->l02.HalfPlaneTest(pReg->pts[1]->toVector()) ==
	pReg->l02.HalfPlaneTest(curr_pt + line_step))
      setPtIterFromLines(pReg->l01, pReg->l12);
    else
      setPtIterFromLines(pReg->l23, pReg->l30);
    //we need this correction because of wonderful floating point rounding
    curr_pt += pt_step.scaleTo(FLOAT_EPS);
    if (!pReg->isPointIn(curr_pt))
      error_flag = true;
  }
}

void RegQuad::RegQuadIterator::operator--()
{
  curr_pt -= pt_step;
  if (!pReg->isPointIn(curr_pt)) {
    //need to advance the line
    curr_line = curr_line.shift_by_vector(-line_step);
    //first decide which pair of lines to use
    //note that curr_pt - line_step is just some point on curr_line
    //Here, we reverse the order of the lines to get some semblence of the smae points back
    if (pReg->l02.HalfPlaneTest(pReg->pts[1]->toVector()) ==
	pReg->l02.HalfPlaneTest(curr_pt - line_step))
      setPtIterFromLines(pReg->l12, pReg->l01);
    else
      setPtIterFromLines(pReg->l30, pReg->l23);
    pt_step = -pt_step; //to start at the end of that row
    //we need this correction because of wonderful floating point rounding
    curr_pt -= pt_step.scaleTo(FLOAT_EPS);
    if (!pReg->isPointIn(curr_pt))
      error_flag = true;
  }
}

void RegQuad::RegQuadIterator::resetToBegin()
{
  curr_line = pReg->l02.parallelThrough(pReg->pts[1]->toVector());
  line_step = curr_line.perpendicularVectorTowards(pReg->pts[3]->toVector()).scaleTo(dist_pt);
  setPtIterFromLines(pReg->l01, pReg->l12);
  error_flag = false;
}

void RegQuad::RegQuadIterator::resetToEnd()
{
  curr_line = pReg->l02.parallelThrough(pReg->pts[3]->toVector());
  //we have to use - here because we still want the line_step to point towards the
  // "end" of the rectangle
  line_step = -curr_line.perpendicularVectorTowards(pReg->pts[1]->toVector()).scaleTo(dist_pt);
  setPtIterFromLines(pReg->l23, pReg->l30);
  pt_step = -pt_step; //to start at the end of that row
  error_flag = false;
}

void RegQuad::RegQuadIterator::setPtIterFromLines(Line lA, Line lB)
{
  Vector ptA = lA.intersection(curr_line);
  Vector ptB = lB.intersection(curr_line);
  curr_pt = ptA;
  pt_step = ptB - ptA;
  if (pt_step.mod2() <= FLOAT_EPS)
    pt_step = curr_line.getVector(); //just make sure line will advance next time
  else
    pt_step = pt_step.scaleTo(dist_pt);
}




/********************** RegArc **********************/

RegArc::RegArc()
{
  center = NULL;
  start_rad = end_rad = 0.0;
  start_ang = span_ang = 0.0;
}

RegArc::RegArc(RegPoint* center, float start_rad, float end_rad, float start_ang, float span_ang)
{
  this->center = center;
  this->start_rad = start_rad;
  this->end_rad = end_rad;
  this->start_ang = start_ang;
  this->span_ang = span_ang;
}

RegArc::RegArc(Circle c)
{
  center = new RegPointSimple(c.center.x, c.center.y);
  start_rad = 0.0;
  end_rad = c.radius;
  start_ang = 0;
  span_ang = 360;
}

RegArc::~RegArc()
{
  if (center) delete center;
}

bool RegArc::isPointIn(Vector pt) 
{
  Vector c = center->toVector();
  double dist = c.dist(pt);
  if (dist < start_rad || dist > end_rad)
    return false;
  double ang = (pt - c).dir();
  ang -= start_ang;
  NormalizeAngleDeg(&ang);
  if (ang < 0) ang += 360;
  if (ang > span_ang)
    return false;
  return true;
}

Vector RegArc::getSomePointIn() 
{
  /*updateStoredValues();*/
  return center->toVector() +
    Polar2Vector((start_rad + end_rad) / 2.0, start_ang + span_ang / 2.0);
}

Vector RegArc::getClosestPointTo(Vector pt, float *pdist = NULL)
{
  Vector c = center->toVector();
  bool strict_dist_inside = true;
  bool strict_ang_inside = true;
  double dist = c.dist(pt);
  double neardist=0, nearang=0;
  Vector close_pt;
  if (dist < start_rad) {
    strict_dist_inside = false;
    neardist = start_rad;
  } else if (dist > end_rad) {
    strict_dist_inside = false;
    neardist = end_rad;
  }
  double ang = (pt - c).dir();
  ang -= start_ang;
  NormalizeAngleDeg(&ang);
  if (ang < 0) ang += 360;
  if (ang > span_ang) {
    strict_ang_inside = false;
    if (ang - span_ang < 360 - ang)
      nearang = start_ang + span_ang;
    else
      nearang = start_ang;
  }

  /* now we have the four cases for the strict_ bools */
  if (strict_dist_inside) {
    if (strict_ang_inside) {
      close_pt = pt;
    } else {
      // !strict_ang_inside
      close_pt = GetClosestPtInBetween(pt, 
				       c+Polar2Vector(start_rad, nearang),
				       c+Polar2Vector(end_rad, nearang));
    }
  } else {
    // !strict_dist_inside
    if (strict_ang_inside) {
      close_pt = c+Polar2Vector(neardist, ang+start_ang);
    } else {
      // !strict_ang_inside
      close_pt = c+Polar2Vector(neardist, nearang);
    }
  }

  if (pdist)
    *pdist = close_pt.dist(pt);
  return close_pt;
}

bool RegArc::entirelyXLess(float v) 
{
  my_error("RegArc::entirelyXLess: I am not implemented correctly!");
  return (center->toVector().x + end_rad < v);
}

bool RegArc::entirelyXGreater(float v) 
{
  my_error("RegArc::entirelyXGreater: I am not implemented correctly!");
  return (center->toVector().x - end_rad > v);
}

bool RegArc::entirelyYLess(float v) 
{
  my_error("RegArc::entirelyYLess: I am not implemented correctly!");
  return (center->toVector().y + end_rad < v);
}

bool RegArc::entirelyYGreater(float v) 
{
  my_error("RegArc::entirelyYGreater: I am not implemented correctly!");
  return (center->toVector().y - end_rad > v);
}


void RegArc::Print(ostream& out) const
{
  out << "(arc"
      << " " << *center
      << " " << start_rad 
      << " " << end_rad 
      << " " << start_ang 
      << " " << span_ang
      << ")";
}

void RegArc::PrintPretty(ostream& out, string line_header) const
{
  out << line_header
      << "Arc: "
      << "center=";
  center->PrintSimple(out);
  out << "\tradius="
      << start_rad
      << " to "
      << end_rad
      << "\tangle="
      << start_ang
      << " for "
      << span_ang
      << endl;
}

void RegArc::iterate(CoachMessageIterator* pIter, IteratorType it)
{
  if (it == IT_ParentFirst) pIter->iterRegion(this);
  if (center) center->iterate(pIter, it);
  if (it == IT_ParentLast) pIter->iterRegion(this);
}

bool RegArc::setCenter(RegPoint* pt)
{
  if (center) delete center;
  this->center = center;
  return true;
}

bool RegArc::setRad(float start_rad, float end_rad)
{
  if (start_rad > end_rad)
    return false;
  this->start_rad = start_rad;
  this->end_rad = end_rad;
  return true;
}

bool RegArc::setAng(float start_ang, float span_ang)
{
  if (start_ang > 180 || start_ang < -180)
    return false;
  this->start_ang = start_ang;
  this->span_ang = span_ang;
  return true;
}

#ifdef WITH_DRAW
void RegArc::draw(FieldImage* pfi, const Magick::Color* pc)
{
  Vector startpt, endpt;
  center->draw(pfi, pc);

  std::list<Magick::Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(*pc));
  drawList.push_back(DrawableFillColor(*pc));

  list<Magick::VPath> l;
  Vector c = center->toVector();
  //the main arc area:

  drawList.push_back(DrawableStrokeWidth(REGION_BORDER_STROKE_WIDTH));
  drawList.push_back(DrawableStrokeOpacity((double)0));
  drawList.push_back(DrawableFillOpacity((double)100));
  //The sides of the region
  if (span_ang < 360) {
    startpt = c + Polar2Vector(start_rad, start_ang);
    endpt = c + Polar2Vector(end_rad, start_ang);  
    drawList.push_back(DrawableLine(startpt.x, startpt.y, endpt.x, endpt.y));
    startpt = c + Polar2Vector(start_rad, start_ang+span_ang);
    endpt = c + Polar2Vector(end_rad, start_ang+span_ang);  
    drawList.push_back(DrawableLine(startpt.x, startpt.y, endpt.x, endpt.y));
  }
  
  /* the arc parts and the center shading
     We draw this in two segments in order to get span angles more than 180
     correct */
  float ravg = (start_rad + end_rad) / 2.0;
  float sta = start_ang;
  float spa = Min(span_ang, 180);
  for (int i=0; i<2; i++) {
    l.clear();
    if (start_rad > 1.0) {
      l.push_back(PathMovetoAbs(Vector2Coord(c + Polar2Vector(start_rad, sta))));
      endpt = c + Polar2Vector(start_rad, sta + spa);
      l.push_back(PathArcAbs(PathArcArgs(start_rad, start_rad, 0.0,
					 false, true,
					 endpt.x, endpt.y)));
      drawList.push_back(DrawableStrokeWidth(REGION_BORDER_STROKE_WIDTH));
      drawList.push_back(DrawableStrokeOpacity((double)0));
      drawList.push_back(DrawableFillOpacity((double)100));
      drawList.push_back(DrawablePath(l));
      l.clear();
    }
    l.push_back(PathMovetoAbs(Vector2Coord(c + Polar2Vector(end_rad, sta))));
    endpt = c + Polar2Vector(end_rad, sta + spa);
    l.push_back(PathArcAbs(PathArcArgs(end_rad, end_rad, 0.0,
				       false, true,
				       endpt.x, endpt.y)));
    drawList.push_back(DrawableStrokeWidth(REGION_BORDER_STROKE_WIDTH));
    drawList.push_back(DrawableStrokeOpacity((double)0));
    drawList.push_back(DrawableFillOpacity((double)100));
    drawList.push_back(DrawablePath(l));
    l.clear();

    l.push_back(PathMovetoAbs(Vector2Coord(c + Polar2Vector(ravg, sta))));
    endpt = c + Polar2Vector(ravg, sta + spa);
    l.push_back(PathArcAbs(PathArcArgs(ravg, ravg, 0.0,
				       false, true,
				       endpt.x, endpt.y)));
    drawList.push_back(DrawableStrokeWidth(end_rad - start_rad));
    drawList.push_back(DrawableStrokeOpacity((double)80));
    drawList.push_back(DrawableFillOpacity((double)100));
    drawList.push_back(DrawablePath(l));
    l.clear();
    
    if (span_ang < 180)
      break;
    sta += spa;
    spa = span_ang - 180;
    if (spa <= 1.0)
      break;
  }
  
  //now actually draw all this stuff!
  pfi->addDrawableObjects(drawList);
}
#endif

void RegArc::rotate180()
{
  center->rotate180();
  span_ang += 180;
  NormalizeAngleDeg(&span_ang);
}


/* RegArcIterator */
RegArc::RegArcIterator::RegArcIterator(float d, RegArc* p)
  : RegionIterator(d), pReg(p)
{  
  resetToBegin();
}

Vector RegArc::RegArcIterator::getPt() const
{
  return pReg->center->toVector() +
    Polar2Vector(curr_dist, pReg->start_ang+curr_ang);
}

void RegArc::RegArcIterator::operator++()
{
  curr_ang += curr_ang_incr;
  if (curr_ang > pReg->span_ang) {
    //advance to the next row
    curr_ang = 0;
    curr_dist += dist_pt;
    if (curr_dist > pReg->end_rad)
      error_flag = true;
    curr_ang_incr = get_ang_incr_for_dist(curr_dist);
  }
}

void RegArc::RegArcIterator::operator--()
{
  curr_ang -= curr_ang_incr;
  if (curr_ang < 0) {
    //advance to the next row
    curr_ang = pReg->span_ang;
    curr_dist -= dist_pt;
    if (curr_dist < pReg->start_rad)
      error_flag = true;
    curr_ang_incr = get_ang_incr_for_dist(curr_dist);
  }
}

void RegArc::RegArcIterator::resetToBegin()
{
  curr_ang = 0;
  curr_dist = pReg->start_rad;
  curr_ang_incr = get_ang_incr_for_dist(curr_dist);
  error_flag = false;
}

void RegArc::RegArcIterator::resetToEnd()
{
  curr_ang = pReg->span_ang;
  curr_dist = pReg->end_rad;
  curr_ang_incr = get_ang_incr_for_dist(curr_dist);
  error_flag = false;
}

float RegArc::RegArcIterator::get_ang_incr_for_dist(float d)
{
  //This is a law of cosines calculation
  float tmp = 2.0 * Sqr(d);
  float tmp2 =  (tmp - Sqr(dist_pt)) / tmp;
  if (tmp2 < -1.0)
    return 500.0; //anything greater than 180 so the ++ and -- see that it's not valid
  return Rad2Deg(acos(tmp2));
}



/********************** RegUnion **********************/
RegUnion::~RegUnion()
{
  /* smartptr, don't need to delete! */
  /*
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    if (*iter != NULL) delete *iter;
  */
}

bool RegUnion::isPointIn(Vector pt)
{
  /*updateStoredValues();*/
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    if ((*iter)->isPointIn(pt))
      return true;
  return false;
}

Vector RegUnion::getClosestPointTo(Vector pt, float *pdist = NULL)
{
  /*updateStoredValues();*/
  MinMaxSummary results;
  vector<Vector> close_pts;
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter) {
    float d;
    close_pts.push_back((*iter)->getClosestPointTo(pt, &d));
    results.AddPoint(d);
  }
  if (pdist)
    *pdist = results.GetMin();
  return close_pts[results.GetMinIndex()];
}

bool RegUnion::entirelyXLess(float v) 
{
  /*updateStoredValues();*/
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    if (!(*iter)->entirelyXLess(v))
      return false;
  return true;
}

bool RegUnion::entirelyXGreater(float v) 
{
  /*updateStoredValues();*/
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    if (!(*iter)->entirelyXGreater(v))
      return false;
  return true;
}

bool RegUnion::entirelyYLess(float v) 
{
  /*updateStoredValues();*/
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    if (!(*iter)->entirelyYLess(v))
      return false;
  return true;
}

bool RegUnion::entirelyYGreater(float v) 
{
  /*updateStoredValues();*/
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    if (!(*iter)->entirelyYGreater(v))
      return false;
  return true;
}


void RegUnion::Print(ostream& out) const
{
  out << "(reg";
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    out << " " << **iter;
  out << ")";
}
void RegUnion::PrintPretty(ostream& out, string line_header) const
{
  out << line_header << "Region Union:" << endl;
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->PrintPretty(out, line_header + "o ");
}

void RegUnion::linkNamedElements(const CoachMessageRepository* pCMR)
{
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->linkNamedElements(pCMR);
}

void RegUnion::updateNamedRegion(const char* name, RegionPtr p)
{
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->updateNamedRegion(name, p);
}

void RegUnion::iterate(CoachMessageIterator* pIter, IteratorType it)
{
  if (it == IT_ParentFirst) pIter->iterRegion(this);
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->iterate(pIter, it);
  if (it == IT_ParentLast) pIter->iterRegion(this);
}

RegionPtr RegUnion::getRegion(unsigned i) const
{
  if (i < 0 || i >= vecReg.size())
    return RegionPtr();
  return vecReg[i];
}

bool RegUnion::addRegion(RegionPtr r)
{
  vecReg.push_back(r);
  return true;
}

bool RegUnion::replaceRegion(unsigned i, RegionPtr r)
{
  if (i < 0 || i >= vecReg.size())
    return false;
  //delete vecReg[i];/* smartptr, don't need to delete! */
  vecReg[i] = r;
  return true;
}

void RegUnion::setRegions(vector<RegionPtr>& vecReg) 
{
  /* smartptr, don't need to delete! */
  /*
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
  delete (*iter); */
  this->vecReg = vecReg;
}

void RegUnion::updateStoredValues()
{
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->updateStoredValues();
}

RegionPtr RegUnion::clone() const
{
  vector<RegionPtr> newV;
  vector<RegionPtr>::const_iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    newV.push_back((*iter)->clone());
  return RegionPtr(new RegUnion(newV));
  return RegionPtr(); //to avoid a warning about no return
}

void RegUnion::rotate180()
{
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->rotate180();
}


#ifdef WITH_DRAW
void RegUnion::draw(FieldImage* pfi, const Magick::Color* pc)
{
  vector<RegionPtr>::iterator iter;
  for (iter = vecReg.begin(); iter != vecReg.end(); ++iter)
    (*iter)->draw(pfi, pc);
}
#endif

/* RegUnionIterator */
RegUnion::RegUnionIterator::RegUnionIterator(float d, RegUnion* p)
  : RegionIterator(d), pReg(p), lIter()
{
  vector<RegionPtr>::const_iterator iter;
  for (iter = pReg->vecReg.begin(); iter != pReg->vecReg.end(); ++iter)
    lIter.push_back((*iter)->newIterator(dist_pt));
  resetToBegin();
}

RegUnion::RegUnionIterator::~RegUnionIterator()
{
  list<RegionIterator*>::iterator iter;
  for (iter = lIter.begin(); iter != lIter.end(); ++iter)
    delete (*iter);
}

/* we maintain an invariant that iter_curr always points to a valid
   iterator, unless error_flag is set */
void RegUnion::RegUnionIterator::operator++()
{
  (**iter_curr)++;
  forward_to_valid();
}

void RegUnion::RegUnionIterator::operator--()
{
  (**iter_curr)--;
  backward_to_valid();
}

void RegUnion::RegUnionIterator::resetToBegin()
{
  iter_curr = lIter.begin();
  error_flag = false;
  (*iter_curr)->resetToBegin();
  forward_to_valid();
}

void RegUnion::RegUnionIterator::resetToEnd()
{
  iter_curr = lIter.end();
  iter_curr--;
  (*iter_curr)->resetToEnd();  
  error_flag = false;
  backward_to_valid();
}

void RegUnion::RegUnionIterator::forward_to_valid()
{
  while (!(*iter_curr)->valid()) {
    iter_curr++;
    if (iter_curr == lIter.end()) {
      error_flag = true;
      break;
    }
    (*iter_curr)->resetToBegin();
  }  
}

void RegUnion::RegUnionIterator::backward_to_valid()
{
  while (!(*iter_curr)->valid()) {
    if (iter_curr == lIter.begin()) {
      error_flag = true;
      break;
    }
    iter_curr--;
    (*iter_curr)->resetToEnd();
  }
}



/********************** RegNamed **********************/
void RegNamed::linkNamedElements(const CoachMessageRepository* pCMR)
{
  pReg = pCMR->lookupRegion(name);
}
void RegNamed::iterate(CoachMessageIterator* pIter, IteratorType it)
{
  if (it == IT_ParentFirst) pIter->iterRegion(this);
  if (pReg) pReg->iterate(pIter, it);
  if (it == IT_ParentLast) pIter->iterRegion(this);
}

/********************** RegPoint **********************/

Vector RegPoint::getClosestPointTo(Vector pt, float *pdist = NULL)
{
  Vector v = toVector();
  if (pdist)
    *pdist = v.dist(pt);
  return v;
}


#ifdef WITH_DRAW
void RegPointSimple::draw(FieldImage* pfi, const Magick::Color* pc)
{
  pfi->addPoint(toVector(), *pc);
}
#endif

void RegPointRelative::PrintPretty(ostream& out, string line_header) const
{
  out << line_header << "Point-Relative(" << x << ", " << y << " ";
  if (pt)
    pt->PrintSimple(out);
  else
    out << "NULL";
  out << ")" << endl;
}

void RegPointRelative::PrintSimple(ostream& out) const
{
  out << "Point-Relative(" << x << ", " << y << " ";
  if (pt)
    pt->PrintSimple(out);
  else
    out << "NULL";
  out << ")" << endl;
}

void RegPointRelative::iterate(CoachMessageIterator* pIter, IteratorType it)
{
  if (it == IT_ParentFirst) pIter->iterRegion(this);
  pt->iterate(pIter, it);
  if (it == IT_ParentLast) pIter->iterRegion(this);
}

#ifdef WITH_DRAW
void RegPointRelative::draw(FieldImage* pfi, const Magick::Color* pc)
{
  Vector relpt = pt->toVector();
  Vector end_pt = toVector();
  pfi->addPoint(end_pt, *pc);
  pfi->addArrow(relpt, end_pt, *pc, .5);
  pt->draw(pfi, pc);
}
#endif

#ifdef WITH_DRAW
void RegPointBall::draw(FieldImage* pfi, const Magick::Color* pc)
{
  Vector pt = toVector();
  pfi->addPoint(pt, *pc);
  std::list<Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(Magick::Color("black")));
  drawList.push_back(DrawableFillColor(Magick::Color("black")));
  drawList.push_back(DrawablePointSize(2));
  drawList.push_back(DrawableText(pt.x, pt.y, "B"));
  pfi->addDrawableObjects(drawList);
}
#endif

Vector RegPointBall::toVector() const
{
  return getEvaluator()->getBallPos();
}

#ifdef WITH_DRAW
void RegPointPlayer::draw(FieldImage* pfi, const Magick::Color* pc)
{
  Vector pt = toVector();
  pfi->addPoint(pt, *pc);
  std::list<Magick::Drawable> drawList;
  drawList.push_back(DrawableStrokeColor(Magick::Color("black")));
  drawList.push_back(DrawableFillColor(Magick::Color("black")));
  drawList.push_back(DrawablePointSize(2));
  char str[4];
  sprintf(str, "%c%d", (our_side ? 'T' : 'O'), num);
  drawList.push_back(DrawableText(pt.x, pt.y, str));
  pfi->addDrawableObjects(drawList);
}
#endif

Vector RegPointPlayer::toVector() const
{
  return getEvaluator()->getPlayerPos(our_side, num);
}

