/* -*- Mode: C++ -*- */

/*
    libscoach: Soccer Coach library for use with the SoccerServer system 

    Copyright (C) 2001  Patrick Riley, Paul Carpenter, Gal Kaminka, Manuela Veloso
    Copyright (C) 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Itsuki Noda,	Mikhail Prokopenko, Jan Wendler 
    Copyright (C) 2000, 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda,
	Mikhail Prokopenko, Jan Wendler 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* This is a modified version of a file distributed with the SoccerServer system
   http://sserver.sourceforge.net/
   Patrick Riley (pfr@cs.cmu.edu) 08/2001
*/

/* This file defines a class region to specify regions for the coach messages */

//This must be up here to avoid weird circular reference stuff
#include "region_ptr.h"

#ifndef _H_REGION
#define _H_REGION

#include <stream.h>
#include <vector>
#include <list>
#include <string>

#include "geometry.h"
#include "coach_lang_iter.h"

#ifdef WITH_DRAW
class FieldImage;
namespace Magick {
  class Color;
}
#endif

class CoachMessageEvaluator;

//using namespace std;

const int max_name_len = 40;

/* a small value that used as a buffer to determine if two floating point numbers
   are the same. */
//const double FLOAT_EPS = .000001;

class CoachMessageRepository; //define in coach_mess.h

enum RegionType {
  RT_Null,
  RT_PointSimple,
  RT_PointRelative,
  RT_PointBall,
  RT_PointPlayer,
  RT_Quad,
  RT_Arc,
  RT_Union,
  RT_Named
};

/* A scheme must be set up so that updateStoredValues is called on all the regions
   every time any of the relative points can change */

/******************************************************************************/
/* In general, you get the iterator from the Region* and then ++iter until !iter.valid()
   *iter gives you a Vector for the current point 
   Important notes:
   * Going forwards and backwards may give you different points
   * The point distance is only approximate
   * Don't forget to delete the iterator when you are done!
   * In RegUnion, the same point can come up multiple times
   * Calling getPt on an iterator that is not valid has undefined results
   * check out the printPointsOf and printPointsOfReverse functions for examples
   * IMPORTANT: a RegionIterator is only valid in the cycle in which it is created!
                this is because the the relative points can can affect regions
*/
class RegionIterator 
{
public:
  RegionIterator(float d = 1.0)
  { this->dist_pt = d; error_flag = false; }
  virtual ~RegionIterator() {}

  virtual Vector getPt() const = 0;
  virtual void operator++() = 0;
  void operator++(int) { ++(*this); }
  virtual void operator--() = 0;
  void operator--(int) { --(*this); }

  //iter.valid == false iff the iterator was advanced past one end or the other
  bool valid() const { return !error_flag; }

  //it is the the inherited class's responsibility to reset error_flag
  virtual void resetToBegin() = 0;
  virtual void resetToEnd() = 0;

  void setDistBetweenPoints(float d) { dist_pt = d; }
  float getDistBetweenPoints(float d) const { return dist_pt; }

protected:
  float dist_pt; //the distance requested between points
  bool error_flag;
};

/******************************************************************************/
class Region;
class RegNull;
class RegQuad;
class RegArc;
class RegUnion;
class RegNamed;
class RegPoint;
class RegPointSimple;
class RegPointRelative;
class RegPointBall;
class RegPointPlayer;

class RegionPtr; //defined in region_ptr.h

class Region 
{
public:
  Region() {reference_count = 0;}
  virtual ~Region() { if (reference_count!=0) my_error("uh oh, you are deleting a region which still has references (%d)", reference_count); }

  virtual RegionType getType() const = 0;

  virtual bool isPointIn(Vector pt) = 0;
  //max_dist is the maximum from the region for a non-zero value
  //this just used the getClosestPointTo call below
  double isPointInFuzzy(Vector pt, double max_dist);

  /* This function returns some point inside of the region
     It's really just intended for testing */
  virtual Vector getSomePointIn() = 0;

  virtual Vector getClosestPointTo(Vector pt, float *pdist = NULL) = 0;
  
  virtual bool entirelyXLess(float v)  = 0;
  virtual bool entirelyXGreater(float v)  = 0;
  virtual bool entirelyYLess(float v)  = 0;
  virtual bool entirelyYGreater(float v)  = 0;
  
  virtual void Print(ostream& out) const = 0;
  virtual void PrintPretty(ostream& out, string line_header) const = 0;
  friend ostream& operator <<(ostream & os, const Region& r) 
    { r.Print(os); return os; } 

  virtual void linkNamedElements(const CoachMessageRepository* pCMR) {}
  virtual void updateNamedRegion   (const char* name, RegionPtr p) {}

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual RegionPtr clone() const = 0;

  virtual void rotate180() = 0;
  
  //d is the desired distance between points
  virtual RegionIterator* newIterator(float d) = 0;

  //this function runs a simple evaluation to see if all relative points are known
  // return true if region is good
  bool check();
  
  static void setEvaluator(CoachMessageEvaluator* pe);
  static CoachMessageEvaluator* getEvaluator();
  static void forgetEvaluator();

#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi, const Magick::Color* pc) = 0;
#endif

  //this function updates any stored values the regions have to improve
  //speed of access
  virtual void updateStoredValues() = 0;

protected:
  static CoachMessageEvaluator* pEval;

  //these friend are needed because they point to other regions
  friend RegNamed;
  friend RegUnion;

  friend RegionPtr;
  int reference_count; //this only refers to references by RegionPtr
  
} ;

class RegNull
  : public Region
{
public:
  RegNull() : Region() {}
  ~RegNull() {}

  RegionType getType() const { return RT_Null; }

  Vector getSomePointIn() 
  { my_error("Can't give you a point in the null region"); return Vector(1000,1000); }
  Vector getClosestPointTo(Vector pt, float *pdist = NULL) 
  { my_error("Can't give you closest point in the null region"); return Vector(1000,1000); }

  bool isPointIn(Vector pt) { return false; }

  bool entirelyXLess(float v)     { return true; }
  bool entirelyXGreater(float v)  { return true; }
  bool entirelyYLess(float v)     { return true; }
  bool entirelyYGreater(float v)  { return true; }

  void Print(ostream& out) const 
    { out << "(null)"; }
  void PrintPretty(ostream& out, string line_header) const 
    { out << line_header << "null region" << endl; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterRegion(this); }

  RegionPtr clone() const { return RegionPtr(new RegNull()); }

  void rotate180() {}

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc) {}
#endif

  class RegNullIterator : public RegionIterator 
  {
  public:
    RegNullIterator(float d = 1.0) : RegionIterator(d)
    { error_flag = true; }
    ~RegNullIterator() {}

    Vector getPt() const { my_error("Can't dereference a RegNullIterator"); return Vector(0,0); }
    void operator++() { my_error("Can't ++ a RegNullIterator"); }
    void operator--() { my_error("Can't -- a RegNullIterator"); }

    //the error flag should always be true, but this doesn't hurt
    void resetToBegin() { error_flag = true; } 
    void resetToEnd() { error_flag = true; } 

  };
  RegionIterator* newIterator(float d) { return new RegNullIterator(d); }

protected:
  void updateStoredValues() {}

} ;

class RegQuad
  : public Region
{
public:
  RegQuad() : Region() { pts[0] = pts[1] = pts[2] = pts[3] = NULL; }
  RegQuad(RegPoint* pt0, RegPoint* pt1, RegPoint* pt2, RegPoint* pt3) : Region()
  { pts[0] = pt0; pts[1] = pt1; pts[2] = pt2; pts[3] = pt3; correctPointOrder(); }
  //Rectangle is an axis aligned rectangle from geometry.h
  RegQuad(const Rectangle& r);
  ~RegQuad();

  RegionType getType() const { return RT_Quad; }

  bool isPointIn(Vector pt) ;

  //this takes advantage of the invariant discussed below
  Vector getSomePointIn() ; 

  Vector getClosestPointTo(Vector pt, float *pdist = NULL);

  bool entirelyXLess(float v) ;
  bool entirelyXGreater(float v) ;
  bool entirelyYLess(float v) ;
  bool entirelyYGreater(float v) ;

  void Print(ostream& out) const ;
  void PrintPretty(ostream& out, string line_header) const ;

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  RegionPtr clone() const;

  void rotate180();
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);  
#endif

  class RegQuadIterator : public RegionIterator 
  {
  public:
    RegQuadIterator(float d, RegQuad* p);
    ~RegQuadIterator() {}

    Vector getPt() const { return curr_pt; }
    void operator++();
    void operator--();

    void resetToBegin();
    void resetToEnd();

  private:
    RegQuad* pReg;
    Line curr_line;
    Vector line_step;
    Vector curr_pt;
    Vector pt_step;

    //curr_line and line_step must already be set
    void setPtIterFromLines(Line lA, Line lB);
  };
  RegionIterator* newIterator(float d)
  { /*updateStoredValues();*/ return new RegQuadIterator(d, this); }
  friend RegQuadIterator;

  /* class specific */
  RegPoint* getPt(int i) { return (i>=0 && i<=3) ? pts[i] : NULL; }
  bool     setPt(int i, RegPoint* pt);
  void     setAllPts(RegPoint* pt0, RegPoint* pt1, RegPoint* pt2, RegPoint* pt3);
  
protected:
  void updateStoredValues() { correctPointOrder(); }

  RegPoint* pts[4];
  /* for all the lines except l02, we want a negative eval to have point inclusion
     for l02, we want a negative for 012 traingle inclusion and positive for 023
     triangle */
  Line l02, l01, l12, l23, l30;
  /* an invariant we must maintain is that the line segment from the 0 to the 2
     point divides the quadrangle into 2 triangles;
     Also, all the Line vars have to be maintained;
     this function will verify/correct this */
  //returns whether the points had to be shifted
  bool correctPointOrder();
} ;

class RegArc
  : public Region
{
public:
  RegArc();
  RegArc(RegPoint* center, float start_rad, float end_rad, float start_ang, float span_ang);
  RegArc(Circle c);
  ~RegArc();

  RegionType getType() const { return RT_Arc; }

  bool isPointIn(Vector pt) ;

  Vector getSomePointIn() ;
  
  Vector getClosestPointTo(Vector pt, float *pdist = NULL);

  bool entirelyXLess(float v) ;
  bool entirelyXGreater(float v) ;
  bool entirelyYLess(float v) ;
  bool entirelyYGreater(float v) ;

  void Print(ostream& out) const ;
  void PrintPretty(ostream& out, string line_header) const ;

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  RegionPtr clone() const;

  void rotate180();
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);
#endif

  class RegArcIterator : public RegionIterator 
  {
  public:
    RegArcIterator(float d, RegArc* p);
    ~RegArcIterator() {}

    Vector getPt() const ;
    void operator++();
    void operator--();

    void resetToBegin();
    void resetToEnd();

  private:
    RegArc* pReg;
    //curr_ang will represent displacement from the start_ang
    float curr_ang;
    float curr_ang_incr; //based on curr_dist
    float curr_dist;
    float get_ang_incr_for_dist(float d);
  };
  RegionIterator* newIterator(float d) { return new RegArcIterator(d, this); }
  friend RegArcIterator;

  /* class specific */
  RegPoint* getCenter()   const { return center; }
  float    getStartRad() const { return start_rad; }
  float    getEndRad()   const { return end_rad; }
  float    getStartAng() const { return start_ang; }
  float    getSpanAng()  const { return span_ang; }
  bool     setCenter(RegPoint* pt);
  bool     setRad(float start_rad, float end_rad);
  bool     setAng(float start_ang, float span_ang);
  
protected:
  void updateStoredValues() {}

  RegPoint* center;
  /* start rad <= end_rad */
  float start_rad, end_rad;
  float start_ang, span_ang;
} ;


class RegUnion
  : public Region
{
public:
  RegUnion() : Region() { }
  RegUnion(vector<RegionPtr>& vecReg) : Region() { this->vecReg = vecReg; }
  ~RegUnion();

  RegionType getType() const { return RT_Union; }

  bool isPointIn(Vector pt) ;

  Vector getSomePointIn() 
  { return vecReg[0]->getSomePointIn(); }

  Vector getClosestPointTo(Vector pt, float *pdist = NULL);

  bool entirelyXLess(float v) ;
  bool entirelyXGreater(float v) ;
  bool entirelyYLess(float v) ;
  bool entirelyYGreater(float v) ;

  void Print(ostream& out) const ;
  void PrintPretty(ostream& out, string line_header) const ;

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedRegion   (const char* name, RegionPtr p);

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  RegionPtr clone() const;

  void rotate180();
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);
#endif

  class RegUnionIterator : public RegionIterator 
  {
  public:
    RegUnionIterator(float d, RegUnion* p);
    ~RegUnionIterator();

    Vector getPt() const { return (*iter_curr)->getPt(); }
    void operator++();
    void operator--();

    void resetToBegin();
    void resetToEnd();

  private:
    RegUnion* pReg;
    list<RegionIterator*> lIter;
    list<RegionIterator*>::iterator iter_curr;

    void forward_to_valid();
    void backward_to_valid();
  };
  RegionIterator* newIterator(float d)
  { /*updateStoredValues();*/ return new RegUnionIterator(d, this); }
  friend RegUnionIterator;

  /* class specific */
  int getNumRegions() const { return vecReg.size(); }
  RegionPtr getRegion(unsigned i) const;
  bool    addRegion(RegionPtr r);
  bool    replaceRegion(unsigned i, RegionPtr r);
  void    setRegions(vector<RegionPtr>& vecReg);
  
protected:
  void updateStoredValues();

  vector<RegionPtr> vecReg;
} ;

class RegNamed
  : public Region
{
public:
  RegNamed(const char* name = NULL) : Region(), pReg() { setName(name); }
  ~RegNamed() {}

  RegionType getType() const { return RT_Named; }

  bool isPointIn(Vector pt) 
  { if (pReg == NULL) {my_error("RegNamed::isPointIn NULL"); return false;}
  return pReg->isPointIn(pt); }

  Vector getSomePointIn() 
  { if (pReg == NULL) {my_error("RegNamed::getSomePointIn NULL"); return Vector(0,0);}
  return pReg->getSomePointIn(); }
  Vector getClosestPointTo(Vector pt, float *pdist = NULL)
  { if (pReg == NULL) {my_error("RegNamed::getClosestPointTo NULL"); return Vector(0,0);}
  return pReg->getClosestPointTo(pt, pdist); }

  bool entirelyXLess(float v) 
  { if (pReg == NULL) {my_error("RegNamed::entirelyXLess"); return false;}
  return pReg->entirelyXLess(v); }
  bool entirelyXGreater(float v) 
  { if (pReg == NULL) {my_error("RegNamed::entirelyXLess"); return false;}
  return pReg->entirelyXGreater(v); }
  bool entirelyYLess(float v) 
  { if (pReg == NULL) {my_error("RegNamed::entirelyXLess"); return false;}
  return pReg->entirelyYLess(v); }
  bool entirelyYGreater(float v) 
  { if (pReg == NULL) {my_error("RegNamed::entirelyXLess"); return false;}
  return pReg->entirelyYGreater(v); }

  void Print(ostream& out) const 
    { out << "\"" << name << "\""; }
  void PrintPretty(ostream& out, string line_header) const 
    { out << line_header << "region named \"" << name << "\"" << endl;
      if (pReg) pReg->PrintPretty(out, line_header+" "); 
      else out << line_header << " UNLINKED" << endl; }

  void iterate(CoachMessageIterator* pIter, IteratorType it);

  RegionPtr clone() const { return RegionPtr(new RegNamed(name)); }

  void rotate180()
  { if (pReg == NULL) my_error("RegNamed::rotate180 NULL"); 
    else pReg->rotate180(); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc)
  { if (pReg == NULL) my_error("Asked to draw an unlinked named region");
  else return pReg->draw(pfi, pc); }
#endif

  RegionIterator* newIterator(float d) 
  { if (pReg == NULL) {my_error("Asked for iterator on unlinked named region"); return NULL;}
  else return pReg->newIterator(d); }

  /* class specific */
  char* getName() { return name; }
  void  setName(const char* name)
  { if (name == NULL) this->name[0] = 0;
    else strncpy(this->name, name, max_name_len); this->name[max_name_len] = 0; }

  void linkNamedElements(const CoachMessageRepository* pCMR);
  void updateNamedRegion   (const char* name, RegionPtr p)
  { if (strcmp(name, this->name) == 0) { pReg = p; } }
  
protected:
  void updateStoredValues() { pReg->updateStoredValues(); }

  char name[max_name_len+1];
  RegionPtr pReg;
} ;


/**** The various RegPoint s ****/
class RegPoint
  : public Region
{
public:
  RegPoint() : Region() { }
  virtual ~RegPoint() {}

  virtual RegionType getType() const = 0;

  bool isPointIn(Vector pt) { return pt.ApproxEqual(toVector()); }

  Vector getSomePointIn() { return toVector(); }
  Vector getClosestPointTo(Vector pt, float *dist = NULL);
  
  bool entirelyXLess(float v)     { return toVector().x <= v; }
  bool entirelyXGreater(float v)  { return toVector().x >= v; }
  bool entirelyYLess(float v)     { return toVector().y <= v; }
  bool entirelyYGreater(float v)  { return toVector().y >= v; }

  virtual void Print(ostream& out) const = 0;
  virtual void PrintPretty(ostream& out, string line_header) const = 0;
  //like PrintPretty, but no line header
  virtual void PrintSimple(ostream& out) const = 0;

  virtual void iterate(CoachMessageIterator* pIter, IteratorType it) = 0;

  virtual RegionPtr clone() const = 0;

  virtual void rotate180() = 0;
  
  virtual Vector toVector() const = 0;

#ifdef WITH_DRAW
  virtual void draw(FieldImage* pfi, const Magick::Color* pc) = 0;
#endif

  class RegPointIterator : public RegionIterator 
  {
  public:
    RegPointIterator(float d, RegPoint* p) : RegionIterator(d), pReg(p) {}
    ~RegPointIterator() {}

    Vector getPt() const { return pReg->toVector(); }
    void operator++() { error_flag = true; }
    void operator--() { error_flag = true; }

    void resetToBegin() { error_flag = false; } 
    void resetToEnd() { error_flag = false; } 

  private:
    RegPoint* pReg;
  };
  RegionIterator* newIterator(float d) { return new RegPointIterator(d, this); }
  friend RegPointIterator;

protected:
  void updateStoredValues() {}

} ;

class RegPointSimple
  : public RegPoint
{
public:
  RegPointSimple(float x=0.0, float y=0.0) : RegPoint() { this->x = x; this->y = y;}
  RegPointSimple(Vector v) : RegPoint() { this->x = v.x; this->y = v.y;}
  ~RegPointSimple() {}

  RegionType getType() const { return RT_PointSimple; }

  void Print(ostream& out) const
    { out << "(pt " << x << " " << y << ")"; }
  void PrintPretty(ostream& out, string line_header) const
    { out << line_header << "Point(" << x << ", " << y << ")" << endl; }
  void PrintSimple(ostream& out) const
    { out << "Point(" << x << ", " << y << ")"; }

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterRegion(this); }

  RegionPtr clone() const { return RegionPtr(new RegPointSimple(x,y)); }

  void rotate180() { x=-x; y=-y; }

  Vector toVector() const { return Vector(x,y); }  
  
#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);  
#endif

  /* class specific */
  float getX() { return x; }
  float getY() { return y; }
  void  setPt(float x, float y) { this->x = x; this->y = y;}

private:
  float x;
  float y;
} ;

class RegPointRelative
  : public RegPoint
{
public:
  RegPointRelative(float x=0.0, float y=0.0, RegPoint* pt=NULL) : RegPoint()
   { this->x = x; this->y = y; this->pt = pt; }
  RegPointRelative(Vector v, RegPoint* pt=NULL) : RegPoint()
   { this->x = v.x; this->y = v.y; this->pt = pt; }
  ~RegPointRelative() { if (pt) delete pt; }

  RegionType getType() const { return RT_PointRelative; }

  void Print(ostream& out) const
  { out << "(pt " << x << " " << y << " " << *pt << ")"; }
  void PrintPretty(ostream& out, string line_header) const;
  void PrintSimple(ostream& out) const;
  
  void iterate(CoachMessageIterator* pIter, IteratorType it);

  RegionPtr clone() const { return RegionPtr(new RegPointRelative(x,y,(RegPoint*)pt->clone().forget())); }

  void rotate180() { x=-x; y=-y; }

  Vector toVector() const { return Vector(x,y) + pt->toVector(); }

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);  
#endif

  /* class specific */
  float getRelX() { return x; }
  float getRelY() { return y; }
  void  setRelVector(float x, float y) { this->x = x; this->y = y;}
  RegPoint* getPt() { return pt; }
  void      setPt(RegPoint* p) { if (pt) delete pt; pt = p; }
  
private:
  float x;
  float y;
  RegPoint* pt;
} ;

class RegPointBall
  : public RegPoint
{
public:
  RegPointBall() : RegPoint() {}
  ~RegPointBall() {}

  RegionType getType() const { return RT_PointBall; }

  void Print(ostream& out) const
  { out << "(pt ball)"; }
  void PrintPretty(ostream& out, string line_header) const
    { out << line_header << "Point-Ball" << endl; }
  void PrintSimple(ostream& out) const
    { out << "Point-Ball"; }    

  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterRegion(this); }

  RegionPtr clone() const { return RegionPtr(new RegPointBall()); }

  void rotate180() {}
  
  Vector toVector() const;

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);  
#endif

} ;

class RegPointPlayer
  : public RegPoint
{
public:
  RegPointPlayer(bool our_side, int num) : RegPoint()
   { this->our_side = our_side; this->num = num; }
  ~RegPointPlayer() { }

  RegionType getType() const { return RT_PointPlayer; }

  void Print(ostream& out) const
  { out << "(pt " << (our_side ? "our" : "opp") << " " << num << ")"; }
  void PrintPretty(ostream& out, string line_header) const
    { out << line_header << "Point-Player(" << (our_side ? "our team" : "opponent")
	  << " " << num << ")" << endl; }
  void PrintSimple(ostream& out) const
    { out << "Point-Player(" << (our_side ? "our team" : "opponent")
	  << " " << num << ")"; }
  
  void iterate(CoachMessageIterator* pIter, IteratorType it) { pIter->iterRegion(this); }

  RegionPtr clone() const { return RegionPtr(new RegPointPlayer(our_side, num)); }

  void rotate180() {}
  
  Vector toVector() const;

#ifdef WITH_DRAW
  void draw(FieldImage* pfi, const Magick::Color* pc);  
#endif

  /* class specific */
  bool isOurSide()  { return our_side; }
  bool isTheirSide(){ return !our_side; }
  int  getNum() { return num; }
  
  void setOurSide(bool our_side) { this->our_side = our_side; }
  void setNum(int num) { this->num = num; }
  
private:
  bool our_side;
  int  num;
} ;

/******************************************************************************/
/* inline functions */
inline RegionPtr RegArc::clone() const
{ 
  return RegionPtr(new RegArc((RegPoint*)center->clone().forget(), start_rad, end_rad,
			      start_ang, span_ang));
}

/******************************************************************************/
/****** utility function *******/

RegionPtr convertEllipseToRegion(Ellipse e);

void printIsPointIn(ostream& out, Region* pReg,
		    double xmin, double xmax, double xstep,
		    double ymin, double ymax, double ystep);

void printIsPointInFuzzy(ostream& out, Region* pReg, double max_dist,
			 double xmin, double xmax, double xstep,
			 double ymin, double ymax, double ystep);

void printPointsOf(ostream& out, Region* pReg, double dist_pt);

void printPointsOfReverse(ostream& out, Region* pReg, double dist_pt);

/******************************************************************************/

class CMI_UpdateRegion : public CoachMessageIterator
{
public:
  CMI_UpdateRegion() : CoachMessageIterator() { }
  ~CMI_UpdateRegion() {}

  void iterMessage(CM_Message* pMess) { }
  void iterMetaToken(CM_MetaToken* pMTok) { }
  void iterAction(CM_Action* pAct) { }
  void iterDirective(CM_Directive* pDir) { }
  void iterCondition(CM_Condition* pCond) { }
  void iterToken(CM_Token* pTok) { }
  void iterDefineToken(CM_DefineToken* pDTok) { }
  void iterRegion(Region* pReg) { pReg->updateStoredValues(); }

};

#endif
